package de.renew.ccpn.engine.simulator;

import de.renew.engine.common.StepIdentifier;
import de.renew.engine.simulator.Simulator;

/**
 * The simulator for the Curry Net Compiler formalism.
 *
 * Little more than a proxy that forwards all calls to the {@link CCPNProcess} object
 * and by extension to the CCPN Haskell binary.
 *
 * @author Michael Simon
 */
public class CurrySimulator implements Simulator {
    public static org.apache.log4j.Logger logger = org.apache.log4j.Logger
                    .getLogger(CurrySimulator.class);
    private CCPNProcess process;

    private static final long[] EMPTY_RUN_IDS = new long[0];

    public CurrySimulator(boolean wantEventQueueDelay) {
        this();
    }

    public CurrySimulator() {
    }

    @Override
    public boolean isActive() {
        return true;
    }

    private synchronized boolean processStopped() {
        if (logger.isDebugEnabled()) {
            logger.debug(CurrySimulator.class.getSimpleName()
                            + ": determining if CCPN process is stopped: "
                            + process);
        }
        return process == null || process.stopped;
    }

    @Override
    public synchronized void startRun() {
        if (!processStopped()) {
            process.continuousUpdate = true;
            process.run();
        }
    }

    public void setProcess(CCPNProcess process) {
        this.process = process;
    }

    @Override
    public synchronized void stopRun() {
        if (!processStopped()) {
            process.continuousUpdate = false;
            process.stop();
        }
    }

    @Override
    public synchronized void terminateRun() {
        if (process != null) {
            process.terminate();
            process = null;
        }
    }

    @Override
    public synchronized int step() {
        if (processStopped()) {
            return Simulator.statusStopped;
        }

        if (process.continuousUpdate) {
            stopRun();
        } else {
            process.step();
        }

        // Since we do not know if additional transitions are enabled
        // we just indicate that the step was completed.
        return Simulator.statusStepComplete;
    }

    @Override
    public void refresh() {
    }

    @Override
    public boolean isSequential() {
        return false;
    }

    @Override
    public StepIdentifier nextStepIdentifier() {
        /** null needed for {@link de.renew.application.SimulatorPlugin#createNetInstance(String)}. **/
        return null;
    }

    @Override
    public StepIdentifier currentStepIdentifier() {
        /** Not needed, but kept for consistency to {@link #nextStepIdentifier()}. **/
        return null;
    }

    @Override
    public long[] collectSimulationRunIds() {
        /** Not needed, but implemented for completeness sake. **/
        return EMPTY_RUN_IDS;
    }
}