package CH.ifa.draw.gui;

import java.awt.*;
import java.io.StringReader;
import java.util.*;
import java.util.List;

import javax.swing.*;


/**
 * A customizable layout for the toolbar.
 * <p>
 * Depends on the buttons having a unique name.
 */
public class StorableButtonContainerLayout {

    protected final List<List<String>> layout;
    protected final Set<String> requiredPalettes;
    protected final String layoutName;

    private StorableButtonContainerLayout(List<List<String>> layout,
                                          Set<String> requiredPalettes,
                                          String layoutName) {
        this.layout = layout;
        this.requiredPalettes = requiredPalettes;
        this.layoutName = layoutName;
    }

    @Override
    public String toString() {
        return layoutName;
    }

    @Override
    public boolean equals(Object o) {
        if (!(o instanceof StorableButtonContainerLayout)) {
            return false;
        }
        return (this.layoutName.equals(
                        ((StorableButtonContainerLayout) o).layoutName));
    }

    /**
     * Check if this layout can be used with the available palettes.
     *
     * @param availablePalettes
     * @return
     */
    public boolean isAvailable(Set<String> availablePalettes) {
        return availablePalettes.containsAll(requiredPalettes);
    }

    public static StorableButtonContainerLayout createLayout(Component[] components,
                                                             Map<String, String> groupLookup,
                                                             String layoutName) {

        List layout = new ArrayList();
        Component[] buttons;
        Set<String> requiredPalettes = new HashSet<>();

        for (Component component : components) {
            if (component instanceof ToolButtonContainer) {

                List<String> container = new ArrayList();
                buttons = ((ToolButtonContainer) component).getButtons();

                for (Component button : buttons) {
                    String name = button.getName();
                    if (name != null && name != "") {

                        container.add(name);
                        String group = groupLookup.get(name);
                        if (group != null) {
                            requiredPalettes.add(group);
                        }
                    }

                }
                layout.add(container);

            }

        }

        return new StorableButtonContainerLayout(layout, requiredPalettes,
                        layoutName);
    }

    public static List<Set<AbstractButton>> readLayout(StorableButtonContainerLayout layout,
                                                       Map<String, AbstractButton> buttonLookup) {

        List<Set<AbstractButton>> result = new ArrayList<>();

        for (List<String> set : layout.layout) {

            Set<AbstractButton> buttonSet = new LinkedHashSet<>();

            for (String name : set) {
                buttonSet.add(buttonLookup.get(name));
            }
            result.add(buttonSet);
        }
        return result;
    }

    /**
     * Creates a list of {@link StorableButtonContainerLayout} from a properties file.
     *
     * @param properties
     * @return
     */
    public static List<StorableButtonContainerLayout> createLayoutsFromProperties(Properties properties) {

        String name;
        String[] splitString;
        Set<String> requiredPalettes;
        ButtonLayoutParser parser;
        List<List<String>> layout;
        List<StorableButtonContainerLayout> result = new ArrayList<>();

        for (Map.Entry entry : properties.entrySet()) {

            try {
                name = (String) entry.getKey();
                String data = (String) entry.getValue();

                splitString = data.split("%");

                // An empty layout might have been saved, so we have to check here.
                if (splitString.length > 1) {
                    // Find required palettes
                    data = splitString[1];
                    parser = new ButtonLayoutParser(new StringReader(data));
                    requiredPalettes = parser.parseRequiredPalettes();

                    // Find containers
                    data = splitString[0];
                    parser = new ButtonLayoutParser(new StringReader(data));
                    layout = parser.parseButtons();

                    result.add(new StorableButtonContainerLayout(layout,
                                    requiredPalettes, name));
                } else {
                    result.add(new StorableButtonContainerLayout(
                                    new ArrayList<List<String>>(),
                                    new HashSet<String>(), name));
                }
            } catch (ParseException e) {
                e.printStackTrace();
            } catch (TokenMgrError error) {
                System.out.println(error.getMessage());
            }
        }

        return result;
    }

    /**
     * Write the provided layout into the {@link Properties} file.
     *
     * @param properties
     * @param layout
     */
    public static void writeToProperties(Properties properties,
                                         StorableButtonContainerLayout layout) {

        StringBuilder builder = new StringBuilder();

        for (List<String> container : layout.layout) {
            for (String button : container) {
                builder.append(button);
                builder.append(",");
            }
            builder.append(";");
        }
        builder.append("%");
        for (String group : layout.requiredPalettes) {
            builder.append(group);
            builder.append(",");
        }

        properties.put(layout.layoutName, builder.toString());
    }
}
