package CH.ifa.draw.gui;

import CH.ifa.draw.framework.Toolbar;

import bibliothek.gui.DockUI;
import bibliothek.gui.dock.util.laf.LookAndFeelColors;

import de.renew.plugin.PluginManager;
import de.renew.windowmanagement.DragCanvas;
import de.renew.windowmanagement.DraggableComponentAdapter;
import de.renew.windowmanagement.DraggableComponentDropTarget;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;
import java.util.List;
import java.util.concurrent.Semaphore;

import javax.imageio.ImageIO;
import javax.swing.*;


/**
 * I provide a container for tool buttons.
 * Buttons registered with me can be dragged and dropped within my confines.
 */
public class ToolButtonScrollPane extends DragCanvas implements Toolbar {

    private final static String images = "CH/ifa/draw/images/";
    private final static String layoutPath = "CH/ifa/draw/gui/";

    private Properties persistantLayouts;

    private final ToolButtonPanel content;

    // All registered buttons in their original groupings
    private final Map<String, Set<AbstractButton>> registeredButtons = new HashMap<>();

    // For a button name stores the original grouping the button is found in
    private final Map<String, String> groupingLookup = new HashMap<>();

    // Used to look up buttons by their name
    private final Map<String, AbstractButton> buttonLookup = new HashMap<>();

    // The tooltips for the original button groupings
    private final Map<String, String> tooltips = new HashMap<>();

    private final JComboBox<String> chooseToolsComboBox;
    private JComboBox<StorableButtonContainerLayout> chooseLayoutBox;

    // Layouts that are known but can currently not be applied because there are buttons missing
    private final Map<String, StorableButtonContainerLayout> unavailableLayouts = new HashMap<>();

    private final Semaphore semaphore = new Semaphore(1);


    /**
     * @param panel Usually a new empty panel is used.
     */
    public ToolButtonScrollPane(ToolButtonPanel panel) {

        super(panel);
        this.setMinimumSize(new Dimension(70, 0));
        this.getViewport()
                        .setScrollMode(javax.swing.JViewport.BLIT_SCROLL_MODE);
        this.getVerticalScrollBar().setUnitIncrement(20);
        this.getHorizontalScrollBar().setUnitIncrement(20);
        content = panel;
        content.setLayout(new BoxLayout(content, BoxLayout.Y_AXIS));
        chooseToolsComboBox = new JComboBox<>();
        initComboBox();
        content.add(chooseToolsComboBox);
        content.add(new CreateNewContainerTarget());
        content.add(createLayoutPanel());

        loadInitialButtonLayouts();
    }

    /**
     * Writes the properties to disk.
     */
    public void storeLayoutsToDisk() {

        try {

            File file = new File(PluginManager.getPreferencesLocation(),
                            "renewButtonLayout.properties");
            FileOutputStream outputStream = new FileOutputStream(file);
            persistantLayouts.store(outputStream, "");
            outputStream.close();

        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private void loadInitialButtonLayouts() {

        try {

            persistantLayouts = new Properties();
            File file = new File(PluginManager.getPreferencesLocation(),
                            "renewButtonLayout.properties");
            // Create new file if it doesn't exist
            file.createNewFile();
            FileInputStream stream = new FileInputStream(file);
            persistantLayouts.load(stream);
            stream.close();

            List<StorableButtonContainerLayout> layouts = StorableButtonContainerLayout
                            .createLayoutsFromProperties(persistantLayouts);
            semaphore.acquire();
            for (StorableButtonContainerLayout layout : layouts) {
                // Initially there are no buttons, so every layout should be unavailable
                unavailableLayouts.put(layout.layoutName, layout);
            }
            semaphore.release();
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
    }

    private void initComboBox() {

        chooseToolsComboBox.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                JComboBox cb = (JComboBox) e.getSource();
                String id = (String) cb.getSelectedItem();
                reAddTools(id);
            }
        });

        // Used to display tooltips on the items
        DefaultListCellRenderer renderer = new DefaultListCellRenderer() {

            @Override
            public Component getListCellRendererComponent(JList list,
                                                          Object value,
                                                          int index,
                                                          boolean isSelected,
                                                          boolean cellHasFocus) {

                JComponent comp = (JComponent) super.getListCellRendererComponent(
                                list, value, index, isSelected, cellHasFocus);

                if (index > -1 && value != null && tooltips != null) {
                    String key = chooseToolsComboBox.getItemAt(index);
                    String tooltip = tooltips.get(key);
                    list.setToolTipText(tooltip);
                }
                return comp;
            }
        };
        chooseToolsComboBox.setRenderer(renderer);
    }

    /**
     * Creates the panel with all options regarding storable layouts of the tools panel.
     */
    private JPanel createLayoutPanel() {

        final JTextField textField = new JTextField();
        textField.setEditable(true);
        final ToolButtonScrollPane handle = this;

        JButton saveButton = new JButton();
        Icon icon = new ImageIcon(loadImage(images + "buttonlayout_save.png"));
        saveButton.setIcon(icon);
        //saveButton.setText("Save");
        saveButton.setToolTipText(
                        "Saves the current layout under the specified name.");
        saveButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                String name = textField.getText();
                switch (name) {
                case "":
                    textField.setText("Enter name");
                case "Enter name":
                    break;
                default:
                    handle.saveLayout(name);
                    break;

                }
            }
        });

        JButton loadButton = new JButton();
        icon = new ImageIcon(loadImage(images + "buttonlayout_load.png"));
        loadButton.setIcon(icon);
        //loadButton.setText("Load");
        loadButton.setToolTipText("Loads the selected Layout.");
        loadButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                StorableButtonContainerLayout layout = (StorableButtonContainerLayout) chooseLayoutBox
                                .getSelectedItem();
                if (layout != null) {
                    loadLayout(layout);
                }

            }
        });

        JButton deleteButton = new JButton();
        icon = new ImageIcon(loadImage(images + "buttonlayout_delete.png"));
        deleteButton.setIcon(icon);
        deleteButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                StorableButtonContainerLayout layout = (StorableButtonContainerLayout) chooseLayoutBox
                                .getSelectedItem();
                if (layout != null) {
                    chooseLayoutBox.removeItem(layout);
                    persistantLayouts.remove(layout.layoutName);
                }
            }
        });

        chooseLayoutBox = new JComboBox<StorableButtonContainerLayout>() {

            @Override
            public void removeItem(Object item) {
                super.removeItem(item);

                if (getItemCount() == 0) {
                    setVisible(false);
                }
            }

            @Override
            public void addItem(StorableButtonContainerLayout item) {

                StorableButtonContainerLayout duplicate;
                int count = getItemCount();
                for (int i = 0; i < count; i++) {
                    duplicate = getItemAt(i);
                    if (duplicate != null) {
                        if (duplicate.equals(item)) {
                            removeItemAt(i);
                        }
                    }

                }

                super.addItem(item);
                this.setVisible(true);
                this.revalidate();
            }
        };
        // Initially we have no layouts, so the combobox is empty
        chooseLayoutBox.setVisible(false);

        JPanel panel = new JPanel();
        panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));

        JLabel label = new JLabel("Panel Layout");
        label.setHorizontalAlignment(JLabel.CENTER);
        label.setAlignmentX(Component.CENTER_ALIGNMENT);
        panel.add(label);
        panel.add(textField);

        JPanel buttonPanel = new JPanel();
        buttonPanel.setLayout(new GridLayout(0, 2));
        buttonPanel.add(saveButton);
        buttonPanel.add(loadButton);
        buttonPanel.add(deleteButton);
        panel.add(buttonPanel);
        panel.add(chooseLayoutBox);
        panel.setBorder(BorderFactory.createLoweredSoftBevelBorder());
        return panel;
    }


    public void addTools(Set<AbstractButton> buttons, String id) {

        addTools(buttons, id, id);
    }

    public void addTools(Set<AbstractButton> buttons, String id,
                         String description) {

        try {
            semaphore.acquire();
            DraggableComponentAdapter adapter;

            for (AbstractButton button : buttons) {
                adapter = new DraggableComponentAdapter(this, button);
                button.addMouseListener(adapter);
                button.addMouseMotionListener(adapter);

                // This retains information about where a button originally was
                // It is used to later create customizable layouts.
                groupingLookup.put(button.getName(), id);
                buttonLookup.put(button.getName(), button);
            }

            registeredButtons.put(id, buttons);
            tooltips.put(id, description);
            chooseToolsComboBox.addItem(id);

            // Check if new layouts are available
            Collection<StorableButtonContainerLayout> layouts = new HashSet<>(
                            unavailableLayouts.values());
            for (StorableButtonContainerLayout layout : layouts) {
                if (layout.isAvailable(registeredButtons.keySet())) {
                    chooseLayoutBox.addItem(layout);
                    unavailableLayouts.remove(layout.layoutName);
                }
            }

            createToolButtonContainer(buttons);
            semaphore.release();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }

    }

    private void reAddTools(String id) {
        Set<AbstractButton> buttons = registeredButtons.get(id);
        if (buttons != null) {
            createToolButtonContainer(buttons);
        }
    }


    public void removeTools(String id) {
        try {
            semaphore.acquire();

            Set<AbstractButton> buttons = registeredButtons.remove(id);
            chooseToolsComboBox.removeItem(id);
            tooltips.remove(id);

            if (buttons != null) {

                // Adding the buttons to a temporary panel removes them from all other components.
                JPanel tempPanel = new JPanel();
                for (AbstractButton button : buttons) {
                    tempPanel.add(button);
                    groupingLookup.remove(button.getName());
                    buttonLookup.remove(button.getName());
                }

                // Check if layouts are still available
                int count = chooseLayoutBox.getItemCount();
                for (int i = 0; i < count; i++) {
                    StorableButtonContainerLayout layout = chooseLayoutBox
                                    .getItemAt(i);
                    if (!layout.isAvailable(registeredButtons.keySet())) {
                        unavailableLayouts.put(layout.layoutName, layout);
                    }
                }
                for (StorableButtonContainerLayout value : unavailableLayouts
                                .values()) {
                    chooseLayoutBox.removeItem(value);
                }
                revalidate();
            }
            semaphore.release();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
    }

    private void createToolButtonContainer(Set<AbstractButton> buttons) {

        ToolButtonContainer container = new ToolButtonContainer(buttons);
        DraggableComponentAdapter adapter = new DraggableComponentAdapter(this,
                        container);
        container.addMouseListener(adapter);
        container.addMouseMotionListener(adapter);

        // Add one before last position
        int count = content.getComponentCount();
        content.add(container, count - 1);
        revalidate();
    }

    @Override
    public void dropComponent(Component draggedComponent,
                              Point locationOnScreen) {
        if (content.isVisible()) {
            content.dropComponent(draggedComponent, locationOnScreen);
        }
    }

    /**
     * Saves the current layout of the tool panel.
     */
    public void saveLayout(String layoutName) {

        final StorableButtonContainerLayout layout;

        try {
            semaphore.acquire();
            layout = StorableButtonContainerLayout.createLayout(
                            content.getComponents(), groupingLookup,
                            layoutName);

            // layout should be available if it has just been created
            chooseLayoutBox.addItem(layout);

            // remove old layout if it exists
            unavailableLayouts.remove(layoutName);
            semaphore.release();

            if (layout != null) {
                // We can run this concurrently, no reason to do this on the AWT thread
                new Thread(new Runnable() {
                    @Override
                    public void run() {
                        StorableButtonContainerLayout.writeToProperties(
                                        persistantLayouts, layout);
                    }
                }).start();
            }

        } catch (InterruptedException e) {
            e.printStackTrace();
        }

    }

    /**
     * Loads the requested layout.
     *
     * @param layout
     */
    private void loadLayout(StorableButtonContainerLayout layout) {

        try {
            semaphore.acquire();
            List<Set<AbstractButton>> buttons = StorableButtonContainerLayout
                            .readLayout(layout, buttonLookup);
            content.removeAllToolButtonContainers();
            for (Set<AbstractButton> set : buttons) {
                createToolButtonContainer(set);
            }
            semaphore.release();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }

    }

    private class CreateNewContainerTarget extends JPanel
                    implements DraggableComponentDropTarget {

        CreateNewContainerTarget() {

            super();

            final JLabel handle = new JLabel("<html>Mouse over</html>");
            handle.setHorizontalAlignment(JLabel.CENTER);
            this.addMouseListener(new MouseAdapter() {
                @Override
                public void mouseEntered(MouseEvent e) {
                    handle.setText("<html>Drop Button here to create new Container</html>");
                }

                @Override
                public void mouseExited(MouseEvent e) {
                    handle.setText("<html>Mouse over</html>");
                }
            });
            this.setLayout(new BorderLayout());
            this.add(handle, BorderLayout.CENTER);
            this.setBorder(BorderFactory.createLoweredSoftBevelBorder());
        }

        @Override
        public void dropComponent(Component draggedComponent,
                                  Point locationOnScreen) {
            if (draggedComponent instanceof AbstractButton) {
                Set<AbstractButton> set = new HashSet<AbstractButton>();
                set.add((AbstractButton) draggedComponent);
                createToolButtonContainer(set);
            }
        }

        @Override
        public void paintComponent(Graphics g) {
            super.paintComponent(g);

            Color background = DockUI.getColor(
                            LookAndFeelColors.TITLE_SELECTION_BACKGROUND);
            Color brighter = DockUI
                            .getColor(LookAndFeelColors.TITLE_BACKGROUND);

            Graphics2D g2 = (Graphics2D) g;
            GradientPaint backgroundPaint = new GradientPaint(0, 0, background,
                            getWidth() / 2, getHeight() / 2, brighter, true);
            g2.setPaint(backgroundPaint);
            g2.fill(new Rectangle(0, 0, getWidth(), getHeight()));
        }
    }

    /**
     * Tries to load an image from a pathname.
     * Used for later creating icon images.
     */
    private BufferedImage loadImage(final String path) {

        URL url = PluginManager.getInstance().getBottomClassLoader().getResource(path);
        if (url == null) {
            File file = new File(path);
            if (file != null) {
                try {
                    url = file.toURI().toURL();
                } catch (MalformedURLException e) {
                    e.printStackTrace();
                }
            }
        }
        if (url != null) {
            try {
                return ImageIO.read(url);
            } catch (IOException e) {
                e.printStackTrace();
            }
        }

        return null;
    }
}
