package CH.ifa.draw.standard;

import CH.ifa.draw.DrawPlugin;
import CH.ifa.draw.IOHelper;
import CH.ifa.draw.application.AbstractFileDragDropListener;
import CH.ifa.draw.application.DrawingViewContainer;
import CH.ifa.draw.application.DrawingViewFrame;
import CH.ifa.draw.contrib.DiamondFigure;
import CH.ifa.draw.contrib.PolygonTool;
import CH.ifa.draw.contrib.TriangleFigure;
import CH.ifa.draw.figures.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.gui.ToolButtonPanel;
import CH.ifa.draw.gui.ToolButtonScrollPane;
import CH.ifa.draw.io.DrawingFileHelper;
import CH.ifa.draw.io.ImportHolder;
import CH.ifa.draw.io.SimpleFileFilter;
import CH.ifa.draw.io.StatusDisplayer;
import CH.ifa.draw.io.importFormats.ImportFormat;
import CH.ifa.draw.util.*;

import bibliothek.gui.dock.common.CControl;
import bibliothek.gui.dock.common.DefaultSingleCDockable;
import bibliothek.gui.dock.common.event.CControlListener;
import bibliothek.gui.dock.common.event.CFocusListener;
import bibliothek.gui.dock.common.intern.CDockable;

import de.renew.windowmanagement.RenewShutdownHook;
import de.renew.windowmanagement.UnknownMenuSectionException;
import de.renew.windowmanagement.Workbench;
import de.renew.windowmanagement.WorkbenchImpl;

import java.awt.*;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetListener;
import java.awt.event.KeyEvent;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.util.*;
import java.util.List;

import javax.swing.*;


public class StandardDrawingEditor
                implements DrawingEditor, PaletteListener, CControlListener,
                CFocusListener, AutosaveSaver, StatusDisplayer {

    public static org.apache.log4j.Logger logger = org.apache.log4j.Logger
                    .getLogger(StandardDrawingEditor.class);

    // the image resource path
    private static final String fgDrawPath = "CH/ifa/draw/";

    /**
     * The image folder path
     */
    public static final String IMAGES = fgDrawPath + "images/";

    /**
     * The minimum time interval that has to elapse before a drawing is saved
     * automatically.
     */
    private static final int AUTOSAVE_INTERVAL = 120000;

    /**
     * The autosave manager that is associated to this DrawingEditor.
     */
    public AutosaveManager autosaveManager = new AutosaveManager(this,
                    AUTOSAVE_INTERVAL);

    /**
     * The id of the factory used to create {@link DrawingView}s.
     */
    public static final String FACTORY_ID = "StandardDrawingViewFactory";

    private final Map<String, String> factoryIDLookup = new HashMap();

    private StandardDrawingLookup fLookup;
    private ToolButton fCurrentToolButton;
    private ToolButton fDefaultToolButton;
    private UndoRedoManager fUndoRedoManager;
    protected Workbench fWorkbench;
    private StandardDrawingViewFactory fFactory;
    private List<SelectionChangeListener> fSelectionChangeListeners = new ArrayList<SelectionChangeListener>();
    private ToolButton fTextTB;
    private ToolButton fConnTextTB;
    private boolean fStickyTools = false;
    private ToolButtonScrollPane _toolbar;

    public StandardDrawingEditor(Workbench workbench) {

        fDefaultToolButton = createToolButton(IMAGES + "SEL", "Selection Tool",
                        createSelectionTool());
        fCurrentToolButton = fDefaultToolButton;
        fUndoRedoManager = new UndoRedoManager(this);
        fLookup = new StandardDrawingLookup();
        fWorkbench = workbench;

        fWorkbench.registerControlListener(this);
        fWorkbench.registerFocusListener(this);
        fFactory = new StandardDrawingViewFactory(this);
        fWorkbench.registerEditorFactory(fFactory, FACTORY_ID);

        _toolbar = new ToolButtonScrollPane(new ToolButtonPanel());
        _toolbar.setAutoscrolls(true);
        _toolbar.setHorizontalScrollBarPolicy(
                        JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        _toolbar.setVerticalScrollBarPolicy(
                        JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        fWorkbench.addViewWindow(new DefaultSingleCDockable("Tools", _toolbar),
                        WorkbenchImpl.UPPER_LEFT_GROUP);
        createToolPanel();
        fWorkbench.registerShutdownHook(new RenewShutdownHook() {
            @Override
            public boolean canClose() {
                _toolbar.saveLayout("Last Layout");
                _toolbar.storeLayoutsToDisk();
                return true;
            }

            @Override
            public void exit() {
            }
        });

        DropTargetListener dragDropListener = new AbstractFileDragDropListener() {
            @Override
            protected void handleFiles(final File[] files, Point loc) {
                for (File file : files) {
                    if (!file.isDirectory()) {
                        openOrLoadDrawing(file);
                    }
                }
            }
        };
        new DropTarget(fWorkbench.getMainFrame(), dragDropListener);

    }

    /**
     * Register a {@link SelectionChangeListener} to be notified when the selection has changed.
     *
     * @param listener
     */
    public void registerSelectionChangeListener(SelectionChangeListener listener) {
        fSelectionChangeListeners.add(listener);
    }

    /**
     * Creates the standard selection tool.
     *
     * @return
     */
    protected Tool createSelectionTool() {
        return new SelectionTool(this);
    }

    private void createToolPanel() {
        final LinkedHashSet<AbstractButton> set = new LinkedHashSet<>();
        ToolButton button;

        set.add(fDefaultToolButton.button());

        Tool tool = new CreationTool(this, new RectangleFigure());
        button = createToolButton(IMAGES + "RECT", "Rectangle Tool", tool);
        set.add(button.button());
        button.addAdditionalComponent(
                        ((CreationTool) tool).initialAttributesMenu());

        tool = new CreationTool(this, new RoundRectangleFigure());
        button = createToolButton(IMAGES + "RRECT", "Round Rectangle Tool",
                        tool);
        set.add(button.button());
        button.addAdditionalComponent(
                        ((CreationTool) tool).initialAttributesMenu());

        tool = new CreationTool(this, new EllipseFigure());
        button = createToolButton(IMAGES + "ELLIPSE", "Ellipse Tool", tool);
        set.add(button.button());
        button.addAdditionalComponent(
                        ((CreationTool) tool).initialAttributesMenu());

        tool = new CreationTool(this, new PieFigure());
        button = createToolButton(IMAGES + "PIE", "Elliptical Arc/Pie Tool",
                        tool);
        set.add(button.button());
        button.addAdditionalComponent(
                        ((CreationTool) tool).initialAttributesMenu());

        tool = new CreationTool(this, new DiamondFigure());
        button = createToolButton(IMAGES + "DIAMOND", "Diamond Tool", tool);
        set.add(button.button());
        button.addAdditionalComponent(
                        ((CreationTool) tool).initialAttributesMenu());

        tool = new CreationTool(this, new TriangleFigure());
        button = createToolButton(IMAGES + "TRIANGLE", "Triangle Tool", tool);
        set.add(button.button());
        button.addAdditionalComponent(
                        ((CreationTool) tool).initialAttributesMenu());

        tool = new CreationTool(this, new LineFigure());
        button = createToolButton(IMAGES + "LINE", "Line Tool", tool);
        set.add(button.button());
        button.addAdditionalComponent(
                        ((CreationTool) tool).initialAttributesMenu());

        tool = new ConnectionTool(this, new LineConnection());
        button = createToolButton(IMAGES + "CONN", "Connection Tool", tool);
        set.add(button.button());
        button.addAdditionalComponent(new JMenuItem("Test"));
        button.setAlternativeMode(ToolButton.POPUP_MENU);

        tool = new ConnectionTool(this, new ElbowConnection());
        button = createToolButton(IMAGES + "OCONN", "Elbow Connection Tool",
                        tool);
        set.add(button.button());

        tool = new ScribbleTool(this);
        button = createToolButton(IMAGES + "SCRIBBL", "Scribble Tool", tool);
        set.add(button.button());

        tool = new PolygonTool(this);
        button = createToolButton(IMAGES + "POLYGON", "Polygon Tool", tool);
        set.add(button.button());

        tool = new ImageFigureCreationTool(this, fWorkbench.getMainFrame());
        button = createToolButton(IMAGES + "IMAGE", "Image Tool", tool);
        set.add(button.button());

        TextFigure prototype = new TextFigure(false);
        prototype.setAlignment(TextFigure.LEFT);
        tool = new TextTool(this, prototype);
        fTextTB = createToolButton(IMAGES + "TEXT", "Text Tool", tool);
        set.add(fTextTB.button());

        prototype = new TextFigure();
        prototype.setAlignment(TextFigure.CENTER);
        tool = new ConnectedTextTool(this, prototype);
        fConnTextTB = createToolButton(IMAGES + "ATEXT", "Connected Text Tool",
                        tool);
        set.add(fConnTextTB.button());

        tool = new TargetTool(this);
        button = createToolButton(IMAGES + "TARGETTEXT", "Target Tool", tool);
        set.add(button.button());

        String description = "<html><b>Standard Tools</b><br>"
                        + "The standard drawing tools<br>"
                        + "used to create basic shapes.</html>";
        _toolbar.addTools(set, "Standard Tools", description);
    }

    /**
     * Returns the current toolbar.
     *
     * @return
     */
    public Toolbar getToolbar() {
        return _toolbar;
    }

    /**
     * Creates a tool button with the given image, tool, and text
     *
     * @param iconName the icon name for the tool button
     * @param toolName the name of the tool button
     * @param tool     the tool associated with the tool button
     * @return the created tool button
     */
    public ToolButton createToolButton(String iconName, String toolName,
                                       Tool tool) {
        return new ToolButton(this, iconName, toolName, tool);
    }

    /**
     * create a button for with the given icons, name and tool
     * @param icon the icon for the button
     * @param selectedIcon the icon for the pressed button
     * @param toolName the name of the tool
     * @param tool the tool to be triggered with this button
     */
    public ToolButton createToolButton(Icon icon, Icon selectedIcon,
                                       String toolName, Tool tool) {
        // logger.debug("creating a toolbutton with icon " + iconName + ", toolname " + toolName);
        logger.debug("the delegator is " + this);
        return new ToolButton(this, icon, selectedIcon, toolName,
                tool);
    }

    /**
     * Set the {@link StandardDrawingViewFactory} which is responsible for creating
     * new view containers for drawings.
     *
     * @param factory
     */
    public void setDrawingViewFactory(StandardDrawingViewFactory factory,
                                      String factoryID) {

        fFactory = factory;
        // We have to keep the old factory registered, in case we have already used it
        // to create drawings.
        fWorkbench.registerEditorFactory(factory, factoryID);
    }

    @Override
    public DrawingView view() {
        DrawingView view = fLookup.getCurrentView();
        if (view != null) {
            return view;
        } else {
            return NullDrawingView.INSTANCE;
        }
    }

    @Override
    public DrawingView previousView() {
        return fLookup.getPreviousView();
    }

    @Override
    public Drawing drawing() {
        return view().drawing();
    }

    @Override
    public Tool tool() {
        return fCurrentToolButton.tool();
    }

    @Override
    public Tool defaultTool() {
        return fDefaultToolButton.tool();
    }

    /**
     * Set if tools are sticky.
     * Usually we always want sticky tools, but some tools, like the image creation tool, require non-sticky tools to be set.
     *
     * @param sticky
     */
    @Override
    public void setStickyTools(boolean sticky) {
        fStickyTools = sticky;
    }

    /**
     * The concept of sticky tools was not really intuitive, so we just return true
     * here and use sticky tools all the time.
     *
     * @return true
     */
    @Override
    public boolean isStickyTools() {
        return fStickyTools;
    }

    @Override
    public Image getIconImage() {

        Image icon = fWorkbench.getMainFrame().getIconImage();

        assert icon != null : "Icon of mainframe may not be null.";

        return icon;
    }

    @Override
    public void dispatchEvent(KeyEvent evt) {
        fWorkbench.getMainFrame().dispatchEvent(evt);

    }

    /**
     * Gets tool button dependent of the type of the given figure.
     *
     * @param figure the figure for which the tool button is searched.
     * @return the tool button for the figure.
     */
    public ToolButton toolButtonForTextFigure(TextFigure figure) {
        if (figure.parent() == null) {
            return fTextTB;
        } else {
            return fConnTextTB;
        }
    }

    @Override
    public void toolDone() {

        if (isStickyTools() && tool() != null) {
            tool().deactivate();
            tool().activate();
        } else if (fDefaultToolButton != null) {
            setToolButton(fDefaultToolButton);
        }


    }

    @Override
    public void selectionChanged(DrawingView view) {
        menuStateChanged();

        for (SelectionChangeListener listener : fSelectionChangeListeners) {
            listener.selectionStateChanged();
        }
    }

    /**
     * Call if some action has occurred after which a
     * {@link CH.ifa.draw.util.CommandMenuItem} might be disabled.
     */
    @Override
    public void menuStateChanged() {

        JFrame frame = fWorkbench.getMainFrame();

        if (frame != null) {
            JMenuBar mb = frame.getJMenuBar();
            int count = mb.getMenuCount();

            for (int i = 0; i < count; i++) {
                JMenu menu = (mb.getMenu(i));

                if (menu instanceof CommandMenu) {
                    ((CommandMenu) menu).checkEnabled();
                    ((CommandMenu) menu).updateCommandText();
                }
            }
        }

    }

    @Override
    public void showStatus(String string) {
        fWorkbench.showStatus(string);
    }

    @Override
    public void prepareUndoSnapshot() {
        Drawing currentDrawing = drawing();
        fUndoRedoManager.prepareUndoSnapshot(currentDrawing);
    }

    @Override
    public void commitUndoSnapshot() {
        Drawing currentDrawing = drawing();
        fUndoRedoManager.commitUndoSnapshot(currentDrawing);
    }

    @Override
    public void prepareAccumulatedUndoSnapshot() {
        Drawing currentDrawing = drawing();
        fUndoRedoManager.prepareAccumulatedUndoSnapshot(currentDrawing);
    }

    @Override
    public void triggerAccumulatedUndoSnapshot() {
        Drawing currentDrawing = drawing();
        fUndoRedoManager.triggerAccumulatedUndoSnapshot(currentDrawing);
    }

    @Override
    public UndoRedoManager getUndoRedoManager() {
        return fUndoRedoManager;
    }

    /**
     * This is not used anymore. Instead use {@link StandardDrawingEditor#focusGained}.
     *
     * @param viewContainer
     */
    @Override
    public void drawingViewContainerActivated(DrawingViewContainer viewContainer) {
        // DrawingViewContainers are not used anymore.
    }

    /**
     * This is not used anymore. Instead use {@link StandardDrawingEditor#closed} or
     * {@link StandardDrawingEditor#removed} if removeOnClose is not set to true.
     *
     * @param viewContainer
     */
    @Override
    public void drawingViewContainerClosing(DrawingViewContainer viewContainer) {
        // DrawingViewContainers are not used anymore.
    }

    /**
     * This is not used anymore. Instead use
     * {@link StandardDrawingLookup#setCurrentView}.
     *
     * @param dvc
     */
    @Override
    public void setCurrentDrawing(DrawingViewContainer dvc) {
        // DrawingViewContainers are not used anymore.
    }

    /**
     * Don't use. If you need the frame size for some reason, just call the
     * workbench directly.
     * <p>
     * Currently only used for calculating the position of new
     * {@link DrawingViewFrame}s. With the new window sizes, this will not work
     * anymore.
     *
     * @return
     */
    @Deprecated
    @Override
    public Dimension getSize() {
        return fWorkbench.getMainFrame().getSize();
    }

    /**
     * Don't use. If you need the frame location for some reason, just call the
     * workbench directly.
     * <p>
     * Currently only used for calculating the position of new
     * {@link DrawingViewFrame}s. With the new window sizes, this will not work
     * anymore.
     *
     * @return
     */
    @Deprecated
    @Override
    public Point getLocationOnScreen() {
        return fWorkbench.getMainFrame().getLocationOnScreen();
    }

    /**
     * Reacts to user selections on the palette panel. Sticky tools are not used
     * anymore, so doubleclick does nothing different.
     *
     * @param button
     * @param doubleclick
     */
    @Override
    public void paletteUserSelected(ToolButton button, boolean doubleclick) {
        setToolButton(button);

        fStickyTools = doubleclick;
        showStatus("Selected tool: " + button.name());
    }

    /**
     * Deactivate the old tool button and activate the new one.
     *
     * @param button
     */
    private void setToolButton(ToolButton button) {
        assert button != null;

        //fCurrentToolButton.button().setSelected(false);
        ToolButton.ToolButtonModel model = (ToolButton.ToolButtonModel) fCurrentToolButton
                        .button().getModel();
        model.setTrulySelected(false);
        fCurrentToolButton.tool().deactivate();
        fCurrentToolButton = button;

        model = (ToolButton.ToolButtonModel) button.button().getModel();
        model.setTrulySelected(true);
        button.tool().activate();
    }

    @Override
    public void paletteUserOver(ToolButton button, boolean inside) {
        // We could show the name of the hovered tool on the status line.
        // But that is kind of pointless, since the buttons already have tooltips.
        // Also it would erase any other more important information displayed there.
    }

    @Override
    public void added(CControl cControl, CDockable cDockable) {
        if (cDockable instanceof StandardDrawingViewContainer) {

            logger.debug("Dockable of type StandardDrawingViewContainer was added.");

            StandardDrawingViewContainer container = (StandardDrawingViewContainer) cDockable;
            Drawing drawing = container.getView().drawing();
            fLookup.addViewContainer(drawing, container);
            fUndoRedoManager.newUndoHistory(drawing);
        }
    }

    @Override
    public void removed(CControl cControl, CDockable cDockable) {
        if (cDockable instanceof StandardDrawingViewContainer) {
            logger.debug("Dockable of type StandardDrawingViewContainer was removed.");

            StandardDrawingViewContainer container = (StandardDrawingViewContainer) cDockable;
            Drawing drawing = container.getView().drawing();
            fLookup.removeViewContainer(drawing);
            noUndoHistoryFor(drawing);
            if (fLookup.getCurrentView() == container.getView()) {
            	fLookup.setCurrentView(NullDrawingView.INSTANCE);
            	selectionChanged(null);
            }
        }
    }

    @Override
    public void opened(CControl cControl, CDockable cDockable) {
        // Use added instead
    }

    @Override
    public void closed(CControl cControl, CDockable cDockable) {
        // Use removed instead
    }

    @Override
    public void focusGained(CDockable cDockable) {
        if (cDockable instanceof StandardDrawingViewContainer) {
            logger.debug(((StandardDrawingViewContainer) cDockable).getView()
                            .drawing().getName() + " gained focus.");
            StandardDrawingViewContainer container = (StandardDrawingViewContainer) cDockable;
            fLookup.setCurrentView(container.getView());
            selectionChanged(container.getView());
        }
    }

    @Override
    public void focusLost(CDockable cDockable) {
        // We want to retain the current view, if a tool panel has been selected.
        // So unless another View has gained focus, do nothing
    }

    /**
     * Open the given {@link Drawing} by creating a new
     * {@link StandardDrawingViewContainer}. If {@link Drawing} is already open,
     * instead focus the view container.
     *
     * @param drawing
     * @return the same drawing given
     */
    public Drawing openDrawing(Drawing drawing) {

        assert drawing != null : "Drawing may not be null.";

        StandardDrawingViewContainer container = fLookup
                        .getViewContainerForDrawing(drawing);
        if (container != null) {

            if (fLookup.getCurrentView().drawing().equals(drawing)) {
                // case: view already focused, do nothing
            } else {
                // case: drawing is already open, but not focused
                container.toFront();
            }
        } else {
            // case: drawing has not been opened yet
            createNewViewContainerForDrawing(drawing);

        }

        return drawing;
    }

    /**
     * Calls the {@link Workbench} to create a {@link StandardDrawingViewContainer} via a previously registered factory.
     * Once the drawing is opened, this class will be informed via {@link #added(CControl, CDockable)}.
     *
     * @param drawing The drawing to be opened
     */
    protected void createNewViewContainerForDrawing(Drawing drawing) {

        String id = factoryIDLookup.getOrDefault(drawing.getClass().getName(),
                        FACTORY_ID);
        fWorkbench.addEditorWindow(new DrawingLayoutWrapper(drawing), id);
    }

    /**
     * Opens the drawing from the given file. If it is not in the drawing list it
     * will be loaded.
     *
     * @param file the file of the drawing.
     */
    public void openOrLoadDrawing(File file) {

        Drawing drawing = fLookup.getLoadedDrawingForFile(file);
        if (drawing != null) {
            openDrawing(drawing);
        } else {
            getIOHelper().loadAndOpenDrawings(file);
        }
    }

    /**
     * Closes the specified {@link Drawing}.
     *
     * @param drawing
     */
    public void closeDrawing(Drawing drawing) {
        StandardDrawingViewContainer container = fLookup
                        .getViewContainerForDrawing(drawing);
        fLookup.removeViewContainer(drawing);
        fWorkbench.close(container);
    }

    /**
     * Creates a new empty {@link Drawing}. The specific subtype of the
     * {@link Drawing} is defined by the registered factory.
     *
     * @return
     */
    public Drawing newDrawing() {
        Drawing drawing = fFactory.create().drawing();
        return openDrawing(drawing);
    }

    /**
     * Saves the given drawing to the given autosave file. The drawing's dirty flag
     * is not cleared.
     *
     * @param drawing the drawing to save.
     * @param file    the name of the autosave file.
     * @param loc     the location of the drawing's view window; it will be saved to the
     *                file, too.
     * @param size    width and height of the drawing's view window; they will be saved
     *                to the file, too.
     * @throws IOException if an I/O-failure occurs.
     **/
    public void saveAutosaveFile(Drawing drawing, File file, Point loc,
                                 Dimension size)
                    throws IOException {
        DrawingFileHelper.saveAsStorableOutput(drawing, file, loc, size, false);
    }

    /**
     * Prohibits the management of undo and redo snapshots for the given drawing.
     **/
    protected void noUndoHistoryFor(Drawing drawing) {
        fUndoRedoManager.removeUndoHistory(drawing);
    }

    public void addMenu(JMenu menu) {
        fWorkbench.registerMenu(menu);
    }

    public void addMenu(JMenu menu, int index) {

        try {
            fWorkbench.registerMenu(menu, index);
        } catch (UnknownMenuSectionException e) {
            e.printStackTrace();
        }
    }

    public void removeMenu(JMenu menu) {
        fWorkbench.removeMenu(menu);
    }

    public DrawingView getView(Drawing drawing) {
        StandardDrawingViewContainer container = fLookup
                        .getViewContainerForDrawing(drawing);
        DrawingView view = container.getView();

        assert view != null;

        return view;
    }

    /**
     * Get the lookup managing all currently known drawings and their respective
     * visual representations.
     *
     * @return
     */
    public StandardDrawingLookup getDrawingLookup() {
        return fLookup;
    }

    private IOHelper getIOHelper() {
        return DrawPlugin.getCurrent().getIOHelper();
    }

    private ImportHolder getImportHolder() {
        return DrawPlugin.getCurrent().getImportHolder();
    }

    private boolean canOpen(URI path) {
        Iterator<SimpleFileFilter> iterator = getIOHelper().getFileFilter()
                        .getFileFilters().iterator();
        while (iterator.hasNext()) {
            SimpleFileFilter filter = (SimpleFileFilter) iterator.next();
            if (filter.accept(new File(path.getPath()))) {
                return true;
            }
        }
        ImportFormat[] allImportFormats = getImportHolder().allImportFormats();
        for (ImportFormat importFormat : allImportFormats) {
            if (importFormat.canImport(path)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Opens the text edit for the given figure.
     *
     * @param textFigure the figure which should be edited.
     */
    public void doTextEdit(TextFigure textFigure) {
        doTextEdit(textFigure, 0, 0, 0, 0);
    }

    /**
     * Edit a given {@link TextFigure} with the appropriate {@link TextTool}.
     *
     * @param textFigure  The figure to edit
     * @param startLine   if {@code startLine > 0 && startColumn > 0}, set caret position
     * @param startColumn
     * @param endLine     if {@code endLine > startLine or (startLine == endLine) && (startColumn <
     *                    endColumn)}, select text in between
     * @param endColumn
     */
    public void doTextEdit(TextFigure textFigure, int startLine,
                           int startColumn, int endLine, int endColumn) {

        ToolButton tb = toolButtonForTextFigure(textFigure);
        paletteUserSelected(tb, false);
        TextTool tt = (TextTool) tb.tool();
        tt.beginEdit(textFigure);

        if ((startLine < endLine) || ((startLine == endLine)
                        && (startColumn < endColumn))) {
            tt.select(startLine, startColumn, endLine, endColumn);
        } else {
            if (startLine > 0 && startColumn > 0) {
                tt.setCaretPosition(startLine, startColumn);
            }
        }

    }

    /**
     * Loads a drawing from the given file and inserts all its figures into
     * the current drawing.
     *
     * @param file the file name where to retrieve the drawing.
     **/
    public synchronized void loadAndInsertDrawing(File file) {

        Drawing existingDrawing = drawing();
        DrawingView existingView = view();
        Drawing newDrawing = DrawingFileHelper.loadDrawing(file, null);
        if (newDrawing != null) {
            FigureEnumeration figures = newDrawing.figures();
            existingView.clearSelection();
            while (figures.hasMoreElements()) {
                Figure fig = figures.nextFigure();
                existingDrawing.add(fig);
                existingView.addToSelection(fig);
            }
            existingView.checkDamage();
        }

    }

    /**
     * Prints the drawing.
     */
    public void print() {

        fCurrentToolButton.tool().deactivate();
        PrinterJob printerJob = PrinterJob.getPrinterJob();
        printerJob.setPrintable(view());
        if (printerJob.printDialog()) {
            try {
                printerJob.print();
            } catch (PrinterException e) {
                logger.debug(e.getMessage(), e);
                JOptionPane.showMessageDialog(fWorkbench.getMainFrame(),
                                "Printing of this Drawing failed.\n"
                                                + "Reason: " + e.getMessage(),
                                "Printer Error", JOptionPane.ERROR_MESSAGE);
            }
        }
        fCurrentToolButton.tool().activate();
    }

    /**
     * Close the currently focused drawing.
     */
    public void closeCurrentDrawing() {
        Drawing drawing = drawing();
        StandardDrawingViewContainer container = fLookup
                        .getViewContainerForDrawing(drawing);
        fWorkbench.close(container);
    }

    @Override
    public void registerDrawingViewFactoryID(String key, String id) {
        factoryIDLookup.put(key, id);
    }
}
