package de.renew.extras.commands;

import CH.ifa.draw.DrawPlugin;
import CH.ifa.draw.application.DrawApplication;
import CH.ifa.draw.figures.AttributeFigure;
import CH.ifa.draw.framework.Drawing;
import CH.ifa.draw.framework.Figure;
import CH.ifa.draw.framework.FigureWithID;

import de.renew.extras.commands.StyleStore.StyleInformation;
import de.renew.plugin.command.CLCommand;

import java.awt.Dimension;
import java.awt.Rectangle;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;


public class ApplyCopiedStyleCommand implements CLCommand {
    public final static String CMD = "applystyle";
    private StyleStore store;
    private static org.apache.log4j.Logger logger = org.apache.log4j.Logger
                    .getLogger(ApplyCopiedStyleCommand.class);

    public ApplyCopiedStyleCommand(StyleStore store) {
        this.store = store;
    }

    @Override
    public void execute(final String[] args, final PrintStream response) {
        CommandLineParser parser = new DefaultParser();
        Options opts = new Options();
        opts.addOption("a", "no-attributes", false,
                        "do not apply style attributes from style store");
        opts.addOption("d", "no-dimensions", false,
                        "do not apply dimensions from style store");
        opts.addOption("r", "no-reset", false,
                        "do not reset attributes that are unset in style store");
        opts.addOption("l", "list", false, "list available style identifiers");
        opts.addOption("h", "help", false, "print help");


        CommandLine line = null;
        try {
            line = parser.parse(opts, args);
        } catch (ParseException e1) {
            response.append("Could not parse command.\n");
            printHelp(response, opts);
            return;
        }

        if (line.hasOption("l")) {
            response.append("Available identifiers: \n");
            for (String identifier : this.store.getKnownIdentifiers()) {
                response.append(identifier + "\n");
            }
            return;
        }

        if (line.hasOption("h")) {
            printHelp(response, opts);
            return;
        }

        List<String> argList = line.getArgList();

        DrawApplication app = DrawPlugin.getGui();
        Vector<Figure> selection = app.view().selection();
        if (selection.isEmpty()) {
            response.append("Command " + CMD
                            + ": You need to select at least one Figure\n");
            printHelp(response, opts);
        } else if (argList.size() > 1) {
            response.append("Command " + CMD
                            + ": You can use at most one style identifier\n");
            printHelp(response, opts);
        } else {
            DrawApplication application = DrawPlugin.getGui();
            Drawing drawing = application.drawing();
            application.getUndoRedoManager().prepareUndoSnapshot(drawing);

            StyleInformation style;

            if (argList.isEmpty()) {
                style = this.store.getStyleInformation();
            } else {
                style = this.store.getStyleInformation(argList.get(0));
            }

            if (style == null) {
                response.append("Command " + CMD
                                + ": no style stored under the selected identifier\n");
                return;
            }

            for (Figure f : selection) {
                String figureID = (f instanceof FigureWithID)
                                ? " " + ((FigureWithID) f).getID()
                                : "";
                boolean changed = false;

                // attributes
                if (!line.hasOption("a") && style.getAttributes() != null) {
                    if (f instanceof AttributeFigure) {
                        AttributeFigure afig = (AttributeFigure) f;
                        if (!line.hasOption("r")) {
                            for (String attribute : StyleStore.STYLE_ATTRIBUTES) {
                                changed |= setAttribute(afig, attribute, style);
                            }
                        }
                        response.append("Applied style attributes from style store to Figure"
                                        + figureID + "\n");
                    } else {
                        response.append("Did not apply style attributes from style store to Figure"
                                        + figureID
                                        + ", because the Figure is not an AttributeFigure\n");
                    }
                }

                // dimensions
                if (!line.hasOption("d") && style.getDimensions() != null) {
                    changed |= setDimensions(f, style);
                    response.append("Applied dimensions from style store to Figure"
                                    + figureID + "\n");
                }
                if (changed) {
                    app.view().checkDamage();
                    application.getUndoRedoManager()
                                    .commitUndoSnapshot(drawing);
                }
            }
        }
    }

    private boolean setDimensions(Figure f, StyleInformation style) {
        Dimension dimensions = style.getDimensions();
        Rectangle r = f.displayBox();
        Rectangle newDisplayBox = new Rectangle(r);
        newDisplayBox.width = (int) dimensions.getWidth();
        newDisplayBox.height = (int) dimensions.getHeight();
        f.displayBox(newDisplayBox); // this might have no effect (e.g. for NetComponentFigures)
        boolean displayBoxChanged = (!r.equals(f.displayBox()));
        return displayBoxChanged;
    }

    private boolean setAttribute(AttributeFigure afig, String attribute,
                                 StyleInformation style) {
        boolean changed = false;
        Map<String, Object> attributes = style.getAttributes();
        if (attributes.containsKey(attribute)) {
            Object value = attributes.get(attribute);
            logger.debug("Setting attribute " + attribute + " to " + value);
            afig.setAttribute(attribute, value);
            changed = true;
        } else if (afig.hasAttribute(attribute)) {
            Object defaultValue = AttributeFigure
                            .getDefaultAttribute(attribute);
            logger.debug("Setting attribute " + attribute + " to "
                            + defaultValue);
            afig.setAttribute(attribute, defaultValue);
            changed = true;
        }
        return changed;
    }

    @Override
    public String getDescription() {
        return "Apply the copied style to the selected figures";
    }

    @Override
    public String getArguments() {
        return null;
    }

    private void printHelp(PrintStream response, Options opts) {
        HelpFormatter formatter = new HelpFormatter();
        String header = getDescription();
        formatter.printHelp(new PrintWriter(response, true),
                        HelpFormatter.DEFAULT_WIDTH, CMD + " [identifier]",
                        header, opts, HelpFormatter.DEFAULT_LEFT_PAD,
                        HelpFormatter.DEFAULT_DESC_PAD, null, true);
    }
}