package de.renew.extras.commands;

import de.renew.plugin.IPlugin;
import de.renew.plugin.PluginManager;
import de.renew.plugin.PluginProperties;
import de.renew.plugin.command.CLCommand;
import de.renew.plugin.command.CLCommandHelper;
import de.renew.plugin.command.LoadCommand;
import de.renew.plugin.command.UnloadCommand;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.List;


/**
 * This command reloads the plugin with the given name.
 * If the argument "-v" is given, verbose output is printed,
 * especially concerning load errors regarding dependency problems.
 *
 * @author Eva M&uuml;ller
 */
public class ReloadCommand implements CLCommand {
    public static org.apache.log4j.Logger logger = org.apache.log4j.Logger
                    .getLogger(ReloadCommand.class);
    private final int PLUGIN_NAME_PROP_INDEX = 0;
    private final int PLUGIN_PROJECT_FOLDER_PROP_INDEX = 1;
    private final int PLUGIN_FOLDER_NAME_PROP_INDEX = 2;
    private static final String PROP_PLUGIN_MATCHES = "de.renew.reloadcommand.pluginMatches";
    private static final String PROP_WORKSPACES = "de.renew.reloadcommand.workspaces";
    private static final String PROP_BUILD_XML_FOLDER = "de.renew.reloadcommand.buildxml.folder";
    private static final String PROP_BUILD_XML_TARGET = "de.renew.reloadcommand.buildxml.target";
    private static final String DS = File.separator;
    private static final String FLAG_I = "-i";
    private static final String FLAG_V = "-v";

    @Override
    public void execute(String[] args, PrintStream response) {
        response.println("\n\n------- RELOAD -------\n\n");
        if (args == null) {
            response.println("ReloadCommand: please set name of plugin!");
            return;
        }
        if (args.length == 1 && ("h".equals(args[0]) || "-h".equals(args[0])
                        || "--help".equals(args[0]) || "--h".equals(args[0])
                        || "-help".equals(args[0]))) {
            response.println("usage: reload [" + FLAG_I + " | " + FLAG_V
                            + "] <PluginName>\n" + "Description: \n"
                            + "Reloads given plugin. I.e., the plugin (and dependent plugin) will be unloaded, \n"
                            + "the stated plugin will be build (if build.xml is available), \n"
                            + "and all unloaded plugins will be loaded again.\n\n"
                            + "                  This command is highly EXPERIMENTAL!!!\n\n");
            response.println(FLAG_I + ": Start reload process immediately");
            response.println(FLAG_V + ": Be verbose");
            response.println("");
            response.println(
                            "To use this command, extend your local .renew.properties file by the following properties:");
            response.println("# Semicolon separated pathes to workspaces");
            response.println(
                            "de.renew.reloadcommand.workspaces=/path/to/Mulan;/path/to/Renew");
            response.println("");
            response.println(
                            "# Semicolon separated entries stated as => <Plugin name>:<Project folder name>:<Plugin folder name>");
            response.println(
                            "de.renew.reloadcommand.pluginMatches=Navigator:Renew:Navigator;Mulan Viewer Gui:Mulan:MulanViewerGui");
            response.println("");
            response.println("---------------------");
            response.println(
                            "To trigger build process from within another project, use the following properties (e.g. from RenewMulanBinary)");
            response.println("# Use build.xml from Project RenewMulanBinary");
            response.println(
                            "de.renew.reloadcommand.buildxml.folder=/path/to/RenewMulanBinary");
            response.println(
                            "de.renew.reloadcommand.buildxml.target=buildIndivdualPlugin");
            return;
        }

        String pluginName = CLCommandHelper.getPluginName(args);
        if (pluginName == null || pluginName.trim().length() == 0) {
            response.println("ReloadCommand: please set name of plugin!");
            return;
        }

        IPlugin pluginToReload = PluginManager.getInstance()
                        .getPluginByName(pluginName);
        if (pluginToReload == null) {
            LoadCommand loadCommand = new LoadCommand();
            loadCommand.execute(new String[] { pluginName.toLowerCase() + "*" },
                            response);
            return;
        }
        File buildxml = null;
        try {
            buildxml = getPluginBuildXml(pluginName);
        } catch (NullPointerException e) {
            logger.error(e.getMessage());
        }
        try {
            boolean startReload = CLCommandHelper.isFlagSet(args, FLAG_I) ? true
                            : confirm(pluginName, response);
            if (!startReload) {
                return;
            }
            List<IPlugin> dependers = CLCommandHelper
                            .getDependers(pluginToReload, false);

            // Unload first, else build will not replace resulting jar file
            UnloadCommand unloadCommand = new UnloadCommand();
            unloadCommand.execute(new String[] { "-r", pluginName }, response);

            // Ensure that plugin is unloaded
            boolean doReload = waitUnloaded(pluginName, response);

            // trigger build
            if (buildxml != null) {
                if (pluginToReload != null && !executeBuildFile(pluginName,
                                buildxml, response,
                                CLCommandHelper.isFlagSet(args, FLAG_V))) {
                    doReload = false;
                }
            } else {
                response.println("---> No build will be triggered.");
            }

            // trigger reload
            if (doReload) {
                reload(dependers, pluginToReload, response);
            }
        } catch (ArrayIndexOutOfBoundsException e) {
            logger.error("Reload failed.", e);
        } catch (NumberFormatException e) {
            logger.error("Reload failed.", e);
        } catch (URISyntaxException e) {
            logger.error("Reload failed.", e);
        }
    }

    private void reload(List<IPlugin> dependers, IPlugin toShutDown,
                        PrintStream response)
                    throws URISyntaxException {
        LoadCommand loadCommand = new LoadCommand();
        loadCommand.execute(new String[] { extractName(toShutDown) }, response);
        if (waitLoaded(toShutDown.getName(), response)) {
            for (IPlugin entry : dependers) {
                logger.debug("Dependent plugin: " + entry.getName() + " -> "
                                + entry.getProperties().getURL().getPath());
                if (entry.getName().equals(toShutDown.getName())) {
                    continue;
                }
                loadCommand.execute(new String[] { extractName(entry) },
                                response);
                if (!waitLoaded(entry.getName(), response)) {
                    logger.error("-----> Dependent plugin " + entry.getName()
                                    + " could not be loaded.");
                }
            }
        } else {
            logger.error("-----> Plugin " + toShutDown.getName()
                            + " could not be loaded.");
        }
    }

    private boolean waitUnloaded(String pluginName, PrintStream response) {
        int max = 0;
        while (max < 10 && PluginManager.getInstance()
                        .getPluginByName(pluginName) != null) {
            max++;
            try {
                response.println("[" + max + "] .... wait ....");
                Thread.sleep(1000);
            } catch (Exception e) {
                // Nothing to do
            }
        }
        if (PluginManager.getInstance().getPluginByName(pluginName) != null) {
            logger.error("-----> Plugin " + pluginName + " still loaded.");
            return false;
        }
        response.println("-----> Plugin " + pluginName + " unloaded.");
        return true;
    }

    private boolean waitLoaded(String pluginName, PrintStream response) {
        int max = 0;
        while (max < 10 && PluginManager.getInstance()
                        .getPluginByName(pluginName) == null) {
            max++;
            try {
                response.println("[" + max + "] .... wait ....");
                Thread.sleep(1000);
            } catch (Exception e) {
                // Nothing to do
            }
        }
        if (PluginManager.getInstance().getPluginByName(pluginName) == null) {
            logger.error("-----> Plugin " + pluginName
                            + " could not be loaded.");
            return false;
        }
        response.println("-----> Plugin " + pluginName + " loaded.");
        return true;
    }

    private String extractName(IPlugin plugin) throws URISyntaxException {
        URL url = plugin.getProperties().getURL();
        File tmp = new File(url.toURI());
        return tmp.getName();
    }

    private boolean executeBuildFile(String pluginName, File buildXml,
                                     PrintStream response, boolean verbose) {
        String target = "";
        String properties = "";
        String tmpTarget = getProperty(PROP_BUILD_XML_TARGET);
        if (tmpTarget != null) {
            target = " " + tmpTarget;
            String projectFolder = getPluginProjectLocation(pluginName);
            String pluginFolderName = getPluginFolder(pluginName);
            if (projectFolder == null || pluginFolderName == null) {
                response.println("Build could not be executet. Project folder ("
                                + projectFolder
                                + ") and/or plugin folder name ("
                                + pluginFolderName + ") are missing.");
                return false;
            }
            properties = " -Dtmp.checkout=" + projectFolder
                            + " -DpluginFolderName=" + pluginFolderName;
        }
        StringBuffer sb = new StringBuffer();
        try {
            Process p = Runtime.getRuntime()
                            .exec("ant" + properties + " -buildfile "
                                            + buildXml.getAbsolutePath()
                                            + target);
            p.waitFor();
            BufferedReader reader = new BufferedReader(
                            new InputStreamReader(p.getInputStream()));
            String line = reader.readLine();
            sb.append(line + "\n");
            while (line != null) {
                line = reader.readLine();
                if (line != null) {
                    sb.append(line + "\n");
                    if (line.trim().toUpperCase().equals("BUILD SUCCESSFUL")) {
                        if (verbose) {
                            response.println(sb.toString());
                        } else {
                            response.println("Build successful.");
                        }
                        return true;
                    }
                }
            }
            logger.warn(sb.toString());
        } catch (IOException e) {
            response.println("Build failed.\n" + sb.toString());
            logger.error("Build failed", e);
        } catch (InterruptedException e) {
            response.println("Build failed.\n" + sb.toString());
            logger.error("Build failed", e);
        }
        return false;
    }

    private boolean confirm(String pluginName, PrintStream printStream) {
        InputStream in1 = System.in;
        InputStreamReader in2 = new InputStreamReader(in1);
        BufferedReader in = new BufferedReader(in2);
        PrintStream out = printStream;
        out.println("Do you really want to remove '" + pluginName
                        + "' and all of its dependent plugins?");
        out.println("Type [y|yes] to start replacement process");
        out.println("");
        String answer = null;
        try {
            answer = in.readLine();
            if (answer != null && (answer.trim().toLowerCase().equals("yes")
                            || answer.trim().toLowerCase().equals("y"))) {
                out.println("");
                out.println("Start reloading plugin: '" + pluginName + "'");
                out.println("");
                return true;
            }
        } catch (IOException e) {
            out.println("Could not reload '" + pluginName
                            + "' due to exception : " + e + ".");
            return false;
        }
        out.println("");
        out.println("Cancelled reload process...");
        out.println("");
        return false;
    }

    private File getPluginBuildXml(String toShutDown) {
        File buildxml = new File(
                        getProperty(PROP_BUILD_XML_FOLDER) + DS + "build.xml");
        if (buildxml.exists()) {
            return buildxml;
        }
        String pluginFolderName = getPluginFolder(toShutDown);
        if (pluginFolderName == null) {
            throw new NullPointerException(
                            "Could not determine the plugin folder name");
        }
        String projectLocation = getPluginProjectLocation(toShutDown);
        if (projectLocation != null) {
            buildxml = new File(projectLocation + DS + pluginFolderName + DS
                            + "build.xml");
            if (!buildxml.exists()) {
                throw new NullPointerException(
                                "Could not determine the plugin's build.xml. The following path does not exist: "
                                                + buildxml.getAbsolutePath());
            }
            return buildxml;
        }
        return null;
    }

    private String getPluginProjectLocation(String toShutDown) {
        String projectFolderName = getPluginProjectFolder(toShutDown);
        if (projectFolderName == null) {
            throw new NullPointerException(
                            "Could not determine the plugin's project folder name");
        }
        String workspacesProp = getProperty(PROP_WORKSPACES);
        if (workspacesProp != null && workspacesProp.trim().length() != 0) {
            String[] parts = workspacesProp.split(";");
            if (parts != null) {
                for (String part : parts) {
                    if (part.endsWith(DS + projectFolderName)) {
                        return part;
                    }
                }
            }
        }
        return null;
    }

    private String getPluginFolder(String pluginName) {
        return getPluginReloadProperties(pluginName,
                        PLUGIN_FOLDER_NAME_PROP_INDEX);
    }

    private String getPluginProjectFolder(String pluginName) {
        return getPluginReloadProperties(pluginName,
                        PLUGIN_PROJECT_FOLDER_PROP_INDEX);
    }

    private String getPluginReloadProperties(String pluginName, int type) {
        String workspacesProp = getProperty(PROP_PLUGIN_MATCHES);
        if (workspacesProp != null && workspacesProp.trim().length() != 0) {
            String[] entries = workspacesProp.split(";");
            if (entries != null) {
                for (String entry : entries) {
                    String[] parts = entry.split(":");
                    if (parts != null && parts.length == 3
                                    && parts[PLUGIN_NAME_PROP_INDEX]
                                                    .equals(pluginName)) {
                        return parts[type];
                    }
                }
            }
        }
        return null;
    }

    private String getProperty(String property) {
        String tmp = PluginProperties.getUserProperties().getProperty(property);
        if (tmp != null && tmp.trim().length() == 0) {
            return null;
        }
        return tmp;
    }

    @Override
    public String getDescription() {
        return "Rebuilds and reloads a plugin, i.e. "
                        + "unloads alls dependant plugins and the desired plugin and reloads all previously dependant plugins and the plugin itself";
    }

    /**
     * @see de.renew.plugin.command.CLCommand#getArguments()
     */
    @Override
    public String getArguments() {
        return "locationsName";
    }
}