package de.renew.extras.commands;

import CH.ifa.draw.DrawPlugin;
import CH.ifa.draw.framework.Drawing;
import CH.ifa.draw.framework.DrawingEditor;
import CH.ifa.draw.framework.Figure;
import CH.ifa.draw.standard.NullDrawingEditor;

import de.renew.gui.DeclarationFigure;
import de.renew.plugin.command.CLCommand;
import de.renew.util.StringUtil;

import java.io.PrintStream;
import java.util.Collections;
import java.util.Vector;


/**
 * Command to reorganize the statements of a declaration node.
 * @author David Mosteller
 * @date Sep. 08 2016
 *
 */
public class ReorganizeDeclarationNodeCommand implements CLCommand {
    public static org.apache.log4j.Logger logger = org.apache.log4j.Logger
                    .getLogger(ReorganizeDeclarationNodeCommand.class);
    private static final String PACKAGE_DECLARATION = "package";
    private static final String IMPORT_STATEMENT = "import";
    private static final String ACCESS_STATEMENT = "access";

    @Override
    public void execute(String[] args, PrintStream response) {
        Drawing drawing = getEditor().drawing();
        DeclarationFigure df = null;
        if (drawing != null) {
            for (Figure f : Collections.list(drawing.figures())) {
                if (f instanceof DeclarationFigure) {
                    df = (DeclarationFigure) f;
                    reorganizeDeclarations(df);

                    // There should be only one if any.
                    return;
                }
            }
        } else {
            logger.trace("No drawing is selected for which the declaration node can be reorganized.");
        }
    }

    protected DrawingEditor getEditor() {
        DrawPlugin plugin = DrawPlugin.getCurrent();
        return (plugin == null) ? NullDrawingEditor.INSTANCE
                        : plugin.getDrawingEditor();
    }

    /**
     * Reorganizes the statements of a @DeclarationFigure in the following way:
     * First comes the package declaration, then the import statements,
     * then the access statements and finally the variable declarations.
     * Each group is sorted in alphabetical order and groups are separated by an empty line.
     * @param df The declaration figure to be reorganized.
     */
    public void reorganizeDeclarations(DeclarationFigure df) {
        String text = df.getText();
        if (text != null && !text.isEmpty()) {
            String[] lines = text.split("\n");
            String packageStatement = null;
            Vector<String> importStatements = new Vector<String>();
            Vector<String> accessStatements = new Vector<String>();
            Vector<String> variableDeclarations = new Vector<String>();

            // read the lines into groups of statements
            for (String line : lines) {
                if (!line.isEmpty()) {
                    if (line.startsWith(IMPORT_STATEMENT)) {
                        importStatements.addElement(line);
                    } else if (line.startsWith(ACCESS_STATEMENT)) {
                        accessStatements.addElement(line);
                    } else if (line.startsWith(PACKAGE_DECLARATION)) {
                        packageStatement = line;
                    } else {
                        variableDeclarations.addElement(line);
                    }
                }

                // sort the groups
                Collections.sort(importStatements,
                                String.CASE_INSENSITIVE_ORDER);
                Collections.sort(accessStatements,
                                String.CASE_INSENSITIVE_ORDER);
                Collections.sort(variableDeclarations,
                                String.CASE_INSENSITIVE_ORDER);

                // build reorganized declaration node text
                text = "";
                if (packageStatement != null) {
                    text += packageStatement + "\n\n";
                }

                if (!importStatements.isEmpty()) {
                    text += StringUtil.join(importStatements, "\n");
                    text += "\n\n";
                }

                if (!accessStatements.isEmpty()) {
                    text += StringUtil.join(accessStatements, "\n");
                    text += "\n\n";
                }

                text += StringUtil.join(variableDeclarations, "\n");
                df.setText(text);
                logger.trace("Reorganized declaration node.");
            }
        }
    }

    @Override
    public String getDescription() {
        return "Reorganizes the declaration node of the current drawing by sorting the entries by category and alphabetical order.";
    }

    @Override
    public String getArguments() {
        return "";
    }
}