package de.renew.extras.io.importFormats;

import CH.ifa.draw.DrawPlugin;
import CH.ifa.draw.application.DrawApplication;
import CH.ifa.draw.framework.Drawing;
import CH.ifa.draw.io.DrawingFileHelper;
import CH.ifa.draw.io.ImportHolder;
import CH.ifa.draw.io.SimpleFileFilter;
import CH.ifa.draw.io.StatusDisplayer;
import CH.ifa.draw.io.importFormats.ImportFormat;
import CH.ifa.draw.io.importFormats.ImportFormatMulti;

import de.renew.plugin.command.CLCommand;

import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.io.PrintStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;


public class ImportClCommand implements CLCommand, StatusDisplayer {
    private static final String SYNOPSIS = "Imports a file into a drawing. Usage: import filename";
    private static org.apache.log4j.Logger logger = org.apache.log4j.Logger
                    .getLogger(ImportClCommand.class);

    @Override
    public void execute(String[] args, PrintStream response) {
        ImportHolder importer = DrawPlugin.getCurrent().getImportHolder();

        List<ImportFormat> formats = new ArrayList<ImportFormat>();
        for (ImportFormat importFormat : importer.allImportFormats()) {
            if (importFormat instanceof ImportFormatMulti) {
                formats.addAll(Arrays.asList(((ImportFormatMulti) importFormat)
                                .allImportFormats()));
            } else {
                formats.add(importFormat);
            }
        }

        if (args.length == 0) {
            response.append(SYNOPSIS + "\n");
            response.append("List of available import formats: \n");
            response.append("type   description\n---------------------------\n");
            Object[] formatArgs = new Object[2];
            for (ImportFormat importFormat : formats) {
                FileFilter fileFilter = (FileFilter) importFormat.fileFilter();
                formatArgs[0] = fileFilter instanceof SimpleFileFilter
                                ? ((SimpleFileFilter) fileFilter).getExtension()
                                : "unknown";
                formatArgs[1] = importFormat.formatName();
                response.append(String.format("%-5s  %-20s %n", formatArgs));
            }
        }
        if (args.length == 1) {
            String filename = args[0];
            String writeTo = filename.substring(0, filename.lastIndexOf("."))
                            + ".rnw";
            logger.info("Trying to import file: <" + filename + "> to <"
                            + writeTo + ">.");
            File toImport = new File(filename);
            if (!toImport.exists()) {
                logger.error("File to import not found.");
            } else {
                try {
                    URL file = toImport.toURI().toURL();
                    for (ImportFormat importFormat : formats) {
                        if (importFormat.canImport(file)) {
                            logger.info("Found importer " + importFormat);
                            List<Drawing> drawings = importFormat
                                            .importFiles(new URL[] { file });
                            if (!drawings.isEmpty()) {
                                if (drawings.size() > 1) {
                                    logger.warn("Imported more than one net, but only the first is saved.");
                                }
                                DrawingFileHelper.saveAsStorableOutput(
                                                drawings.get(0),
                                                new File(writeTo), null, null,
                                                false);
                            }
                            DrawApplication gui = DrawPlugin.getGui();
                            if (gui != null) {
                                for (Drawing d : drawings) {
                                    gui.openDrawing(d);
                                }
                            }
                        }
                    }
                } catch (MalformedURLException e) {
                    e.printStackTrace();
                } catch (IOException e) {
                    logger.error("Failed to store " + writeTo);
                    e.printStackTrace();
                } catch (Exception e) {
                    /* thrown by {@see ImportFormat.importFiles(URL[])} */
                    logger.error("Failed to import " + filename);
                    e.printStackTrace();
                }
            }
        }
    }

    @Override
    public void showStatus(String message) {
        logger.info(ImportClCommand.class.getName() + " Status: " + message);
    }

    @Override
    public String getDescription() {
        return SYNOPSIS + "\nType import for a list of available formats.";
    }

    /**
     * @see de.renew.plugin.command.CLCommand#getArguments()
     */
    @Override
    public String getArguments() {
        return "fileNames";
    }
}