package de.renew.extras.statistics;


/**
 * ColoredTransitionEvent.java
 *
 * internalTime is a relative time measure:
 *    - subtract starting time (relativeTime == true)
 *    - normalize (divide time values by timeDivisor)
 *
 * the additional logging mechanisms may be switched off:
 *    - log the populations
 *    - log the delays versus the population
 *   if switched off, the dumping produces empty strings
 *
 * output formats can be switched: text, komma separated text, html (table)
 *
 *
 * Created: Fri May  2 13:31:59 2003
 *
 * @author <a href="mailto: ">Heiko Roelke</a>
 * @version
 */
import java.awt.geom.Point2D;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Locale;
import java.util.PropertyResourceBundle;
import java.util.ResourceBundle;
import java.util.Vector;


public class ColoredTransitionEvent {
    private double startingTime = 0.0;
    private static PropertyResourceBundle prb;

    //properties
    private static final boolean relativeTime = true; //subtract start time
    private static final double timeDivisor = 1.0; //divide time values by me

    private double internalTime(double t) {
        if (relativeTime) {
            return (t - startingTime) / timeDivisor;
        } else {
            return t / timeDivisor;
        }
    }

    private static final boolean logPopulation = true;
    private static final boolean logDelayPerPopulation = true;

    //output and dump formats
    public static final int TEXT = 0;
    public static final int KOMMA_SEPARATED = 1;
    public static final int HTML = 2;
    public static final String SUFFIX_TEXT = ".txt";
    public static final String SUFFIX_KOMMA_SEPARATED = ".txt";
    public static final String SUFFIX_HTML = ".html";

    //values to store
    private Hashtable<Object, Double> inTimes = new Hashtable<Object, Double>();
    private long inCounter = 0; //input
    private long outCounter = 0; //throughput
    private long maxPopulation = 0;
    private long currentPopulation = 0;
    private Vector<Double> delayTimes = new Vector<Double>();

    //used if specified (see properties)
    private Vector<Point2D.Double> populations = new Vector<Point2D.Double>();
    private Hashtable<Long, Vector<Double>> delayPerPopulation = new Hashtable<Long, Vector<Double>>();

    //assuming empty system at start time
    public ColoredTransitionEvent() {
        this(0.0);
    }

    public ColoredTransitionEvent(double time) {
        prb = (PropertyResourceBundle) ResourceBundle.getBundle(
                        "de.renew.extras.statistics.stats",
                        Locale.getDefault());

        startingTime = time;
        if (logPopulation) {
            populations.add(new Point2D.Double(internalTime(time), 0.0));
        }
    }

    private static String outputHeader(String title, int format) {
        String header = "";
        switch (format) {
        case TEXT:
            header = "";
            break;
        case KOMMA_SEPARATED:
            header = prb.getString("heading") + " " + title + "\n";
            break;
        case HTML:
            header = "<html>\n<head>\n<title>" + prb.getString("heading") + " "
                            + title + "\n</title>\n</head>\n"
                            + "<body>\n<table cellspacing=\"0\" cellpadding=\"0\" border=\"1\">\n"
                            + "<tr><th></th></tr>\n";
            break;
        }
        return header;
    }

    private static String outputHeader(int format) {
        return outputHeader("", format);
    }

    private static String outputFooter(int format) {
        String footer = "";
        switch (format) {
        case TEXT:
            footer = "\n";
            break;
        case KOMMA_SEPARATED:
            footer = "\n";
            break;
        case HTML:
            footer = "</table>\n</body>\n</html>\n";
            break;
        }
        return footer;
    }

    private static String outputLineStart(int format) {
        String s = "";
        switch (format) {
        case TEXT:
            break;
        case KOMMA_SEPARATED:
            break;
        case HTML:
            s = "<tr><td>";
            break;
        }
        return s;
    }

    private static String outputColumnSeparator(int format) {
        String s = "";
        switch (format) {
        case TEXT:
            s = " ";
            break;
        case KOMMA_SEPARATED:
            s = ", ";
            break;
        case HTML:
            s = "</td><td>";
            break;
        }
        return s;
    }

    private static String outputLineEnd(int format) {
        String s = "";
        switch (format) {
        case TEXT:
            s = "\n";
            break;
        case KOMMA_SEPARATED:
            s = "\n";
            break;
        case HTML:
            s = "</td></tr>\n";
            break;
        }
        return s;
    }

    //convenience method for empty line (two columns)
    private static String emptyLineTwo(int format) {
        return (outputLineStart(format) + outputColumnSeparator(format)
                        + outputLineEnd(format));
    }

    //convenience method for two columns per line 
    private static String outputLineTwo(int format, String first,
                                        String second) {
        return (outputLineStart(format) + first + outputColumnSeparator(format)
                        + second + outputLineEnd(format));
    }

    public void in(Object o, double t) {
        double time = internalTime(t);

        if (inTimes.containsKey(o)) {
            //doppelter Eintrag...nochmal drueber nachdenken
            throw new RuntimeException(
                            "Duplicate entry in statistical system.\n"
                                            + "Was trying to add: " + o);
        }
        inTimes.put(o, Double.valueOf(time));
        inCounter++;
        currentPopulation++;
        if (currentPopulation > maxPopulation) {
            maxPopulation = currentPopulation;
        }
        if (logPopulation) {
            populations.add(new Point2D.Double(time, currentPopulation));
        }
    }

    public void out(Object o, double oT) {
        double outTime = internalTime(oT);

        outCounter++;
        currentPopulation--;
        double inTime = inTimes.remove(o);
        double delayTime = outTime - inTime;
        delayTimes.add(delayTime);
        if (logPopulation) {
            populations.add(new Point2D.Double(outTime, currentPopulation));
        }
        if (logDelayPerPopulation) {
            Long pop = Long.valueOf(currentPopulation + 1);
            if (!delayPerPopulation.containsKey(pop)) {
                delayPerPopulation.put(pop, new Vector<Double>());
            }
            delayPerPopulation.get(pop).add(delayTime);
        }
    }

    //black token case:
    //  assuming FIFO behaviour, removing the first element of the Hashtable
    public void out(double outTime) {
        out(inTimes.keys().nextElement(), outTime);
    }


    /** ********************************************************************
     *  Analysing / Output
     *
     *
     */
    public String analyse(int format, String title, double sT) {
        double startTime = internalTime(startingTime);
        double stopTime = internalTime(sT);

        StringBuffer sb = new StringBuffer(outputHeader(format));

        sb.append(outputLineTwo(format, prb.getString("startTime"),
                        String.valueOf(startTime)));
        sb.append(outputLineTwo(format, prb.getString("stopTime"),
                        String.valueOf(stopTime)));
        double simTime = stopTime - startTime;
        sb.append(outputLineTwo(format, prb.getString("simTime"),
                        String.valueOf(simTime)));
        sb.append(emptyLineTwo(format));

        if (currentPopulation != 0) {
            sb.append(emptyLineTwo(format) + outputLineTwo(format,
                            prb.getString("notEmpty"), "")
                            + emptyLineTwo(format));
        }

        sb.append(outputLineTwo(format, prb.getString("totalIn"),
                        String.valueOf(inCounter)));
        sb.append(outputLineTwo(format, prb.getString("totalThru"),
                        String.valueOf(outCounter)));
        sb.append(outputLineTwo(format, prb.getString("maxPop"),
                        String.valueOf(maxPopulation)));
        sb.append(emptyLineTwo(format));

        double meanInput = inCounter / simTime;
        double meanThrough = outCounter / simTime;
        sb.append(outputLineTwo(format, prb.getString("meanIn"),
                        String.valueOf(meanInput)));
        sb.append(outputLineTwo(format, prb.getString("meanThru"),
                        String.valueOf(meanThrough)));
        sb.append(emptyLineTwo(format));

        //delay time
        double timeCounter = 0;
        for (int i = 0; i < delayTimes.size(); i++) {
            double dTime = delayTimes.get(i).doubleValue();
            timeCounter += dTime;
        }
        double meanDelay = timeCounter / outCounter;
        sb.append(outputLineTwo(format, prb.getString("meanDelay"),
                        String.valueOf(meanDelay)));
        //approximate variance
        double varCount = 0.0;
        for (int i = 0; i < delayTimes.size(); i++) {
            double distance = delayTimes.get(i).doubleValue()
                            - meanDelay;
            varCount += (distance * distance);
        }
        double approxDelayVariance = varCount / (delayTimes.size() - 1);
        sb.append(outputLineTwo(format, prb.getString("delayVar"),
                        String.valueOf(approxDelayVariance)));
        sb.append(outputLineTwo(format, prb.getString("delayDev"),
                        String.valueOf(Math.sqrt(approxDelayVariance))));
        sb.append(emptyLineTwo(format));

        //mean population
        double meanPopulation = meanThrough * meanDelay;
        sb.append(outputLineTwo(format, prb.getString("meanPop"),
                        String.valueOf(meanPopulation)));

        return sb.toString();
    }

    public String analyse(String title, double sT) {
        return (analyse(TEXT, title, sT));
    }

    public String analyse(double sT) {
        return (analyse(TEXT, "", sT));
    }

    public String dumpPopulations(int format) {
        StringBuffer sb = new StringBuffer(outputHeader(format));
        if (format == HTML) {
            sb.append(outputLineStart(format) + "time"
                            + outputColumnSeparator(format) + "population"
                            + outputLineEnd(format));
        }
        Enumeration<Point2D.Double> allPoints = populations.elements();
        while (allPoints.hasMoreElements()) {
            Point2D.Double p = allPoints.nextElement();
            sb.append(outputLineStart(format) + p.getX()
                            + outputColumnSeparator(format) + p.getY()
                            + outputLineEnd(format));
        }
        sb.append(outputFooter(format));
        return sb.toString();
    }

    public String dumpPopulations() {
        return dumpPopulations(TEXT);
    }

    public String dumpDelayVsPopulations() {
        StringBuffer sb = new StringBuffer();
        Enumeration<Long> allData = delayPerPopulation.keys();
        while (allData.hasMoreElements()) {
            Long key = allData.nextElement();
            Enumeration<Double> delayEnum = (delayPerPopulation.get(key))
                            .elements();
            long count = 0;
            double delaySum = 0.0;
            while (delayEnum.hasMoreElements()) {
                count++;
                delaySum += delayEnum.nextElement().doubleValue();
            }
            sb.append(key.longValue() + " " + (delaySum / count) + "\n");
        }
        return sb.toString();
    }

    public void report(double stopTime) {
        //wegen doppeltem internem Aufruf hier Zeit nicht umwandeln!
        System.out.println(analyse(stopTime));
    }

    public String reportHTML(String title, double stopTime) {
        //wegen doppeltem internem Aufruf hier Zeit nicht umwandeln!
        return (analyse(HTML, title, stopTime));
    }
}