package de.renew.plugin;

import de.renew.plugin.load.ModuleConfigurationException;
import de.renew.plugin.load.ModuleLayerByNameException;
import de.renew.plugin.load.ModuleLayerCreationException;
import de.renew.plugin.load.ModuleLayerRemoveException;
import de.renew.plugin.locate.PluginLocationFinders;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.lang.module.Configuration;
import java.lang.module.ModuleDescriptor;
import java.lang.module.ModuleFinder;
import java.lang.module.ModuleReference;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URISyntaxException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.Vector;
import java.util.jar.JarFile;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

/**
 * Responsible for loading and unloading modules.
 *
 * Before a {@link IPlugin} is loaded, a {@link Module}
 * needs to be created for it, which will reside inside
 * a {@link ModuleLayer}. Each layer has a set of parents
 * that corresponds to the required modules.
 *
 * For example, a plugin C which requires plugins A and B
 * will have the layer containing plugin A's module and
 * the layer containing plugin B's module as parents.
 *
 * This class keeps track of the currently loaded layers
 * and also which layers have no children and could potentially
 * be removed using the {@link de.renew.plugin.command.UnloadCommand}
 */
public class ModuleManager {
    private final List<ModuleLayer> _loadedLayers = new ArrayList<>();
    private final List<ModuleLayer> _removableLayers = new ArrayList<>();
    public static org.apache.log4j.Logger logger = org.apache.log4j.Logger
            .getLogger(ModuleManager.class);
    private final List<IPlugin> modularPlugins = new ArrayList<>();
    private final Path[] moduleLocations;

    private static final String MODULE_LOCATIONS_PROPERTY = "pluginLocations";
    private static final String PLUGIN_FOLDER = "/plugins";

    /**
     * Creates a new ModuleManager. This class is only meant
     * to be instantiated by the {@link PluginManager}.
     */
    ModuleManager() {
        _loadedLayers.add(ModuleLayer.boot());
        moduleLocations = getModuleLocations();
    }

    /**
     * Creates a ModuleLayer for the the given arguments
     * Enforced: 1 plugin = 1 module = 1 module Layer
     *
     * @param moduleJarPath     Path to Jar
     * @param parentClassLoader The plugin Class loader of the Module
     * @return The new Module Layer
     * @author Laif-Oke Clasen
     */
    public ModuleLayer createPluginLayer(Path moduleJarPath, ClassLoader parentClassLoader) {

        ModuleReference moduleToLoad = getModuleReferenceFromPath(moduleJarPath);

        var requiredModules = moduleToLoad.descriptor().requires();

        // Extracts the relevant dependencies for the module layer
        List<String> parentLayerNames = getRequiredModuleNames(requiredModules);

        List<ModuleLayer> parentLayers = new ArrayList<ModuleLayer>();
        List<Configuration> parentConfigurations = new ArrayList<Configuration>();

        for (ModuleLayer loadedLayer : _loadedLayers) {
            for (String parentLayerName : parentLayerNames) {
                List<String> modulesInLoadedLayer = getModuleNamesInLayer(loadedLayer);
                if (modulesInLoadedLayer.contains(parentLayerName) && !parentLayers.contains(loadedLayer)) {
                    parentLayers.add(loadedLayer);
                    _removableLayers.remove(loadedLayer);
                    parentConfigurations.add(loadedLayer.configuration());
                }
            }
        }

        if (parentConfigurations.isEmpty()) {
            // None of the required module layers could be located
            // TODO: How do we determine if >0 but not all layers could be located?
            throw new ModuleLayerCreationException(moduleToLoad, parentLayerNames, parentLayers);
        }

        ModuleFinder pluginFolderFinder = ModuleFinder.of(moduleLocations);

        //"resolve" tries to find all modules that match the given ModuleNames and all the dependencies of the found modules.
        //A Configuration is created, that contains the readability graph.
        Configuration newConfiguration = Configuration.resolve(
                ModuleFinder.of(), // No before-finder necessary, use empty finder
                parentConfigurations,
                pluginFolderFinder,
                Set.of(moduleToLoad.descriptor().name())
        );

        //Based on the Configuration a ModuleLayer along with a the given Classloader is created.
        ModuleLayer newLayer = ModuleLayer
                .defineModulesWithOneLoader(newConfiguration, parentLayers, parentClassLoader)
                .layer();

        if (newLayer.modules().size() != 1) {
            throw new ModuleLayerCreationException(moduleToLoad.descriptor().name(), newLayer.modules());
        }
        ServiceLookupInfrastructure.addServicesOfModules(newLayer.modules());
        _loadedLayers.add(newLayer);
        _removableLayers.add(newLayer);

        return newLayer;
    }

    /**
     * Collects the module locations from:
     * - The default folder {@link ModuleManager#PLUGIN_FOLDER}
     * - The user properties {@link PluginManager#PLUGIN_LOCATIONS_PROPERTY}
     *
     * The default folder is always included, the user properties 
     *
     * @return All found paths
     */
    private Path[] getModuleLocations() {
        Vector<Path> result = new Vector<>();
        Properties userProps = PluginProperties.getUserProperties();
        if (userProps.getProperty(MODULE_LOCATIONS_PROPERTY) == null) {
            logger.info("no additional plugin locations set.");
        }
        Collection<String> locations = PropertyHelper.parsePathListString(
                userProps.getProperty(MODULE_LOCATIONS_PROPERTY, ""));
        Iterator<String> files = locations.iterator();
        files.forEachRemaining(file -> {
            logger.debug("location: " + file);
            result.add(new File(file).toPath());
        });
        Path[] arr = new Path[result.size() + 1]; // + 1 to leave space for the '/plugin' folder
        result.toArray(arr);
        try {
            arr[arr.length - 1] = new File(PluginManager.getLoaderLocation().toURI().getPath() + PLUGIN_FOLDER).toPath();
        } catch (URISyntaxException e) {
            logger.error("Could not add plugin location due to syntax exception.", e);
        }
        return arr;
    }

    /**
     * Returns the reference to the module at the given path.
     *
     * @param moduleJarPath Path to the modular JAR.
     * @return The reference to the module
     */
    private ModuleReference getModuleReferenceFromPath(Path moduleJarPath) {
        ModuleFinder jarModuleFinder = ModuleFinder.of(moduleJarPath);

        Set<ModuleReference> referencedModulesInJar = jarModuleFinder.findAll();

        if (referencedModulesInJar.size() != 1) {
            throw new ModuleConfigurationException("A modular jar must contain exactly one module!");
        }

        return referencedModulesInJar.stream().findFirst().get();
    }

    /**
     * Converts the ModuleDescriptor's requires into a list of Renew plugin modules.
     * <p>{@code transitive de.renew.gui} => {@code de.renew.gui}</p>
     * Any requirements not starting with <i>de.renew</i> will be ignored.
     *
     * @param requires The requirements of the module to load.
     * @return The list of required Renew plugin module names
     */
    private List<String> getRequiredModuleNames(Set<ModuleDescriptor.Requires> requires) {
        List<String> requiredModuleNames = new ArrayList<>();
        for (ModuleDescriptor.Requires require : requires) {
            String requireString = require.toString();
            if (requireString.contains("de.renew") && !requireString.contains("static")) {
                int beginIndex = requireString.indexOf("de.renew");
                String requiredModuleName = requireString.toLowerCase().substring(beginIndex);
                requiredModuleNames.add(requiredModuleName);
            }
            if (requireString.contains("CH.ifa.draw") && !requireString.contains("static")) {
                requiredModuleNames.add("CH.ifa.draw");
            }
        }

        if (requiredModuleNames.isEmpty()) {
            requiredModuleNames.add("de.renew.loader");
        }
        return requiredModuleNames;
    }

    /**
     * Remove the module layer of a single module
     * Enforced : 1 plugin = 1 module = 1 module layer
     *
     * @param moduleName Name of the module must follow this pattern: de.renew.util, Ch.ifa.draw or Renew_Gui
     * @throws ModuleLayerRemoveException Thrown if the layer isn't removable or the module name doesn't follow the conventions
     * @author Gerrit Kieslich
     * @author Laif-Oke Clasen
     */
    void removeLayerOfModule(String moduleName) throws ModuleLayerRemoveException {
        checkModuleName(moduleName);
        String convertedName = getModuleNameFromPluginName(moduleName);

        var layersWithModule = getLayersContainingModule(convertedName);

        if (layersWithModule.size() != 1) {
            throw new ModuleConfigurationException(moduleName + " must be in exactly one layer!");
        }

        var layerToRemove = layersWithModule.get(0);

        if (!_removableLayers.contains(layerToRemove)) {
            throwRemoveException(moduleName);
        }

        ServiceLookupInfrastructure.removeServicesOfModules(layerToRemove.modules());
        _loadedLayers.remove(layerToRemove);
        _removableLayers.remove(layerToRemove);
        System.gc();
        updateRemovableLayers();
    }

    /**
     * Converts the plugin name to module/layer name
     *
     * @param pluginName The name of the plugin
     * @return The module/layer name
     * @author Laif-Oke Clasen
     * @author Jonte Johnsen
     * @author Relana Streckenbach
     */
    public static String getModuleNameFromPluginName(String pluginName) {
        Path pluginPath = null;
        Collection<PluginProperties> propsVector = PluginLocationFinders.getInstance().findPluginLocations();
        for (PluginProperties props : propsVector) {
            if (props.getName().equals(pluginName)) {
                try {
                    pluginPath = Paths.get(props.getURL().toURI()) ;
                } catch (URISyntaxException e) {
                    logger.error("Could not create URI from plugin location: " + e);
                }
                break;
            }
        }
        String moduleName = "";
        ModuleFinder finder = ModuleFinder.of(pluginPath);
        moduleName = finder.findAll().iterator().next().descriptor().name();

        return moduleName;
    }

    /**
     * Filter the currently loaded layers for modules with the queried {@code moduleName}
     *
     * @param moduleName The name of the module to find
     * @return All layers containing the queried module
     */
    private List<ModuleLayer> getLayersContainingModule(String moduleName) {
        return _loadedLayers.stream()
                // Only keep layers that contain the module "moduleName"
                .filter(ml -> ml.modules()
                        .stream()
                        .map(Module::getName)
                        .anyMatch(name -> name.equals(moduleName))
                )
                .collect(Collectors.toList());
    }

    private void throwRemoveException(String moduleName) {
        List<String> childrenList = new ArrayList<String>();
        for (ModuleLayer l : _loadedLayers) {
            for (ModuleLayer l1 : l.parents()) {
                if (l1.toString().contains(moduleName)) {
                    childrenList.add(l.toString());
                }
            }
        }
        throw new ModuleLayerRemoveException(moduleName, childrenList);
    }

    /**
     * Checks the currently loaded layers for elements that
     * are not parents to any other layer, i.e., all layers
     * without any children.
     *
     * @author Laif-Oke Clasen
     */
    void updateRemovableLayers() {
        List<ModuleLayer> allParents = new ArrayList<ModuleLayer>();
        for (ModuleLayer layer : _loadedLayers) {
            allParents.addAll(layer.parents());
        }
        for (ModuleLayer layer : _loadedLayers) {
            if (!_removableLayers.contains(layer) && !allParents.contains(layer)) {
                _removableLayers.add(layer);
            }
        }
    }

    /**
     * Remove the module layers of the modules
     * Enforced : 1 plugin = 1 module = 1 module layer
     *
     * @param moduleNames The List of the module names
     * @throws ModuleLayerRemoveException Thrown if the list of layer isn't removable or the module name doesn't follow the conventions
     * @author Laif-Oke Clasen
     */
    void removeLayersOfModules(List<String> moduleNames) throws ModuleLayerRemoveException {

        for (String moduleName : moduleNames) {
            checkModuleName(moduleName);
        }

        List<String> requestedModulesToRemove = moduleNames.stream()
                .map(ModuleManager::getModuleNameFromPluginName)
                .collect(Collectors.toList());

        List<ModuleLayer> toRemoveLayers = getLayersContainingModuleNames(requestedModulesToRemove);
        List<ModuleLayer> orderedToRemoveLayers = calculateRemovalOrder(toRemoveLayers);

        List<String> sortedModuleNames = new ArrayList<>();
        for (ModuleLayer layer : orderedToRemoveLayers) {
            for (String moduleToRemove : requestedModulesToRemove) {
                List<String> modulesInLoadedLayer = getModuleNamesInLayer(layer);
                if (modulesInLoadedLayer.contains(moduleToRemove)) {
                    sortedModuleNames.add(sortedModuleNames.size(), moduleToRemove);
                }
            }
        }

        for (String moduleName : sortedModuleNames) {
            removeLayerOfModule(moduleName);
        }
    }

    /**
     * Returns a list of module layers containing the given names
     *
     * @param modulesToFind The names of modules to find
     * @return The list of found module layer
     * @author Laif-Oke Clasen
     */
    private List<ModuleLayer> getLayersContainingModuleNames(List<String> modulesToFind) {
        List<ModuleLayer> layersWithFoundModules = new ArrayList<ModuleLayer>();
        List<String> foundModules = new ArrayList<String>();
        for (ModuleLayer loadedLayer : _loadedLayers) {
            List<String> moduleNamesInLayer = getModuleNamesInLayer(loadedLayer);
            for (String moduleName : modulesToFind) {
                if (moduleNamesInLayer.contains(moduleName)) {
                    layersWithFoundModules.add(loadedLayer);
                    foundModules.add(moduleName);
                }
            }
        }

        if (layersWithFoundModules.size() != modulesToFind.size()) {
            // Not all requested modules could be found
            var notFoundModules = modulesToFind.stream()
                    .filter(m -> !foundModules.contains(m))
                    .collect(Collectors.toList());

            throw new ModuleLayerByNameException(notFoundModules);
        }
        return layersWithFoundModules;
    }

    /**
     * Order the toRemoveLayers so that the children of the layer with index X are part of the set of elements
     * contained in the index range from 0 to X-1. This ensures that once layer X is removed, all its children
     * have already been removed.
     *
     * @param toRemoveLayers The list of layers to remove
     * @return A list of module layers which must be removed in the order returned.
     */
    private List<ModuleLayer> calculateRemovalOrder(List<ModuleLayer> toRemoveLayers) {
        List<ModuleLayer> sortedToRemoveLayers = new ArrayList<>();

        List<ModuleLayer> tmpChildren = new ArrayList<>();

        for (int k = 0; k < toRemoveLayers.size(); k++) {
            for (ModuleLayer toRemoveLayer : toRemoveLayers) {
                tmpChildren.clear();
                if (!sortedToRemoveLayers.contains(toRemoveLayer)) {
                    tmpChildren.addAll(getChildLayersOfLayer(toRemoveLayer));
                    for (ModuleLayer sortedToRemoveLayer : sortedToRemoveLayers) {
                        tmpChildren.remove(sortedToRemoveLayer); // Since sortedToRemoveLayer will have been removed in an earlier step
                    }
                    if (tmpChildren.isEmpty()) {
                        sortedToRemoveLayers.add(sortedToRemoveLayers.size(), toRemoveLayer);
                    }
                }
            }
        }
        if (toRemoveLayers.size() != sortedToRemoveLayers.size()) { // We were unable to find all layers!
            HashSet<ModuleLayer> modulesNotFound = new HashSet<>(toRemoveLayers);
            modulesNotFound.removeAll(new HashSet<>(sortedToRemoveLayers));
            throw new ModuleLayerRemoveException(getMapLayerToChildModules(modulesNotFound));
        }
        return sortedToRemoveLayers;
    }

    /**
     * Get a map of {@code ModuleLayer}s as keys and a list of all modules in
     * their child layers as values.
     * @param layers the parent layers which become keys
     * @return a map of layer to child modules
     */
    private Map<ModuleLayer, List<Module>> getMapLayerToChildModules(Set<ModuleLayer> layers) {
        return layers.stream().collect(
                        Collectors.toMap(
                                layer -> layer,
                                this::getModulesOfLayersChildren
                        )
                );
    }

    /**
     * Get all modules in the child layers of the specified parent layer
     * @param parent the layer who's children's modules are to be returned
     * @return the modules contained in all child layers of the parent layer
     */
    private List<Module> getModulesOfLayersChildren(ModuleLayer parent) {
        return getChildLayersOfLayer(parent).stream()
                .flatMap(ml -> ml.modules().stream())
                .collect(Collectors.toList());
    }

    /**
     * Returns the names of all modules contained in the layer
     *
     * @param layer The layer containing the modules.
     * @return A list of Strings representing the module names contained in the layer.
     */
    public static List<String> getModuleNamesInLayer(ModuleLayer layer) {
        return layer.modules().stream()
                .map(Module::getName)
                .collect(Collectors.toList());
    }

    /**
     * Returns the names of all modules contained in the layers
     *
     * @param layers The layers containing the modules.
     * @return A list of Strings representing the module names contained in the layer.
     */
    public static List<String> getModuleNamesInLayers(List<ModuleLayer> layers) {
        return layers.stream()
                .map(ModuleManager::getModuleNamesInLayer)
                .flatMap(Collection::stream)
                .collect(Collectors.toList());
    }

    /**
     * Returns a list of child layers from the {@link ModuleLayer}.
     * A {@link ModuleLayer} B is considered to be a child of layer A,
     * if layer B's parents contains layer A.
     *
     * @param parent The parent layer of the children
     * @return The list of children
     * @author Laif-Oke Clasen
     */
    private List<ModuleLayer> getChildLayersOfLayer(ModuleLayer parent) {
        return _loadedLayers.stream()
                .filter(ml -> ml.parents().contains(parent))
                .collect(Collectors.toList());
    }

    /**
     * Checks if the module name matches with the conventions
     *
     * @param moduleName Name of the module must follow this pattern: de.renew.util, Ch.ifa.draw or Renew_Gui
     * @throws ModuleLayerRemoveException Thrown if the layer isn't removable or the module name doesn't follow the Conventions
     * @author Laif-Oke Clasen
     */
    private void checkModuleName(String moduleName) throws ModuleLayerRemoveException {
        Pattern p = Pattern.compile("(^Renew(_\\w+)+)|(^Renew(\\s\\w+)+)|(^de.renew(.\\w+)+)|(^CH.ifa.draw)|(loader)");
        Matcher m = p.matcher(moduleName);
        boolean b = m.matches();
        if (!b) {
            throw new ModuleLayerRemoveException(moduleName);
        }
    }

    /**
     * This method is called to find which module a class belongs to.
     * The name is extracted from the className by cutting the string.
     *
     * @param className Name of the class, that you need to find the module's ClassLoader for.
     * @return the ClassLoader of the Module, that your class is part of
     * or the classloader of this class if this class is in the unnamed module
     */
    public ClassLoader getModuleClassLoaderForClass(String className) {

        if (!this.getClass().getModule().isNamed()) {
            // This class is in the unnamed module.
            // This means Renew was not started as a modular application, for example by an ant task
            return this.getClass().getClassLoader();
        }

        var noStartingSlash = className.startsWith("/") ? className.substring(1) : className;
        var normalizedClassName = noStartingSlash.replace('/', '.');

        int packageEndIndex = normalizedClassName.lastIndexOf('.');

        if (packageEndIndex == -1) {
            logger.debug("The class " + className + "is not fully qualified. Can be ignored if " +
                    "the call originated from an InscriptionParser.");
            return null;
        }

        String packageName = normalizedClassName.substring(0, packageEndIndex);

        for (ModuleLayer layer : _loadedLayers) {
            for (Module module : layer.modules()) {
                if (module.getPackages().contains(packageName)) {
                    return module.getClassLoader();
                }
            }
        }

        return null;
    }

    /**
     * Adds the plugin to the list of currently loaded
     * plugins that were loaded from a modular Jar file.
     * @param plugin the plugin to register
     */
    public void registerPluginAsModular(IPlugin plugin) {
        modularPlugins.add(plugin);
    }

    /**
     * Removes the plugin from the list of currently loaded
     * modular plugins.
     * @param plugin the plugin to unregister
     */
    public void unregisterModularPlugin(IPlugin plugin) {
        modularPlugins.remove(plugin);
    }

    /**
     * Checks if a plugin was registered as being modular.
     * @param plugin the plugin to be checked
     * @return true if the plugin is modular, else false
     */
    public boolean isModularPlugin(IPlugin plugin) {
        return modularPlugins.contains(plugin);
    }
}