package de.renew.plugin;


import java.lang.module.ModuleDescriptor;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The ServiceLookupInfrastructure (SLI) manages the service interfaces and providers.
 * For this purpose, the modules of a module layer are transferred to the SLI, which searches them for service interfaces and providers.
 * The interfaces and providers found are then managed in the SLI.
 * Removing the interfaces and providers is analogous to this.
 *
 * The SLI can be used by passing it a service interface as a class.
 * The SLI will return the module layers of the service providers.
 * The Java ServiceLoader can then be applied to these.
 *
 * @author Laif-Oke Clasen
 * @version 1.0
 */
public class ServiceLookupInfrastructure {

    // A map from the service interfaces to the module layers of the service providers
    private static final Map<Class<?>, List<ModuleLayer>> _providerLocations = new HashMap<>();
    // A map from the service interface string to the service interface
    private static final Map<String, Class<?>> _serviceInterfaces = new HashMap<>();
    // A map of modules to a list of service interface strings contained in the module
    private static final Map<Module, List<String>> _serviceInterfaceLocations = new HashMap<>();

    /**
     * Adds the service providers and the service interfaces of the given modules to the service lookup infrastructure.
     * @param modules The service provider and service interface modules to be added to the service lookup infrastructure.
     * @author Laif-Oke Clasen
     */
    public static void addServicesOfModules(Set<Module> modules)
    {
        for(Module module: modules)
        {
            addServiceInterfacesOfModule(module);
            addServiceProvidersOfModule(module);
        }
    }

    private static Set<String> extractServiceStringsOfModule(Module module)
    {
        Set<String> usesServiceStrings = module.getDescriptor().uses();
        Set<ModuleDescriptor.Provides> provServiceStrings = module.getDescriptor().provides();

        // Combines the service Strings of the provides- and uses-entries in the module info.
        Set<String> serviceStrings = new HashSet<String>();
        for(ModuleDescriptor.Provides serviceString: provServiceStrings)
        {
            serviceStrings.add(serviceString.service());
        }
        serviceStrings.addAll(usesServiceStrings);
        return serviceStrings;
    }

    private static void addServiceInterfacesOfModule(Module module) {

        Set<String> serviceStrings = extractServiceStringsOfModule(module);

        // For each service string it is checked whether the service interface and the location of the service
        // interface are already included in the service lookup infrastructure.
        for(String serviceString : serviceStrings)
        {
            try {
                Class<?> serviceInterface = module.getClassLoader().loadClass(serviceString);
                Module serviceInterfaceModule = serviceInterface.getModule();

                if(!_serviceInterfaces.containsKey(serviceString))
                {
                    _serviceInterfaces.put(serviceString,serviceInterface);
                }

                if(_serviceInterfaceLocations.containsKey(serviceInterfaceModule))
                {
                    List<String> value = _serviceInterfaceLocations.get(serviceInterfaceModule);
                    if(!value.contains(serviceString))
                    {
                        value.add(serviceString);
                        _serviceInterfaceLocations.put(serviceInterfaceModule,value);
                    }
                }
                else
                {
                    List<String> value = new ArrayList<String>();
                    value.add(serviceString);
                    _serviceInterfaceLocations.put(serviceInterfaceModule,value);
                }
            } catch (ClassNotFoundException e) {
                System.out.println("The service interface of the service provider " + serviceString + "is not found.");
                e.printStackTrace();
            }
        }
    }



    private static void addServiceProvidersOfModule(Module providerModule)
    {
        Set<ModuleDescriptor.Provides> providedServices = providerModule.getDescriptor().provides();
        ModuleLayer providerLayer = providerModule.getLayer();

        // For every "provide" statement in the module-info the service interface is mapped to the layer of the
        // service provider.
        for(ModuleDescriptor.Provides providedService : providedServices)
        {
            String service = providedService.service();
            Class<?> serviceInterface = _serviceInterfaces.get(service);

            // If the Key doesnt exists a new key-value-tuple is added to the map.
            // else the value of the key-value-tuple is updated.
            if(!_providerLocations.containsKey(serviceInterface))
            {
                List<ModuleLayer> value = new ArrayList<ModuleLayer>();
                value.add(providerLayer);
                _providerLocations.put(serviceInterface,value);
            }
            else {
                List<ModuleLayer> oldValue = _providerLocations.get(serviceInterface);
                List<ModuleLayer> newValue = _providerLocations.get(serviceInterface);
                newValue.add(providerLayer);
                _providerLocations.replace(serviceInterface,oldValue,newValue);
            }
        }
    }

    /**
     * Returns a list of module layers.
     * These module layers contain the service provider for the given serviceInterface.
     * @param serviceInterface The service interface as a class.
     * @return List of the service providers module layers.
     * @author Laif-Oke Clasen
     */
    public static List<ModuleLayer> getLocationsOfServiceProvider(Class<?> serviceInterface)
    {
        return _providerLocations.get(serviceInterface);
    }

    /**
     * Removes the service providers and the service interfaces of the given modules from the service lookup infrastructure.
     * @param modules The service provider and the service interfaces of the modules that should be removed from the service lookup infrastructure.
     * @author Laif-Oke Clasen
     */
    public static void removeServicesOfModules(Set<Module> modules)
    {
        for(Module module: modules)
        {
            removeServiceProvidersOfModule(module);
            removeServiceInterfacesOfModule(module);

        }
    }

    private static void removeServiceInterfacesOfModule(Module providerModule)
    {
        if(_serviceInterfaceLocations.containsKey(providerModule))
        {
            for(String service : _serviceInterfaceLocations.get(providerModule))
            {
                _serviceInterfaces.remove(service);
            }
            _serviceInterfaceLocations.remove(providerModule);
        }
    }


    private static void removeServiceProvidersOfModule(Module providerModule)
    {
        Set<ModuleDescriptor.Provides> providedServices = providerModule.getDescriptor().provides();
        ModuleLayer providerLayer = providerModule.getLayer();

        // For every "provide" statement in the module-info the mapping form the service interface to the service
        // provider module layer is removed
        for(ModuleDescriptor.Provides providedService : providedServices)
        {
            String service = providedService.service();
            Class<?> serviceInterface = _serviceInterfaces.get(service);
            List<ModuleLayer> value = _providerLocations.get(serviceInterface);
            // If the list of the service providers module layer bigger than 1, than the mapping entry is updated.
            // Else the mapping entry is removed.
            if(value.size() > 1)
            {
                List<ModuleLayer> oldValue = _providerLocations.get(serviceInterface);
                value.remove(providerLayer);
                _providerLocations.replace(serviceInterface, oldValue, value);
            }
            else {
                _providerLocations.remove(serviceInterface);
            }
        }
    }
}