package de.renew.plugin.load;

import de.renew.plugin.ModuleManager;

import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

public class ModuleLayerRemoveException extends ModuleConfigurationException {

    /**
     * Creates the exception for the case if the module layer that should be removed have children.
     * @param moduleLayerName The module layer that should be removed
     * @param childrenOfModuleLayer The children of the module layer
     * @author Laif-Oke Clasen
     */
    public ModuleLayerRemoveException(String moduleLayerName, List<String> childrenOfModuleLayer)
    {
        super(buildMessage(moduleLayerName,childrenOfModuleLayer));
    }

    /**
     * Creates the exception for the case if the module names doesnt follow the conventions.
     * @param moduleName
     */
    public ModuleLayerRemoveException(String moduleName)
    {
        super("The moduleName "+ moduleName+ " does not follow the conventions. " +
                "It must resemble de.renew.util, CH.ifa.draw or Renew_Gui");
    }

    /**
     * Creates the exception for the case if one or more module layer should be removed and all have children.
     * And these children should not be removed
     * @param layerToChildren The names of children
     * @author Laif-Oke Clasen
     */
    public ModuleLayerRemoveException(Map<ModuleLayer, List<Module>> layerToChildren)
    {
        super(buildMessageWithChildren(layerToChildren));

    }

    /**
     * Creates the exception message for the case if one or more module layer should be removed and all have children.
     * And these children should not be removed
     * @param layerToChildren The names of children
     * @return The exception Message
     * @author Laif-Oke Clasen
     */
    private static String buildMessageWithChildren(Map<ModuleLayer, List<Module>> layerToChildren)
    {
        StringBuilder msg = new StringBuilder();
        layerToChildren.forEach((parentLayer, children) -> {
            msg.append(
                    buildMessage(
                            String.join(", ", ModuleManager.getModuleNamesInLayer(parentLayer)),
                            children.stream().map(Module::getName).collect(Collectors.toList())
                    )
            );
            msg.append("\n");
        });
        return msg.toString();
    }

    /**
     * Creates the exception message for the case that the module layer which should be removed has children.
     * @param moduleLayerName The module layer that should be removed
     * @param childrenOfModuleLayer The children of the module layer
     * @return The exception message
     * @author Laif-Oke Clasen
     */
    private static String buildMessage(String moduleLayerName, List<String> childrenOfModuleLayer)
    {
        return "ModuleLayerRemoveException: " + moduleLayerName
                + " has children: " + String.join(" | ", childrenOfModuleLayer);
    }
}
