<!-- Adapted from: https://www.writethedocs.org/guide/writing/beginners-guide-to-docs/ -->

# Ontology

## User Guide

_Ontology_ manages [Renew][rnw] and [Paose][paose] Ontology.

### Features

- Use ontology providers to load ontology models from ontology resources like feature structure nets or [Protégé][protege] resources
- Use ontology generators to generate artifacts from those ontology models, like Java or JavaScript classes.

### Example Usage

1. Create a sns file at `Ontology/src/main/resources/sns`
2. Run the command `./gradlew Ontology` to generate an Ontology of that sns file
3. The newly created Java File is stored in `Ontology/build/generated/`

### Provided Commands

_Ontology_ does not provide any commands.

### Installation

Install _Ontology_ by putting the JAR file inside the `dist/plugins` directory.

It will be automatically loaded at startup of Renew.

### Dependencies

This plugin depends on the following plugins and external libraries:

<!-- must match the `module-info.java` -->

- `FS`: Used for working with feature structures
- `ant`: Uses Ant tasks for ontology generation
- `Velocity`: Used to render files from ontology models

### Configuration

_Ontology_ does not read any properties from the `renew.properties` file in the Renew root directory.

## Developer Guide

### Core Classes

- `OntologyPlugin`: Manages the initialisation of the plugin
- `FSOntologyProvider`: Provides an ontology model given a feature structure
- `JavaOntologyGenerator`: Generates Java artifacts out of an ontology model
- `JavaNoAgentOntologyGenerator`: Ontology generator for Java artifacts
- `JavaScriptOntologyGenerator`: Ontology generator for JavaScript files
- `JavaScriptAmdOntologyGenerator`: Ontology generator for AMD-compliant JavaScript files

### How to Extend

_Ontology_ exposes two Interfaces, `OntologyProvider` and `OntologyGenerator` which have to be implemented to extend the plugin.
Also, your created classes need to be registered with the `OntologyPlugin` class itself, as seen in the method `registerClasses()` at `de/renew/ontology/OntologyPlugin.java:53`.

### Additional Information

<!-- Common issues, explanations for inconsistencies, etc. -->

Following sections will focus on already implemented providers and generators.

###Providers

Providers implement the `OntologyProvider` interface and provide an ontology model for the Renew application
given by different kind of resources.

#### Ant Usage

Use it as follows in your `build.xml`:

```xml
<provider name="{implementation}" src="{source net}" netPathRef="ref.ontology.netpath"/>
```

Instead of the provider name a fully qualified class name can be specified as a parameter to the providers configuration.

```xml
<provider className="{class name}" src="{source net}" netPathRef="ref.ontology.netpath"/>
```

#### Featured Implementations

Possible values that you can use for `{implementation}`:

* **feature-structure**: Expects [Feature Structure (FS)][fs] shadow net files (*.sns) as input.
* **protege** _(planned)_: Expects [Protégé][protege] files as input.


###Generators

Generators implement the `OntologyGenerator` interface and generate artifacts out of an ontology model from the Renew
application. Those artifacts can be used in other plugins to support them with domain knowledge.

#### Ant Usage

Use it as follows in your `build.xml`:

```xml
<generator name="{implementation}" destinationDir="{destination}"/>
```

Instead of the generator name a fully qualified class name can be specified as a parameter to the generators configuration.

```xml
<generator className="{class name}" destinationDir="{destination}"/>
```

#### Featured Implementations

Possible values that you can use for `{implementation}`:

* **java**: Expects an output directory as parameter, where it will generate Java classes to.
* **java-script**: Expects an output directory as parameter, where it will generate JavaScript functions to.

###Gradle Call

To create an Ontology of an sns file you have to run the command  ``` ./gradlew Ontology ```. In order to get java compiled File the sns must be located at ```Ontology/src/main/ressources/sns```.
The newly created Java File is stored in ```Ontology/build/generated/```. Each Ontology will be stored with its own Path within the file.

###Ant-Task

To easily use the Ontology plugin in your Renew project, there is an [Apache Ant][ant] task which automates triggering
providers and generators to supply your project with ontology artifacts. To use it, you have to configure your
environment with the following task definition:

```xml
<!-- originates from ant/commontasks.xml:40 -->
<taskdef name="ontology"
         classname="de.renew.ontology.ant.OntologyTask"
         classpathref="ref.renewclasspath"/>
```

Afterwards, you can embed the `<ontology>...</ontology>` task into your project's targets. There you have to
specify the provided ontology nets and the generators you want to use. Take for example following configuration:

```xml
<!-- originates from ExerciseWorksheetDefinition/build.xml:199 -->
<ontology>
    <!-- Specifying a Feature Structure source net -->
    <provider name="feature-structure"
              src="${dir.build.classes}/ontology/ExerciseWorksheetDefinitionOntology.sns"
              netPathRef="ref.ontology.netpath"/>

    <!-- Let the plugin generate Java classes from the model -->
    <generator name="java" destinationDir="${dir.build.gensrc}"/>

    <!-- Let the plugin generate JavaScript functions from the model -->
    <generator name="java-script" destinationDir="${dir.build.gensrc}"/>
</ontology>
```

<!-- Contribution Info -->

- Originally written by Konstantin Simon Maria Möllers and Benjamin Teuber
- Contains contributions from Christian Bracker (1bracker)

<!-- License -->

[rnw]: http://renew.de/
[paose]: http://paose.net/
[protege]: http://protege.stanford.edu/
[fs]: http://en.wikipedia.org/wiki/Feature_structure
[ant]: http://ant.apache.org/
