package de.renew.ontology;

import de.renew.ontology.entities.Ontology;
import de.renew.ontology.exception.OntologyGeneratorException;
import de.renew.ontology.exception.OntologyProviderException;

import java.util.LinkedList;
import java.util.List;


/**
 * Converter to load ontology models and generate source code from them.
 *
 * @author Konstantin Simon Maria Moellers
 * @version 2015-01-13
 */
final public class OntologyConverter {

    /**
     * Generator to render files from ontology models.
     */
    private final List<OntologyProvider> providers = new LinkedList<OntologyProvider>();

    /**
     * Generator to render files from ontology models.
     */
    private final List<OntologyGenerator> generators = new LinkedList<OntologyGenerator>();

    /**
     * Converts an ontology net to source code.
     *
     * @throws de.renew.ontology.exception.OntologyProviderException
     * @throws de.renew.ontology.exception.OntologyGeneratorException
     */
    public void convert() throws OntologyProviderException,
                    OntologyGeneratorException {
        // Loads the ontology from all providers.
        List<Ontology> ontologyList = loadOntology();

        for (Ontology ontology : ontologyList) {
            // Generate each ontology.
            generateOntology(ontology);
        }
    }

    /**
     * Generates all ontology files from a single ontology.
     *
     * @param ontology data model to generate
     * @throws OntologyGeneratorException
     */
    private void generateOntology(Ontology ontology)
                    throws OntologyGeneratorException {
        for (OntologyGenerator generator : generators) {
            generator.generate(ontology);
        }
    }

    /**
     * Adds an ontology provider to the chain of used providers.
     *
     * @param provider the provider to add.
     */
    public void addProvider(OntologyProvider provider) {
        providers.add(provider);
    }

    /**
     * Adds an ontology generator to the chain of used generators.
     *
     * @param generator the generator to add.
     */
    public void addGenerator(OntologyGenerator generator) {
        generators.add(generator);
    }

    private List<Ontology> loadOntology() throws OntologyProviderException {
        LinkedList<Ontology> result = new LinkedList<Ontology>();
        for (OntologyProvider provider : providers) {
            result.addAll(provider.provide());
        }

        return result;
    }
}