package de.renew.ontology;

import de.renew.util.StringUtil;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;


/**
 * This class helps the entities to transform their information to different target types.
 *
 * @author Benjamin Teuber, Konstantin Simon Maria Moellers
 * @version 2015-01-16
 */
public class OntologyHelper {
    static private final List<String> NUMERIC_TYPES = Arrays.asList("double",
                    "byte", "int", "short", "char", "long", "float", "integer");
    static private final List<String> BOOLEAN_TYPES = Arrays.asList("boolean");
    static private final List<String> BASIC_TYPES = new ArrayList<String>();

    /**
     * Defines the stroke used in ontology models.
     */
    public static final String ONTOLOGY_MODEL_STROKE = "-";

    static {
        BASIC_TYPES.addAll(NUMERIC_TYPES);
        BASIC_TYPES.addAll(BOOLEAN_TYPES);
    }

    private static final String[] capaTypes = { "concept", "agent-action", "predicate", "agent-identifier" };

    /**
     * Converts a given ontology type to a corresponding Java Type.
     *
     * @param modelType type coming from an ontology model
     * @return a valid Java type
     */
    public static String convertToJavaType(final String modelType) {
        for (String basicType : BASIC_TYPES) {
            if (modelType.equals(basicType)) {
                return modelType;
            }
        }
        return convertToJavaClassName(modelType);
    }

    /**
     * Converts a given ontology type to a corresponding JavaScript Type.
     *
     * @param modelType type coming from an ontology model
     * @return a valid JavaScript type
     */
    public static String convertToJavaScriptType(final String modelType) {
        String javaType = convertToJavaType(modelType);
        String lowerCase = modelType.toLowerCase();
        if (NUMERIC_TYPES.contains(modelType)
                        || NUMERIC_TYPES.contains(lowerCase)) {
            return "Number";
        }

        if (BOOLEAN_TYPES.contains(modelType)
                        || BOOLEAN_TYPES.contains(lowerCase)) {
            return "Boolean";
        }

        return javaType;
    }

    /**
     * Get default value for Java data type
     *
     * @param modelType [String] The Java data type
     * @return Object The default value
     */
    public static String getDefaultValueForJavaType(final String modelType) {
        String lowerCase = modelType.toLowerCase();
        if (lowerCase.equals("byte") || lowerCase.equals("short")
                        || lowerCase.equals("int")
                        || lowerCase.equals("integer")) {
            return "0"; // "0";
        }
        if (lowerCase.equals("long")) {
            return "0"; // "0L";
        }
        if (lowerCase.equals("float")) {
            return "0"; // "0.0f";
        }
        if (lowerCase.equals("double")) {
            return "0"; // "0.0d";
        }
        if (lowerCase.equals("char")) {
            return "0"; // "'\u0000'";
        }
        if (lowerCase.equals("boolean")) {
            return "false"; // "false";
        }
        if (lowerCase.equals("date")) {
            return "null";
        }
        return "''";
    }

    /**
     * Converts a given ontology type to a corresponding JavaScript Type name.
     *
     * @param modelType type coming from an ontology model
     * @return a valid JavaScript type
     */
    public static String convertToJavaScriptTypeName(final String modelType) {
        String lowerCase = modelType.toLowerCase();
        if (NUMERIC_TYPES.contains(modelType)
                        || NUMERIC_TYPES.contains(lowerCase)) {
            return "number";
        }

        if (BOOLEAN_TYPES.contains(modelType)
                        || BOOLEAN_TYPES.contains(lowerCase)) {
            return "boolean";
        }

        if (lowerCase.equals("string")) {
            return "string";
        }

        return modelType;
    }

    /**
     * Converts a ontology model type to a Java accepted form.
     *
     * @param modelType type coming from an ontology model
     * @return a valid Java class name
     */
    public static String convertToJavaClassName(final String modelType) {
        return StringUtil.firstToUpperCase(convertToJavaName(modelType));
    }

    /**
     * Returns whether a model type is capa convenient.
     *
     * @param modelType type coming from an ontology model
     * @return <code>true</code>, if this is a CAPA type.
     * @deprecated should not be used. Refactor the FIPA plugin!
     */
    @Deprecated
    public static boolean isCapaType(final String modelType) {
        for (String capaType : capaTypes) {
            if (modelType.equalsIgnoreCase(capaType)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Converts a ontology model name to a Java accepted form.
     *
     * @param modelName name coming from an ontology model
     * @return Java convenient identifier name
     */
    public static String convertToJavaName(final String modelName) {
        return replaceStrokesByCamelCase(modelName.replace(' ', '_'));
    }

    /**
     * Replaces all stroke-strings by camelCaseStrings.
     *
     * @param string the string to transform
     * @return a camel cased string
     */
    private static String replaceStrokesByCamelCase(final String string) {
        StringBuilder builder = new StringBuilder(string);
        int index = builder.indexOf(ONTOLOGY_MODEL_STROKE);
        while (index >= 0) {
            if (index >= builder.length()) {
                builder.setCharAt(index, '_');
            } else if (!Character.isLetter(builder.charAt(index + 1))) {
                builder.setCharAt(index, '_');
            } else {
                builder.deleteCharAt(index);
                builder.setCharAt(index,
                                Character.toUpperCase(builder.charAt(index)));
            }

            // Find next occurrence
            index = builder.indexOf(ONTOLOGY_MODEL_STROKE, index);
        }

        return builder.toString();
    }
}