package de.renew.ontology.entities;

import de.renew.ontology.OntologyHelper;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;


/**
 * Entity representing a concept of an ontology.
 * <p>
 * Concepts can be represented as classes within a Java or ontology model context.
 * <p>
 * Concepts earn many features of Java classes, e.g. they have a super concept from which they
 * inherit all their features.
 * <p>
 * A concept have many Slots.
 *
 * @author Benjamin Teuber, Konstantin Simon Maria Moellers
 * @author Christian Bracker (1bracker)
 * @version 2015-11-29
 */
public class Concept {

    /**
     * The original ontology
     */
    private Ontology ontology;

    /**
     * The original name of this concept.
     */
    private String name;

    /**
     * The java name of this concept.
     */
    private String javaName;

    /**
     * The direct super of this concept.
     */
    private Concept superConcept;

    /**
     * Is this concept abstract?
     */
    private boolean isAbstract;

    /**
     * The javadoc documentation for this concept.
     */
    private String documentation;
    
    /**
     * The annotation for this concept.
     */
    private String annotation;

    /**
     * All (indirect) superconcepts, starting from
     * the most general one (e.g. Object)
     */
    private List<Concept> allSuperconcepts;

    /**
     * The direct slots of this class.
     */
    private List<Slot> directSlots;

    /**
     * The inherited slots of the class, ordered by slotname.
     */
    private List<Slot> inheritedSlots;

    /**
     * All slots of the concept and its superconcepts,
     * ordered by slotname.
     */
    private List<Slot> allSlots;

    /**
     * Creates a new concept. The superConcept of this concept will be set
     * as abstract as forced by current Capa implementation.
     *
     * @param ontology The {@link Ontology}
     * @param name Name of the concept - usually written like "concept-name"
     * @param superConcept Superconcept - null for no superConcept.
     * @param forceAbstract Shall the concept be abstract even if no subconcepts will
     *            be defined in this ontology? Useful when working with ontology references
     * @param documentation Javadoc documentation for the generated concept class
     */
    public Concept(Ontology ontology, String name, Concept superConcept,
                   boolean forceAbstract, String documentation) {
        this(ontology, name, superConcept, forceAbstract, documentation, null);
    }
    
    /**
     * Creates a new concept. The superConcept of this concept will be set
     * as abstract as forced by current Capa implementation.
     *
     * @param ontology The {@link Ontology}
     * @param name Name of the concept - usually written like "concept-name"
     * @param superConcept Superconcept - null for no superConcept.
     * @param forceAbstract Shall the concept be abstract even if no subconcepts will
     *            be defined in this ontology? Useful when working with ontology references
     * @param documentation Javadoc documentation for the generated concept class
     * @param annotation Java annotations for the generated concept class
     */
    public Concept(Ontology ontology, String name, Concept superConcept,
                   boolean forceAbstract, String documentation, String annotation) {
        this.ontology = ontology;
        this.name = name;
        this.javaName = OntologyHelper.convertToJavaClassName(name);
        this.superConcept = superConcept;
        this.isAbstract = forceAbstract;
        if (superConcept != null) {
            superConcept.setAbstract(); // Superconcepts are always abstract in CAPA
        }
        this.documentation = documentation;
        this.annotation = annotation;
        this.directSlots = new ArrayList<Slot>();
    }

    public void addSlot(Slot slot) {
        directSlots.add(slot);
    }

    public void recalc() {
        Collections.sort(directSlots);
        calcAllSuperconcepts();
        calcSlots();
    }

    public Ontology getOntology() {
        return ontology;
    }

    public String getName() {
        return name;
    }

    /**
     * All required slots of the concept and its superconcepts,
     * ordered by slotname.
     */
    private List<Slot> requiredSlots;

    public String getJavaName() {
        return javaName;
    }

    public String getConstName() {
        return javaName.toUpperCase();
    }

    public Concept getSuperconcept() {
        return superConcept;
    }

    public void setAbstract() {
        isAbstract = true;
    }

    public boolean getIsAbstract() {
        return isAbstract;
    }

    public String getDocumentation() {
        return documentation;
    }
    
    public String getAnnotation() {
        return annotation;
    }

    public List<Concept> getAllSuperconcepts() {
        return allSuperconcepts;
    }

    public List<Slot> getDirectSlots() {
        return directSlots;
    }

    public List<Slot> getInheritedSlots() {
        return inheritedSlots;
    }

    public List<Slot> getAllSlots() {
        return allSlots;
    }

    public List<Slot> getRequiredSlots() {
        return requiredSlots;
    }

    /**
     * Determines whether this concept is a common concept.
     *
     * @return <code>true</code> if is common; <code>false</code> otherwise.
     */
    public boolean getIsCommon() {
        return OntologyHelper.isCapaType(this.getName());
    }

    public boolean getIsSubconcept() {
        return superConcept != null
                        && !OntologyHelper.isCapaType(superConcept.getName());
    }

    public boolean getIsSubtypeOf(String name) {
        if (javaName.equals(name)) {
            return true;
        }
        if (allSuperconcepts != null) {
            for (Concept concept : allSuperconcepts) {
                if (concept.javaName.equals(name)) {
                    return true;
                }
            }
        }
        return false;
    }

    public boolean getIsPredicate() {
        return getIsSubtypeOf("Predicate");
    }

    public boolean getIsAgentAction() {
        return getIsSubtypeOf("AgentAction");
    }

    /**
     * Gets the short package name of this concept.
     *
     * @return the short package name.
     */
    public String getShortPackage() {
        return this.getOntology().getShortPackage();
    }

    public String getSuperclass() {
        if (getIsSubconcept()) {
            return superConcept.javaName;
        } else if (getIsPredicate()) {
            return "GenericVT";
        } else {
            return "GenericKVT";
        }
    }

    public String parameterList(List<Slot> slots) {
        StringBuffer buffer = new StringBuffer();
        Iterator<Slot> iter = slots.iterator();
        while (iter.hasNext()) {
            buffer.append(iter.next().getVarDecl());
            if (iter.hasNext()) {
                buffer.append(", ");
            }
        }
        return buffer.toString();
    }

    public String nullableParameterList(List<Slot> slots) {
        StringBuffer buffer = new StringBuffer();
        Iterator<Slot> iter = slots.iterator();
        while (iter.hasNext()) {
            buffer.append(iter.next().getNullableVarDecl());
            if (iter.hasNext()) {
                buffer.append(", ");
            }
        }
        return buffer.toString();
    }

    /**
     * A contructor that takes all required slots as arguments will be
     * generated if at least one, but not all slots are optional
     *
     * @return <code>true</code>, if one of these slots is required
     */
    public boolean getHasRequiredSlotConstructor() {
        boolean hasOptionalSlot = false;
        boolean hasRequiredSlot = false;

        for (Slot slot : allSlots) {
            if (slot.getIsRequired()) {
                hasRequiredSlot = true;
            } else {
                hasOptionalSlot = true;
            }
        }

        return hasOptionalSlot && hasRequiredSlot;
    }

    /**
     * A contructor that takes all slots as arguments will be
     * generated if at least one slot exists at all.
     *
     * @return <code>true</code>, if not all slots are empty
     */
    public boolean getHasAllSlotConstructor() {
        return !allSlots.isEmpty();
    }

    /**
     * Determines whether this concept has a non-empty documentation.
     *
     * @return <code>true</code> if has documentation; <code>false</code> otherwise.
     */
    public boolean getHasDocumentation() {
        String documentation = this.getDocumentation();
        return (documentation != null) && (documentation.equals("") == false);
    }

    private void calcAllSuperconcepts() {
        allSuperconcepts = new ArrayList<Concept>();
        Concept act = this;
        while ((act = act.superConcept) != null) {
            allSuperconcepts.add(0, act);
        }
    }

    private void calcSlots() {
        inheritedSlots = new ArrayList<Slot>();
        for (Concept superconcept : allSuperconcepts) {
            for (Slot slot : superconcept.directSlots) {
                inheritedSlots.add(slot);
            }
        }
        Collections.sort(inheritedSlots);

        allSlots = new ArrayList<Slot>();
        allSlots.addAll(inheritedSlots);
        allSlots.addAll(directSlots);
        Collections.sort(allSlots);

        requiredSlots = new ArrayList<Slot>();
        for (Slot slot : allSlots) {
            if (slot.getIsRequired()) {
                requiredSlots.add(slot);
            }
        }
    }
}