package de.renew.ontology.entities;

import static de.renew.ontology.OntologyHelper.convertToJavaClassName;

import de.renew.ontology.OntologyHelper;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;


/**
 * Entity representing an ontology.
 * <p>
 * An ontology consists of Concepts.
 *
 * @author Benjamin Teuber, Konstantin Simon Maria Moellers
 * @author Christian Bracker (1bracker)
 * @version 2015-11-29
 */
public class Ontology {

    /**
     * The name of this ontology.
     */
    private final String name;

    /**
     * The Java class name that shall be given to this ontology.
     */
    private final String className;

    /**
     * The package to which this ontology belongs.
     */
    private final String ontologyPackage;

    /**
     * This map associates concept names with their representational objects.
     */
    private final Map<String, Concept> conceptMap = new HashMap<String, Concept>();

    /**
     * Handles all ontology objects being referenced by this one.
     */
    private final List<Ontology> referencedOntologies = new ArrayList<Ontology>();

    /**
     * Handles all imports that have to be generated.
     */
    private final Set<String> imports = new HashSet<String>();

    public Ontology(String name, String ontologyPackage) {
        this.name = name;
        this.className = convertToJavaClassName(name) + "Ontology";
        this.ontologyPackage = ontologyPackage;
    }

    /** {@inheritDoc} */
    @Override
    public boolean equals(Object obj) {
        if (obj instanceof Ontology) {
            Ontology otherOntology = (Ontology) obj;
            return this.isEqualStrings(this.getName(), otherOntology.getName())
                            && this.isEqualStrings(this.getPackage(),
                                            otherOntology.getPackage());
        } else {
            return false;
        }
    }

    public void addImport(String pkg) {
        imports.add(pkg);
    }

    public void addReferencedOntology(Ontology ontology) {
        referencedOntologies.add(ontology);
        if (!ontology.getPackage().equals(this.getPackage())) {
            addImport(ontology.getPackage() + ".*");
        }
    }

    public Concept findConcept(String conceptName) {
        return conceptMap.get(conceptName);
    }

    public void addConcept(Concept concept) {
        conceptMap.put(concept.getName(), concept);
    }

    public String getName() {
        return name;
    }

    public String getJavaName() {
        return className;
    }

    public String getPackage() {
        return ontologyPackage;
    }

    /**
     * Gets the class name of this ontology in PascalCase.
     *
     * @return the class name.
     */
    public String getClassName() {
        return convertToJavaClassName(this.getName());
    }

    /**
     * Gets the short name of this ontology, i.e. without <i>Ontology</i> or
     * <i>-ontology</i> suffix.
     *
     * @return the short name.
     */
    public String getShortName() {
        return this.getName().replaceFirst("(Ontology|-ontology)$", "");
    }

    /**
     * Gets the short class name of this ontology in PascalCase.
     *
     * @return the short class name.
     */
    public String getShortClassName() {
        return convertToJavaClassName(this.getShortName());
    }

    /**
     * Gets the short package of this ontology.
     *
     * @return the short package.
     */
    public String getShortPackage() {
        return this.getShortName().toLowerCase();
    }

    public Collection<Concept> getConcepts() {
        Collection<Concept> concepts = new ArrayList<Concept>();
        for (Concept concept : conceptMap.values()) {
            if (!OntologyHelper.isCapaType(concept.getName())) {
                concepts.add(concept);
            }
        }
        return concepts;
    }

    /**
     * @return list of concepts, ordered by internal dependencies
     */
    public Collection<Concept> getDependencyOrderedConcepts() {
        LinkedList<Concept> concepts = new LinkedList<Concept>();
        for (Concept concept : conceptMap.values()) {
            if (!OntologyHelper.isCapaType(concept.getName())) {
                concepts.add(concept);
            }
        }
        LinkedList<Concept> ordered = new LinkedList<Concept>();

        while (!concepts.isEmpty()) {
            Concept concept = concepts.remove();
            if (concept.getIsSubconcept()) {
                if (ordered.contains(concept.getSuperconcept())
                                || concept.getOntology() != concept
                                                .getSuperconcept()
                                                .getOntology()) {
                    ordered.addLast(concept);
                } else {
                    concepts.addLast(concept);
                }
            } else {
                ordered.addFirst(concept);
            }
        }

        return ordered;
    }

    public String getConceptClassList() {
        StringBuffer buffer = new StringBuffer();
        Iterator<Concept> iter = getConcepts().iterator();
        while (iter.hasNext()) {
            buffer.append(iter.next().getJavaName());
            buffer.append(".class");
            if (iter.hasNext()) {
                buffer.append(",\n");
            }
        }
        return buffer.toString();
    }

    public List<Ontology> getReferencedOntologies() {
        return referencedOntologies;
    }

    public boolean hasReferencedOntology(String ontologyName) {
        for (Ontology onto : referencedOntologies) {
            if (onto.getName().equals(ontologyName)) {
                return true;
            }
        }
        return false;
    }

    public Set<String> getImports() {
        return imports;
    }

    /**
     * Get the output file for a given concept in the ontology-package.
     *
     * @param basePath
     * @param conceptName concept name to look for
     * @param extension file extension that should be given to the output file
     * @param fileNamePrefix
     * @param fileNameSuffix
     * @return a file object for the output file to be generated
     */
    public File getOutputFile(File basePath, String conceptName,
                              String extension, String fileNamePrefix,
                              String fileNameSuffix) {
        // Determine output directory.
        String relativePath = this.getPackage();

        relativePath = relativePath.replace('.', File.separator.charAt(0));

        File dir = new File(basePath, relativePath);

        String prefix = fileNamePrefix != null ? fileNamePrefix : "";
        String suffix = fileNameSuffix != null ? fileNameSuffix : "";

        // Return corresponding java file for ontology.
        return new File(dir, prefix + conceptName + suffix + "." + extension);
    }

    /**
     * Get the output file for the ontology-listing class.
     *
     * @param basePath
     * @param extension file extension that should be given to the output file
     * @param fileNamePrefix
     * @param fileNameSuffix
     * @return a file object for the output file to be generated
     */
    public File getOntologyListingOutputFile(File basePath, String extension,
                                             String fileNamePrefix,
                                             String fileNameSuffix) {
        return getOutputFile(basePath, this.getJavaName(), extension,
                        fileNamePrefix, fileNameSuffix);
    }

    /**
     * Compare the given string for equality.
     *
     * @param str1 The first string to test. (may be null)
     * @param str2 The second string to test. (may be null)
     * @return <code>true</code> if given strings are equal; <code>false</code> otherwise.
     */
    private boolean isEqualStrings(String str1, String str2) {
        if (str1 == null) {
            return (str2 == null);
        } else {
            return str1.equals(str2);
        }
    }
}