package de.renew.ontology.entities;

import de.renew.ontology.OntologyHelper;
import de.renew.util.StringUtil;


/**
 * Entity representing a slot of a concept.
 * <p>
 * Slots are accessory properties of a concept. They have a specific type that cannot be changed at
 * runtime.
 *
 * @author Benjamin Teuber, Konstantin Simon Maria Moellers, 3emuelle
 */
public class Slot implements Comparable<Slot> {

    /**
     * Name of the slot.
     */
    private final String name;

    /**
     * Name used as in Java.
     */
    private final String javaName;

    /**
     * Type of this slot, e.g. a class or primitive type.
     */
    private final String type;

    /**
     * Default value of slot primitiv or class
     */
    private final String defaultValueForJavaType;

    /**
     * Type usable in Java context.
     */
    private final String javaType;

    /**
     * Whether this slot is some kind of collection.
     */
    private final boolean isMultiple;

    /**
     * Whether this slot is required for its concept.
     */
    private final boolean isRequired;

    /**
     * Documentation of this slot.
     */
    private final String documentation;

    /**
     * Annotation of this slot.
     */
    private final String annotation;
    
    public Slot(String name, String type, String documentation, String annotation,
                boolean isMultiple, boolean isRequired) {
        this.name = name;
        this.javaName = OntologyHelper.convertToJavaName(name);
        this.type = type;
        this.javaType = OntologyHelper.convertToJavaType(type);
        this.defaultValueForJavaType = OntologyHelper
                        .getDefaultValueForJavaType(type);
        this.documentation = documentation;
        this.annotation = annotation;
        this.isMultiple = isMultiple;
        this.isRequired = isRequired;
        checkIsBoxed();
    }

    public String getName() {
        return name;
    }

    public String getJavaName() {
        return javaName;
    }

    public String getVarName() {
        char[] chars = this.getJavaName().toCharArray();
        chars[0] = Character.toLowerCase(chars[0]);
        return new String(chars);
    }

    public String getConstName() {
        return "SLOT_" + javaName.toUpperCase();
    }

    public String getMethodSuffix() {
        return StringUtil.firstToUpperCase(javaName);
    }

    public String getType() {
        return type;
    }

    public String getJavaType() {
        return javaType;
    }

    public String getDefaultValueForJavaType() {
        return defaultValueForJavaType;
    }

    public String getJavaScriptType() {
        return OntologyHelper.convertToJavaScriptType(type);
    }

    public String getJavaScriptTypeName() {
        return OntologyHelper.convertToJavaScriptTypeName(type);
    }

    public String getImplType() {
        if (isMultiple) {
            return "VTSequence";
        } else {
            return javaType;
        }
    }

    public String getNullableImplType() {
        if (isMultiple) {
            return "VTSequence";
        } else {
            return boxedType;
        }
    }

    public String getVarDecl() {
        return getImplType() + " " + getVarName();
    }

    public String getNullableVarDecl() {
        return getNullableImplType() + " " + getVarName();
    }

    private String boxedType;

    public String getBoxedType() {
        return boxedType;
    }

    public boolean getNeedsBoxing() {
        return !boxedType.equals(javaType);
    }

    public boolean getIsMultiple() {
        return isMultiple;
    }

    public boolean getIsRequired() {
        return isRequired;
    }

    public String getDocumentation() {
        return documentation;
    }
    
    public String getAnnotation() {
        return annotation;
    }

    /**
     * Checks whether this is a boxed Java type.
     */
    private void checkIsBoxed() {
        if (javaType.equals("float") || javaType.equals("boolean")
                        || javaType.equals("byte")) {
            boxedType = StringUtil.firstToUpperCase(javaType);
        } else if (javaType.equals("int")) {
            boxedType = "Integer";
        } else {
            boxedType = javaType;
        }
    }

    /**
     * Compares two slots by their names.
     */
    @Override
    public int compareTo(Slot sl) {
        return this.getName().compareTo(sl.getName());
    }
}