package de.renew.ontology.entities;


/**
 * Nested POJO element to configure ontology templates being used during generation.
 * <p>
 * Parameters:
 * <p>
 * <code>extension</code>: The file extension which should be given
 * to files being generated by this template.
 * <p>
 * <code>filename</code>: The path specifying the location of the
 * Velocity template being used while
 * generating process.
 *
 * @author Konstantin Simon Maria Moellers
 * @version 2015-01-13
 */
public class TemplateConfiguration {
    private String extension;
    private String filename;
    private String fileNamePrefix;
    private String fileNameSuffix;

    public TemplateConfiguration() {
        this.extension = "java";
    }

    public TemplateConfiguration(String extension, String filename) {
        this();
        this.extension = extension;
        this.filename = filename;
    }

    public TemplateConfiguration(String extension, String filename,
                                 String fileNamePrefix, String fileNameSuffix) {
        this(extension, filename);
        this.fileNamePrefix = fileNamePrefix;
        this.fileNameSuffix = fileNameSuffix;
    }

    /**
     * The file extension which should be given
     * to files being generated by this template.
     *
     * @param extension file extension
     */
    public void setExtension(String extension) {
        this.extension = extension;
    }

    /**
     * Returns the file extension which should be given to files being
     * generated by this template.
     *
     * @return file extension string
     */
    public String getExtension() {
        return extension;
    }

    /**
     * The path specifying the location of the
     * Velocity template being used while
     * generating process.
     *
     * @param filename resource file path
     */
    public void setFilename(String filename) {
        this.filename = filename;
    }

    /**
     * Returns the path specifying the location of the Velocity template.
     *
     * @return Velocity template path string
     */
    public String getFilename() {
        return filename;
    }

    /**
     * Returns the prefix for the file name of the output file.
     *
     * @return File name prefix string
     */
    public String getFileNamePrefix() {
        return fileNamePrefix;
    }

    /**
     * Sets prefix for the file name of the output file.
     * @param fileNamePrefix the file name prefix string.
     */
    public void setFileNamePrefix(String fileNamePrefix) {
        this.fileNamePrefix = fileNamePrefix;
    }


    /**
     * Returns the suffix for the file name of the output file.
     *
     * @return File name suffix string
     */
    public String getFileNameSuffix() {
        return fileNameSuffix;
    }

    /**
     * Sets suffix for the file name of the output file.
     * @param fileNameSuffix the file name suffix string.
     */
    public void setFileNameSuffix(String fileNameSuffix) {
        this.fileNameSuffix = fileNameSuffix;
    }


    /**
     * Loads this configuration from a property string.
     *
     * @param propertyString the property string to load from.
     * @return self
     */
    public TemplateConfiguration setPropertyString(String propertyString) {
        String[] configs = propertyString.split(":");

        if (configs.length > 1) {
            this.setExtension(configs[0]);
            this.setFilename(configs[1]);
        } else { // java is default extension
            this.setExtension("java");
            this.setFilename(configs[0]);
        }

        return this;
    }

    /**
     * Returns the property string representing this configuration.
     *
     * @return configuration property string
     */
    public String getPropertyString() {
        return this.getExtension() + ":" + this.getFilename();
    }
}