package de.renew.ontology.generators;

import de.renew.ontology.entities.Concept;
import de.renew.ontology.entities.Ontology;
import de.renew.ontology.entities.TemplateConfiguration;
import de.renew.ontology.exception.OntologyGeneratorException;

import java.io.File;
import java.util.Collection;
import java.util.List;
import java.util.Set;

import org.apache.velocity.app.VelocityEngine;


/**
 * Ontology generator for AMD-compliant JavaScript files.
 *
 * @author Christian Bracker (1bracker)
 * @version 2015-11-29
 */
public class JavaScriptAmdOntologyGenerator extends VelocityOntologyGenerator {
    public static final String BASE_DIR = "de/renew/ontology/js";

    /**
     * Creates a new JavaScriptAmdOntologyGenerator using default templates.
     */
    public JavaScriptAmdOntologyGenerator() {
        super(new TemplateConfiguration("js", BASE_DIR + "/Ontology.amd.js.vm"),
                        new TemplateConfiguration("js",
                                        BASE_DIR + "/Concept.amd.js.vm"),
                        true);
    }

    /** {@inheritDoc} */
    @Override
    public String getGeneratorName() {
        return "java-script-amd";
    }

    /**
     * Renders the ontology classes.
     * <p>
     * Sets the ontology's java name to be equal to the ontology name avoiding
     * "OntologyOntology" suffix and defines different output file path.
     * </p>
     *
     * @param ontology Ontology to be rendered.
     * @param engine Engine used for rendering.
     * @throws OntologyGeneratorException
     */
    @Override
    protected void renderOntologyClasses(Ontology ontology,
                                         VelocityEngine engine)
                    throws OntologyGeneratorException {
        super.renderOntologyClasses(new JavaScriptOntology(ontology), engine);
    }

    /**
     * Renders the ontology listings.
     * <p>
     * Sets the ontology's java name to be equal to the ontology name avoiding
     * "OntologyOntology" suffix and defines different output file path.
     * </p>
     *
     * @param ontology Ontology to be rendered.
     * @param engine Engine used for rendering.
     * @throws OntologyGeneratorException
     */
    @Override
    protected void renderOntologyListings(Ontology ontology,
                                          VelocityEngine engine)
                    throws OntologyGeneratorException {
        super.renderOntologyListings(new JavaScriptOntology(ontology), engine);
    }

    /**
     * Ontology decorator for java-script specific output.
     *
     * @author Christian Bracker (1bracker)
     * @version 2015-11-29
     */
    private static class JavaScriptOntology extends Ontology {

        /**
         * The ontology being decorated.
         */
        private Ontology innerOntology;

        /**
         * Creates a new JavaScriptOntology decorating the given inner ontology.
         *
         * @param innerOntology The ontology to decorate.
         */
        public JavaScriptOntology(Ontology innerOntology) {
            super(innerOntology.getName(), innerOntology.getPackage());
            this.innerOntology = innerOntology;
        }

        /**
         * Gets the inner ontology of this ontology.
         *
         * @return the inner ontology.
         */
        public Ontology getInnerOntology() {
            return this.innerOntology;
        }

        /** {@inheritDoc} */
        @Override
        public void addImport(String pkg) {
            this.getInnerOntology().addImport(pkg);
        }

        /** {@inheritDoc} */
        @Override
        public void addReferencedOntology(Ontology ontology) {
            this.getInnerOntology().addReferencedOntology(ontology);
        }

        /** {@inheritDoc} */
        @Override
        public Concept findConcept(String conceptName) {
            return this.getInnerOntology().findConcept(conceptName);
        }

        /** {@inheritDoc} */
        @Override
        public void addConcept(Concept concept) {
            this.getInnerOntology().addConcept(concept);
        }

        /** {@inheritDoc} */
        @Override
        public String getName() {
            return this.getInnerOntology().getName();
        }

        /** {@inheritDoc} */
        @Override
        public String getJavaName() {
            return this.getInnerOntology().getClassName();
        }

        /** {@inheritDoc} */
        @Override
        public String getPackage() {
            return this.getInnerOntology().getPackage();
        }

        /** {@inheritDoc} */
        @Override
        public String getShortName() {
            return this.getInnerOntology().getShortName();
        }

        /** {@inheritDoc} */
        @Override
        public String getShortClassName() {
            return this.getInnerOntology().getShortClassName();
        }

        /** {@inheritDoc} */
        @Override
        public String getShortPackage() {
            return this.getInnerOntology().getShortPackage();
        }

        /** {@inheritDoc} */
        @Override
        public Collection<Concept> getConcepts() {
            return this.getInnerOntology().getConcepts();
        }

        /** {@inheritDoc} */
        @Override
        public Collection<Concept> getDependencyOrderedConcepts() {
            return this.getInnerOntology().getDependencyOrderedConcepts();
        }

        /** {@inheritDoc} */
        @Override
        public String getConceptClassList() {
            return this.getInnerOntology().getConceptClassList();
        }

        /** {@inheritDoc} */
        @Override
        public List<Ontology> getReferencedOntologies() {
            return this.getInnerOntology().getReferencedOntologies();
        }

        /** {@inheritDoc} */
        @Override
        public boolean hasReferencedOntology(String ontologyName) {
            return this.getInnerOntology().hasReferencedOntology(ontologyName);
        }

        /** {@inheritDoc} */
        @Override
        public Set<String> getImports() {
            return this.getInnerOntology().getImports();
        }

        /** {@inheritDoc} */
        @Override
        public File getOutputFile(File basePath, String conceptName,
                                  String extension, String fileNamePrefix,
                                  String fileNameSuffix) {
            File dir = new File(basePath, this.getShortPackage());

            String prefix = fileNamePrefix != null ? fileNamePrefix : "";
            String suffix = fileNameSuffix != null ? fileNameSuffix : "";

            return new File(dir,
                            prefix + conceptName + suffix + "." + extension);
        }

        /** {@inheritDoc} */
        @Override
        public File getOntologyListingOutputFile(File basePath,
                                                 String extension,
                                                 String fileNamePrefix,
                                                 String fileNameSuffix) {
            String prefix = fileNamePrefix != null ? fileNamePrefix : "";
            String suffix = fileNameSuffix != null ? fileNameSuffix : "";

            return new File(basePath, prefix + this.getClassName() + suffix
                            + "." + extension);
        }
    }
}