package de.renew.ontology.generators;

import de.renew.ontology.OntologyGenerator;
import de.renew.ontology.entities.Concept;
import de.renew.ontology.entities.Ontology;
import de.renew.ontology.entities.TemplateConfiguration;
import de.renew.ontology.exception.OntologyGeneratorException;

import de.renew.plugin.PluginProperties;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.HashSet;
import java.util.Set;

import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.VelocityEngine;
import org.apache.velocity.tools.generic.DateTool;


/**
 * Generator to render files from ontology models.
 *
 * @author Konstantin Simon Maria Moellers
 * @author Christian Bracker (1bracker)
 * @version 2015-11-29
 */
public abstract class VelocityOntologyGenerator implements OntologyGenerator {
    private final TemplateConfiguration listingConfiguration;
    private final TemplateConfiguration classConfiguration;
    private boolean generatingReferencedOntologies;
    private File destinationDirectory;

    public VelocityOntologyGenerator(TemplateConfiguration listingConfiguration,
                                     TemplateConfiguration classConfiguration) {
        this(listingConfiguration, classConfiguration, false);
    }

    public VelocityOntologyGenerator(TemplateConfiguration listingConfiguration,
                                     TemplateConfiguration classConfiguration,
                                     boolean generatingReferencedOntologies) {
        this.listingConfiguration = listingConfiguration;
        this.classConfiguration = classConfiguration;
        this.generatingReferencedOntologies = generatingReferencedOntologies;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void generate(Ontology ontology) throws OntologyGeneratorException {
        Set<Ontology> ontologiesToGenerate = new HashSet<Ontology>();
        ontologiesToGenerate.add(ontology);

        if (this.isGeneratingReferencedOntologies()) {
            this.addReferencedOntologies(ontologiesToGenerate, ontology);
        }

        for (Ontology currentOntology : ontologiesToGenerate) {
            this.generateWithoutReferencedOntologies(currentOntology);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setDestinationDirectory(File destinationDirectory) {
        this.destinationDirectory = destinationDirectory;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public File getDestinationDirectory() {
        return destinationDirectory;
    }

    /**
     * Get whether this generator also generates all referenced ontologies.
     *
     * @return <code>true</code> if generating referenced ontologies; <code>false</code> otherwise.
     */
    public boolean isGeneratingReferencedOntologies() {
        return this.generatingReferencedOntologies;
    }

    /**
     * Set whether this generator also generates all referenced ontologies.
     *
     * @param generatingReferencedOntologies
     *     <code>true</code> if generator should generate referenced ontologies;
     *     <code>false</code> otherwise.
     */
    public void setGeneratingReferencedOntologies(boolean generatingReferencedOntologies) {
        this.generatingReferencedOntologies = generatingReferencedOntologies;
    }

    /**
     * Loads a velocity template engine.
     *
     * @return the engine loaded.
     * @throws OntologyGeneratorException
     */
    private VelocityEngine createVelocityEngine()
                    throws OntologyGeneratorException {
        // Initialize Velocity.
        VelocityEngine engine = new VelocityEngine();
        engine.setProperty("resource.loader", "file, class");
        engine.setProperty("class.resource.loader.description",
                        "Velocity Classpath Resource Loader");
        engine.setProperty("class.resource.loader.class",
                        "org.apache.velocity.runtime.resource.loader.ClasspathResourceLoader");
        try {
            engine.init();
        } catch (Exception e) {
            throw new OntologyGeneratorException(
                            "Could not initialize Velocity", e);
        }

        return engine;
    }

    /**
     * Generates the given ontology without any referenced ontologies.
     *
     * @param ontology The
     * @throws OntologyGeneratorException
     */
    protected void generateWithoutReferencedOntologies(Ontology ontology)
                    throws OntologyGeneratorException {
        if (!destinationDirectory.exists()) {
            if (!destinationDirectory.mkdirs()) {
                throw new OntologyGeneratorException(
                                "Could not create destination directory");
            }
        }

        if (!destinationDirectory.isDirectory()) {
            throw new OntologyGeneratorException(
                            "Destination must be a directory");
        }

        // Loads the velocity engine.
        VelocityEngine engine = createVelocityEngine();

        // Generate ontology listing files.
        renderOntologyListings(ontology, engine);

        // Generate class files.
        renderOntologyClasses(ontology, engine);
    }

    /**
     * Renders ontology listings.
     *
     * @param ontology ontology which will be rendered.
     * @param engine engine which renders.
     * @throws de.renew.ontology.exception.OntologyGeneratorException
     */
    protected void renderOntologyListings(Ontology ontology,
                                          VelocityEngine engine)
                    throws OntologyGeneratorException {
        // Load ontology into context.
        VelocityContext context = new VelocityContext();
        context.put("ontology", ontology);

        // Read listing configuration.
        String templateFilename = listingConfiguration.getFilename();
        String templateExt = listingConfiguration.getExtension();
        String fileNamePrefix = listingConfiguration.getFileNamePrefix();
        String fileNameSuffix = listingConfiguration.getFileNameSuffix();

        try {
            Template template = engine.getTemplate(templateFilename);

            File file = ontology.getOntologyListingOutputFile(
                            this.getDestinationDirectory(), templateExt,
                            fileNamePrefix, fileNameSuffix);
            renderTemplate(template, context, file);
        } catch (Exception e) {
            throw new OntologyGeneratorException(
                            "Could not generate Ontology listing", e);
        }
    }

    /**
     * Renders ontology listings.
     *
     * @param ontology ontology which will be rendered.
     * @param engine engine which renders.
     * @throws de.renew.ontology.exception.OntologyGeneratorException
     */
    protected void renderOntologyClasses(Ontology ontology,
                                         VelocityEngine engine)
                    throws OntologyGeneratorException {
        // Load ontology into context.
        VelocityContext context = new VelocityContext();
        context.put("ontology", ontology);

        // Read class configuration.
        String templateFilename = classConfiguration.getFilename();
        String templateExt = classConfiguration.getExtension();
        String fileNamePrefix = classConfiguration.getFileNamePrefix();
        String fileNameSuffix = classConfiguration.getFileNameSuffix();

        try {
            Template template = engine.getTemplate(templateFilename);

            // Iterate over all ontology concepts.
            for (Concept concept : ontology.getConcepts()) {
                // Load concept into context.
                concept.recalc();
                context.put("concept", concept);

                File file = ontology.getOutputFile(
                                this.getDestinationDirectory(),
                                concept.getJavaName(), templateExt,
                                fileNamePrefix, fileNameSuffix);

                renderTemplate(template, context, file);
            }
        } catch (Exception e) {
            throw new OntologyGeneratorException(
                            "Could not generate Ontology classes", e);
        }
    }

    /**
     * Add all referenced ontologies of the given ontology to the given set of ontologies
     *
     * @param ontologies The set of ontologies to modify.
     * @param ontologyToProcess The ontology whose references to add.
     */
    private void addReferencedOntologies(Set<Ontology> ontologies,
                                         Ontology ontologyToProcess) {
        for (Ontology reference : ontologyToProcess.getReferencedOntologies()) {
            if (ontologies.contains(reference) == false) {
                ontologies.add(reference);
                this.addReferencedOntologies(ontologies, reference);
            }
        }
    }

    /**
     * Renders a velocity template.
     */
    private void renderTemplate(Template template, VelocityContext context,
                                File outputFile)
                    throws IOException, OntologyGeneratorException {
        // Make sure directory exists
        File targetDir = outputFile.getParentFile();
        if (!targetDir.exists()) {
            targetDir.mkdirs();
        }

        FileWriter fw = new FileWriter(outputFile);
        try {
            context.put("generator", this);
            context.put("date", new DateTool());

            context.put("author", getAuthor());
            template.merge(context, fw);
        } catch (Exception e) {
            throw new OntologyGeneratorException(
                            "Could not render template with Velocity", e);
        } finally {
            fw.close();
        }
    }

    private String getAuthor() {
        PluginProperties userprops = PluginProperties.getUserProperties();
        if (userprops != null && userprops.containsKey("author.name")) {
            return userprops.getProperty("author.name");
        }
        return System.getProperty("user.name");
    }
}