package de.renew.ontology.providers;

import de.renew.formalism.fs.ShadowAssoc;
import de.renew.formalism.fs.ShadowConcept;
import de.renew.formalism.fs.ShadowIsa;
import de.renew.net.NetNotFoundException;
import de.renew.ontology.OntologyProvider;
import de.renew.ontology.entities.Concept;
import de.renew.ontology.entities.Ontology;
import de.renew.ontology.entities.Slot;
import de.renew.ontology.exception.OntologyProviderException;
import de.renew.shadow.DefaultShadowNetLoader;
import de.renew.shadow.ShadowDeclarationNode;
import de.renew.shadow.ShadowNet;
import de.renew.shadow.ShadowNetElement;
import de.renew.shadow.ShadowNetLoader;
import de.renew.shadow.ShadowNetSystem;
import de.renew.util.StringUtil;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;


/**
 * @author Benjamin Teuber, Konstantin Simon Maria Moellers
 * @version 2015-01-13
 */
public class FSOntologyProvider implements OntologyProvider {
    public static final String PACKAGE_STATEMENT = "package ";
    public static final String ACCESS_STATEMENT = "access ";
    public static final String IMPORT_STATEMENT = "import ";
    public static final String ABSTRACT_STEREOTYPE = "<<abstract>>";

    /**
     * The used loader for shadow nets.
     */
    private ShadowNetLoader netLoader = new DefaultShadowNetLoader(null);

    /**
     * Source files of the provisioning.
     */
    private Collection<File> sourceFiles;

    /**
     * Paths to search for nets
     */
    private Collection<String> netPath;

    /**
     * Map caching all known ontology models.
     */
    private final Map<File, Ontology> ontologyMap = new HashMap<File, Ontology>();

    /**
     * Finds ontology models by a net name.
     *
     * @return ontology model representing the net
     */
    @Override
    public List<Ontology> provide() throws OntologyProviderException {
        List<Ontology> result = new ArrayList<Ontology>();
        for (File file : this.sourceFiles) {
            result.add(provide(file));
        }
        return result;
    }

    private Ontology provide(File sourceFile) throws OntologyProviderException {
        Ontology ontology = ontologyMap.get(sourceFile);

        // Ontology not registered?
        if (ontology == null) {
            try {
                ShadowNetSystem system = netLoader
                                .loadShadowNetSystem(sourceFile.getName()
                                                .replaceFirst("[.][^.]+$", ""));
                ontology = findOntologyFromSns(system);

                // Register found ontology
                ontologyMap.put(sourceFile, ontology);
            } catch (NetNotFoundException e) {
                throw new OntologyProviderException(e);
            }
        }

        return ontology;
    }

    @Override
    public String getProviderName() {
        return "feature-structure";
    }

    @Override
    public Collection<File> getSourceFiles() {
        return sourceFiles;
    }

    @Override
    public void setSourceFiles(Collection<File> sourceFiles) {
        this.sourceFiles = sourceFiles;
    }

    @Override
    public void setNetPath(Collection<String> netPath) {
        this.netPath = netPath;
        setupShadowNetLoader();
    }

    private void setNetLoader(ShadowNetLoader shadowNetLoader) {
        this.netLoader = shadowNetLoader;
    }

    private Ontology findOntologyFromSns(ShadowNetSystem system)
                    throws OntologyProviderException {
        // de.renew.gui.ShadowTranslator (erzeugt SNSe)
        // de.renew.ant.CreateShadownetsTask (steuert ShadowTranslator)
        // de.renew.shadow.SNSFinder (liest SNSe)
        ShadowNet net = system.elements().iterator().next();

        String ontologyName = net.getName();

        ShadowDeclarationNode declarationNode = findDeclarationNode(net);
        if (declarationNode == null) {
            throw new RuntimeException(
                            "Could not find declaration node. It is needed to define the package.");
        }

        // Find description lines in declaration node.
        String[] declarations = declarationNode.inscr.split(";");
        if (declarations.length == 0) {
            throw new RuntimeException(
                            "There were no statements in the declaration node.");
        }

        // Find the package name and create ontology model.
        String ontologyPackage = findOntologyPackage(declarations);
        Ontology ontology = new Ontology(ontologyName, ontologyPackage);

        // Import ontology models referenced by access declarations
        // and packages in import declarations.
        for (int i = 1; i < declarations.length; i++) {
            String line = declarations[i].trim();
            if (line.startsWith(ACCESS_STATEMENT)) {
                String refOntoName = line.substring(ACCESS_STATEMENT.length())
                                .trim();
                if (!ontology.hasReferencedOntology(refOntoName)) {
                    // Recursively find referenced ontology.
                    Ontology referencedOntology = provide(
                                    new File(refOntoName));
                    ontology.addReferencedOntology(referencedOntology);
                }
            } else if (line.startsWith(IMPORT_STATEMENT)) {
                String pkg = line.substring(IMPORT_STATEMENT.length()).trim();
                ontology.addImport(pkg);
            }
        }

        Map<ShadowConcept, ShadowConcept> superConceptMap = new HashMap<ShadowConcept, ShadowConcept>();

        // Parse ???is-a??? links.
        for (Object o : net.elements()) {
            ShadowNetElement elem = (ShadowNetElement) o;
            if (elem instanceof ShadowIsa) {
                ShadowIsa shadowIsa = (ShadowIsa) elem;
                ShadowConcept source = shadowIsa.getSource();
                ShadowConcept target = shadowIsa.getTarget();
                if (!shadowIsa.isDisjunctive()) {
                    throw new RuntimeException(
                                    "Found a non-disjunctive is-a link: "
                                                    + source.getName() + " --> "
                                                    + target.getName()
                                                    + ". Only disjunctive is-a links are allowed in Ontologies.");
                }
                if (superConceptMap.containsKey(source)) {
                    throw new RuntimeException(
                                    "Found a concept with multiple superconcepts: "
                                                    + source.getName() + ".");
                }
                superConceptMap.put(source, target);
            }
        }

        // Import concepts.
        for (Object o : net.elements()) {
            ShadowNetElement elem = (ShadowNetElement) o;
            if (elem instanceof ShadowConcept) {
                findConcept(ontology, superConceptMap, (ShadowConcept) elem);
            }
        }

        // Import association slots.
        for (Object o : net.elements()) {
            ShadowNetElement elem = (ShadowNetElement) o;
            if (elem instanceof ShadowAssoc) {
                ShadowAssoc shadowAssoc = (ShadowAssoc) elem;
                Concept concept = ontology.findConcept(
                                getConceptName(shadowAssoc.getConcept()));
                String slotName = shadowAssoc.getFeature().toString();
                String slotType = getConceptName(shadowAssoc.getType());
                int mult = shadowAssoc.getMultiplicity();
                String documentation = ""; // TODO get documentation
                String annotation =""; // TODO get annotation
                boolean isMultiple = false;
                boolean isRequired = false;
                switch (mult) {
                case ShadowAssoc.ONE_OR_MORE:
                    isMultiple = true;
                    slotName += "s";
                    isRequired = true;
                    break;
                case ShadowAssoc.ZERO_OR_MORE:
                    isMultiple = true;
                    slotName += "s";
                    isRequired = false;
                    break;
                case ShadowAssoc.ZERO_OR_ONE:
                    isMultiple = false;
                    isRequired = false;
                    break;
                }

                concept.addSlot(new Slot(slotName, slotType, documentation, annotation,
                                isMultiple, isRequired));
            }
        }

        // Import slots in description.
        for (ShadowNetElement elem : net.elements()) {
            if (elem instanceof ShadowConcept) {
                ShadowConcept shadowConcept = (ShadowConcept) elem;
                Concept concept = ontology
                                .findConcept(getConceptName(shadowConcept));
                List<String> slotDocumentation = shadowConcept
                                .getSlotDocumentation();
                
                List<String> slotAnnotation = shadowConcept.getSlotAnnotation();

                // parse lines:
                String[] slotDefs = shadowConcept.getApprop().split("\n");
                int i = 0;
                for (String slotDef : slotDefs) {
                    if (!slotDef.trim().equals("")) {
                        String[] nameType = slotDef.split(":");
                        String slotName = nameType[0].trim();
                        String slotType = nameType[1].trim();
                        boolean isMultiple = false;
                        boolean isRequired = false;

                        // TODO check for required slots, too
                        if (slotType.endsWith("*")) {
                            isMultiple = true;
                            slotType = slotType.substring(0,
                                            slotType.length() - 1);
                        }
                        String documentation = "";
                        if (slotDocumentation.size() > i) {
                            documentation = slotDocumentation.get(i);
                        }
                        
                        String annotation = "";
                        if(slotAnnotation.size() > i) {
                            annotation = slotAnnotation.get(i);
                        }
                        
                        i++;
                        concept.addSlot(new Slot(slotName, slotType,
                                        documentation, annotation, isMultiple, isRequired));
                    }
                }
            }
        }

        return ontology;
    }

    /**
     * Finds the declaration node within a shadow net.
     */
    private ShadowDeclarationNode findDeclarationNode(ShadowNet net) {
        for (ShadowNetElement elem : net.elements()) {
            if (elem instanceof ShadowDeclarationNode) {
                return (ShadowDeclarationNode) elem;
            }
        }

        return null;
    }

    /**
     * @param declarations description lines found in declaration node.
     */
    private String findOntologyPackage(String[] declarations) {
        if (declarations[0].startsWith(PACKAGE_STATEMENT)) {
            return declarations[0].substring(PACKAGE_STATEMENT.length()).trim();
        }

        throw new RuntimeException(
                        "first description line must be a package declaration: \""
                                        + declarations[0] + "\"");
    }

    /**
     * Finds a ontology concept by a shadow concept.
     *
     * @param ontology the ontology to add the concept to.
     * @param superConceptMap
     * @param shadowConcept
     * @return
     */
    private Concept findConcept(Ontology ontology,
                                Map<ShadowConcept, ShadowConcept> superConceptMap,
                                ShadowConcept shadowConcept) {
        // Skip a concept from another namespaces.
        if (!shadowConcept.getNamespace().equals(ontology.getName())) {
            return null;
        }

        Concept concept = null;
        String conceptName = getConceptName(shadowConcept);
        boolean forceAbstract = getConceptForceAbstract(shadowConcept);

        // only do concepts that were not yet imported
        if (ontology.findConcept(conceptName) == null) {
            ShadowConcept superShadowConcept = superConceptMap
                            .get(shadowConcept);
            Concept superConcept = null;

            if (superShadowConcept != null) {
                String superName = getConceptName(superShadowConcept);
                String superOntoName = superShadowConcept.getNamespace();

                // check if superconcept is from external ontology
                if (!superOntoName.equals(ontology.getName())) {
                    // find superconcept in referenced ontology
                    for (Ontology refOnto : ontology
                                    .getReferencedOntologies()) {
                        if (superOntoName.equals(refOnto.getName())) {
                            superConcept = refOnto.findConcept(superName);
                            if (superConcept == null) {
                                throw new RuntimeException("External concept "
                                                + superName
                                                + " does not exist in referenced ontology "
                                                + refOnto.getName());
                            }
                        }
                    }
                } else {
                    // find superconcept in current ontology
                    superConcept = ontology.findConcept(
                                    getConceptName(superShadowConcept));

                    // if superconcept is not found, it still has to be
                    // imported
                    if (superConcept == null) {
                        superConcept = findConcept(ontology, superConceptMap,
                                        superShadowConcept);
                    }
                }

                if (superConcept == null) {
                    throw new RuntimeException(
                                    "superconcept could not be found/imported");
                }
            }

            String name = getConceptName(shadowConcept);
            String documentation = shadowConcept.getConceptDocumentation();
            if (documentation == null) {
                documentation = "";
            }
            
            String annotation = shadowConcept.getConceptAnnotation();
            if(annotation == null) {
                annotation = "";
            }
            concept = new Concept(ontology, name, superConcept, forceAbstract,
                            documentation, annotation);
            ontology.addConcept(concept);
        }

        return concept;
    }

    /**
     * Returns the filtered concept name for a shadow concept.
     */
    private String getConceptName(ShadowConcept shadowConcept) {
        String conceptName = shadowConcept.getName();
        if (conceptName.contains("\n")) {
            String[] lines = conceptName.split("\n");
            conceptName = lines[lines.length - 1].trim();
        }

        return conceptName;
    }

    /**
     * Checks whether a shadow concept must be abstract.
     */
    private boolean getConceptForceAbstract(ShadowConcept shadowConcept) {
        return doesConceptNameContainStereotype(shadowConcept,
                        ABSTRACT_STEREOTYPE);
    }

    /**
     * Checks whether a shadow concept is flagged by a stereotype.
     */
    private boolean doesConceptNameContainStereotype(ShadowConcept shadowConcept,
                                                     String stereotype) {
        String conceptName = shadowConcept.getName();
        String[] lines = conceptName.split("\n");

        for (int i = 0; i < lines.length - 1; i++) {
            String possibleStereotype = lines[i].trim();
            if (possibleStereotype.equals(stereotype)) {
                return true;
            }

            if (!possibleStereotype.equals(ABSTRACT_STEREOTYPE)) {
                throw new RuntimeException(
                                "Illegal stereotype given: " + conceptName);
            }
        }

        return false;
    }

    /**
     * Sets up the shadow net loader for FS ontology providers.
     */
    private void setupShadowNetLoader() {
        StringBuilder netPathBuilder = new StringBuilder();
        for (File file : this.sourceFiles) {
            String net = file.toString();
            String loc = StringUtil.convertToSystem(StringUtil
                            .getPath(StringUtil.convertToSlashes(net)));
            netPathBuilder.append(loc);
            netPathBuilder.append(File.pathSeparator);
        }

        // Add user net path.
        if (netPath != null) {
            for (String pathElement : netPath) {
                netPathBuilder.append(File.pathSeparator);
                netPathBuilder.append(pathElement);
            }
        }

        // Configure shadow net loader.
        Properties props = new Properties();
        props.setProperty("de.renew.netPath", netPathBuilder.toString());
        ShadowNetLoader shadowNetLoader = new DefaultShadowNetLoader(props);

        // Set net loader on FSOntologyProvider.
        setNetLoader(shadowNetLoader);
    }

}