package de.renew.rgbase.elements;

import de.renew.net.NetInstance;

import java.io.Serializable;
import java.util.*;

/**
 * A node in the reachability graph
 * representing the instances of a root net instance and all contained net instances.
 * Nodes store both outgoing edges as well as predecessor nodes for model checking.
 *
 * @author Michael Simon
 */
public abstract class Node implements Serializable {
    private static final long serialVersionUID = 9208206770140563018L;
    protected transient Collection<Edge> edges; // null until exploreEdges method is called.  Do not persist.
    transient Set<Node> predecessors;           // null until the first predecessor is added. Do not persist.

    private ArrayList<Short> labels;            // null until first element is added.


    /**
     * Returns a Collection of all outgoing edges of the node.
     *
     * @return A Collection of the outgoing edges
     */
    public Collection<Edge> getEdges() {
        if (edges == null) {
            return Collections.emptySet();
        }
        return edges;
    }

    /**
     * Adds an outgoing edge to the Node.
     *
     * @param edge An edge of type Edge.
     */
    public void addEdge(Edge edge) {
        if (edges == null) {
            edges = new HashSet<>();
        }
        edges.add(edge);
    }

    /**
     * Returns a set of all predecessor nodes of the node.
     *
     * @return A Set of all predecessor nodes.
     */
    public Set<Node> getPredecessors() {
        if (predecessors == null) {
            return Collections.emptySet();
        }
        return predecessors;
    }

    /**
     * Adds a predecessor node.
     *
     * @param node A node of Type Node.
     */
    public void addPredecessor(Node node) {
        if (predecessors == null) {
            predecessors = new HashSet<Node>();
        }
        predecessors.add(node);
    }

    public abstract Collection<? extends NetInstance> getNetInstances();

    public abstract NetInstance getRootNetInstance();

    public abstract long getDepth();

    /**
     * Adds a label.
     *
     * @param label A label of type Short.
     */
    public void addLabel(Short label){
        if (labels==null){
            labels = new ArrayList<>();
        }
        labels.add(label);
    }

    /**
     * Verifies if the ArrayList of labels contains the given label.
     *
     * @param label A given label of type Short.
     * @return true if the list contains the given label, false else.
     */
    public boolean hasLabel(Short label){
        if (labels==null){
            return false;
        }
        return labels.contains(label);
    }

    /**
     * Resets the list of labels.
     */
    public void resetLabels(){
        labels=null;
    }
}