package de.renew.rgbase.gui;

import CH.ifa.draw.figures.ArrowTip;
import CH.ifa.draw.figures.AttributeFigure;
import CH.ifa.draw.figures.LineConnection;
import CH.ifa.draw.figures.TextFigure;
import CH.ifa.draw.framework.ConnectionFigure;
import CH.ifa.draw.framework.Connector;
import CH.ifa.draw.framework.Figure;
import CH.ifa.draw.framework.FigureEnumeration;
import CH.ifa.draw.standard.StandardDrawing;
import CH.ifa.draw.util.Storable;
import CH.ifa.draw.util.StorableInput;
import de.renew.gui.AnnealingGraphLayout;
import de.renew.gui.GraphLayout;
import de.renew.gui.LayoutableDrawing;
import de.renew.rgbase.elements.Edge;
import de.renew.rgbase.elements.InscribedEdge;
import de.renew.rgbase.elements.Node;

import java.awt.*;
import java.io.IOException;
import java.util.List;
import java.util.*;

/**
 * @author Michael Simon
 */
public class GraphDrawing extends StandardDrawing implements LayoutableDrawing {
    public final static int INSC_NOTHING = 0;
    public final static int INSC_ROOT = 1;
    public final static int INSC_ALL = 2;
    public final static int INSC_DEPTH = 3;
    public final static String[] INSC_MODES = { "Nothing", "State of Root Net Instance", "State of all Net Instances", "Depth" };
    private static final long serialVersionUID = 0L;
    private static final int CONNECTION_LENGTH = 90;
    public static org.apache.log4j.Logger logger = org.apache.log4j.Logger
                    .getLogger(GraphDrawing.class);
    private boolean displayEdgeInscriptions = true;
    private final Color regularNodeColor = Color.WHITE;
    private final Color startNodeColor = Color.ORANGE;


    /**
     * For loading as a {@link Storable}.
     */
    public GraphDrawing() {
    }

    public GraphDrawing(Aborter aborter, final Collection<Node> nodes, final Node startNode,
                        int inscMode, boolean displayEdgeInscriptions) {
        this.displayEdgeInscriptions = displayEdgeInscriptions;
        Map<Node, AttributeFigure> nodeFigures = new HashMap<>();

        for (Node node : nodes) {
                if (aborter.abort()) return;
                addExploredFigure(node, nodeFigures, inscMode);
                nodeFigures.get(node).setFillColor(regularNodeColor);
        }

        nodeFigures.get(startNode).setFillColor(startNodeColor);


        LinkedList<LineConnection> loops = new LinkedList<>();

        for (Node node : nodes) {
            if (aborter.abort()) return;
            addEdges(node, nodeFigures, loops);
        }

        (new AnnealingGraphLayout()).annealingLayout(this);

        //Draw loop edges after graph is layouted
        drawLoops(loops);
    }

    protected void drawLoops(LinkedList<LineConnection> loops) {
        for (LineConnection connection : loops){
            final Point startPoint = connection.startPoint();
            int x = startPoint.x;
            int y = startPoint.y;
            connection.insertPointAt(new Point(x-10,y+25),1);
            connection.insertPointAt(new Point(x-20,y+10),2);
            connection.updateConnection();
        }
    }

    protected ExploredNodeFigure addExploredFigure(Node node,
                                                   Map<Node, AttributeFigure> nodeFigures,
                                                   int inscMode) {
        ExploredNodeFigure nodeFig = new ExploredNodeFigure(node);
        nodeFigures.put(node, nodeFig);
        add(nodeFig);

        processNode(node, nodeFig);

        if (inscMode != INSC_NOTHING) {
            TextFigure fig = nodeFig.createTextFigure(inscMode);
            add(fig);
            Rectangle d = fig.displayBox();
            int width = d.width;
            int height = d.height;
            nodeFig.displayBox(new Point(0, 0),
                    new Point(width + 20, height + 20));
        }

        return nodeFig;
    }


    protected AttributeFigure addUnexploredFigure() {
        AttributeFigure nodeFig = new UnexploredNodeFigure();
        add(nodeFig);

        return nodeFig;
    }

    protected void addEdges(Node node, Map<Node, AttributeFigure> nodeFigures, List<LineConnection> loops) {
        Collection<Edge> edges = node.getEdges();
        for (Edge edge : edges) {
            Node targetNode = edge.getTargetNode();
            AttributeFigure nextFig;
            if (targetNode == null) {
                // unexplorable
                nextFig = addUnexploredFigure();
            } else {
                // explorable
                nextFig = nodeFigures.get(targetNode);
            }

            if(nodeFigures.get(node) == nextFig){
                connectLoop(nextFig, edge, loops);
            }else {
                connect(nodeFigures.get(node), nextFig, edge);
            }
        }
    }

    protected void connect(AttributeFigure start, AttributeFigure end,
                         Edge edge) {
        // Create the connection from start to end.
        LineConnection connection = new LineConnection(null, new ArrowTip(),
                        "");
        connection.startPoint(0, 0);
        connection.endPoint(0, 0);
        connection.connectStart(start.connectorAt(start.center()));
        connection.connectEnd(end.connectorAt(end.center()));
        connection.updateConnection();

        // Add the connection to the displayed figures.
        add(connection);

        processConnection(edge, connection);

        if (displayEdgeInscriptions && edge instanceof InscribedEdge) {
            // Create inscription with the given text at the center of the connection.
            EdgeInscription edgeInscription = new EdgeInscription(
                            (InscribedEdge) edge, connection);
            // Add the inscription to the displayed figures.
            add(edgeInscription);
        }
    }


    protected void connectLoop(AttributeFigure node, Edge edge, List<LineConnection> loops){
        // Create the connection from start to end.
        LineConnection connection = new LineConnection(null, new ArrowTip(),
                "");
        connection.startPoint(0, 0);
        connection.endPoint(0, 0);
        final Connector connector = node.connectorAt(node.center());
        connection.connectStart(connector);
        connection.connectEnd(connector);
        connection.updateConnection();

        // Add the connection to the displayed figures.
        add(connection);
        processConnection(edge, connection);

        loops.add(connection);


        if (displayEdgeInscriptions && edge instanceof InscribedEdge) {
            // Create inscription with the given text at the center of the connection.
            EdgeInscription edgeInscription = new EdgeInscription(
                    (InscribedEdge) edge, connection);
            // Add the inscription to the displayed figures.
            add(edgeInscription);
        }
    }

    /* Used for subclasses to process the connection */
    protected void processConnection(Edge edge, LineConnection connection) { }
    protected void processNode(Node node, AttributeFigure nodeFig) { }

    @Override
    synchronized public void fillInGraph(GraphLayout layout) {
        FigureEnumeration k = figures();

        while (k.hasMoreElements()) {
            Figure f = k.nextFigure();

            if (f instanceof ExploredNodeFigure
                            || f instanceof UnexploredNodeFigure) {
                layout.addNode(f);
            } else if (f instanceof ConnectionFigure) {
                layout.addEdge((ConnectionFigure) f, CONNECTION_LENGTH);
            }
        }
    }

    /**
     * Reads the contained figures from StorableInput.
     * <p>
     * For graph drawings generated by this version of the code
     * {@link StandardDrawing#read(StorableInput)} would suffice,
     * but for backward compatibility this method is still needed.
     */
    @Override
    public void read(StorableInput dr) throws IOException {
        int size = dr.readInt();
        fFigures = new Vector<Figure>(size);
        for (int i = 0; i < size; i++) {
            try {
                final Figure f = (Figure) dr.readStorable();

                // This is needed for backward compatibility.
                if (f instanceof EdgeInscription) {
                    final Figure parent = ((EdgeInscription) f).parent();
                    if (!fFigures.contains(parent)) {
                        add(parent);
                        i++;
                    }
                }

                add(f);
            } catch (IOException e) {
                final String desc = GraphDrawing.class.getSimpleName()
                                + ": could not read in "
                                + Figure.class.getSimpleName() + " object:";
                if (logger.isDebugEnabled()) {
                    // long version
                    logger.error(desc, e);
                } else {
                    // short version
                    logger.error(desc + " " + e);
                }
                break;
            }
        }
    }
}