package de.renew.windowmanagement;

import java.awt.*;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;

import javax.swing.*;

/**
 * responsible class for the dragging action to be displayed on the cursor.
 */
public class DraggableComponentAdapter extends MouseAdapter {

    /**
     * The parent container in which this component can be dragged via a mouse drag event.
     */
    final private DragCanvas canvas;

    /**
     * The component to be dragged.
     */
    final private Component dragTarget;

    /**
     * Default mouse cursor for dragging action
     */
    protected Cursor draggingCursor = Cursor
                    .getPredefinedCursor(Cursor.HAND_CURSOR);

    /**
     * This is the constructor for DraggableComponentAdapter.
     *
     * @param canvas is the drawable where we display our components
     * @param target is the object that can be moved around
     */
    public DraggableComponentAdapter(DragCanvas canvas, Component target) {
        this.dragTarget = target;
        this.canvas = canvas;
    }

    @Override
    public void mouseDragged(MouseEvent e) {
        if (dragTarget.isShowing()) {
            Point absolute = e.getLocationOnScreen();
            dragTarget.setCursor(draggingCursor);

            BufferedImage image = new BufferedImage(dragTarget.getWidth(),
                            dragTarget.getHeight(), BufferedImage.TYPE_INT_RGB);
            Graphics g = image.getGraphics();
            dragTarget.printAll(g);
            g.setColor(new Color(0.6f, 0.6f, 0.6f, 0.6f));
            g.fillRect(0, 0, dragTarget.getWidth(), dragTarget.getHeight());
            canvas.requestOverlayingShadow(absolute, image);
            Rectangle r = new Rectangle(e.getX(), e.getY(), 1, 1);
            ((JComponent) e.getSource()).scrollRectToVisible(r);

            canvas.repaint();
        }
    }

    @Override
    public void mouseReleased(MouseEvent e) {
        Point absolute = e.getLocationOnScreen();

        if (wasDragged(absolute)) {
            Point canvasAbsolute = canvas.getLocationOnScreen();
            Point adjustedLocation = new Point(absolute.x - canvasAbsolute.x,
                            absolute.y - canvasAbsolute.y);
            if (canvas.contains(adjustedLocation)) {
                canvas.dropComponent(dragTarget, absolute);
            }
        }

        canvas.requestOverlayingShadow(null, null);
        canvas.repaint();
        dragTarget.setCursor(Cursor.getDefaultCursor());
        canvas.validate();
    }

    /**
     * Helper method to determine if the component has actually changed positions after being dragged.
     *
     * @param newPosition
     * @return
     */
    private boolean wasDragged(Point newPosition) {
        Point oldPosition = dragTarget.getLocationOnScreen();
        if (oldPosition.x + dragTarget.getWidth() < newPosition.x
                        || oldPosition.x > newPosition.x
                        || oldPosition.y + dragTarget
                                        .getHeight() < newPosition.y
                        || oldPosition.y > newPosition.y) {
            return true;
        } else {
            return false;
        }
    }
}