package de.renew.windowmanagement;

import bibliothek.extension.gui.dock.preference.PreferenceModel;
import bibliothek.gui.dock.common.*;
import bibliothek.gui.dock.common.event.CControlListener;
import bibliothek.gui.dock.common.event.CFocusListener;
import bibliothek.gui.dock.common.event.CKeyboardListener;
import bibliothek.gui.dock.common.event.CVetoClosingListener;
import bibliothek.gui.dock.common.grouping.PlaceholderGrouping;
import bibliothek.gui.dock.common.intern.CDockFrontend;
import bibliothek.gui.dock.common.intern.CDockable;
import bibliothek.gui.dock.common.menu.CLayoutChoiceMenuPiece;
import bibliothek.gui.dock.common.menu.CLookAndFeelMenuPiece;
import bibliothek.gui.dock.common.menu.CPreferenceMenuPiece;
import bibliothek.gui.dock.common.menu.SingleCDockableListMenuPiece;
import bibliothek.gui.dock.common.perspective.CDockablePerspective;
import bibliothek.gui.dock.common.perspective.CGridPerspective;
import bibliothek.gui.dock.common.perspective.CPerspective;
import bibliothek.gui.dock.common.theme.ThemeMap;
import bibliothek.gui.dock.facile.lookandfeel.DockableCollector;
import bibliothek.gui.dock.facile.menu.RootMenuPiece;
import bibliothek.gui.dock.support.lookandfeel.ComponentCollector;
import bibliothek.gui.dock.support.lookandfeel.LookAndFeelList;
import bibliothek.gui.dock.support.menu.SeparatingMenuPiece;
import bibliothek.util.Path;
import de.renew.plugin.PluginManager;

import javax.swing.*;
import java.awt.*;
import java.io.*;


/**
 * This class provides the main frame of the Renew application.
 * It controls all contained windows and allows listening to changes to the view parts.
 */
public class WorkbenchImpl implements Workbench {

    private JFrame mainframe;
    private CControl controller;
    private RenewMenuBar fMenu;
    private JTextField fStatusLine;
    private JPanel fStaticPanel;

    /**
     * creates a constant String LOWER_LEFT_GROUP where th path for the lower left drawing layout is stored.
     */
    public final static Path LOWER_LEFT_GROUP = new Path("drawinglayout",
                    "lowerleft");
    /**
     * creates a constant String UPPER_LEFT_GROUP where th path for the upper left drawing layout is stored.
     */
    public final static Path UPPER_LEFT_GROUP = new Path("drawinglayout",
                    "upperleft");
    /**
     * creates a constant String UPPER_RIGHT_GROUP where th path for the upper right drawing layout is stored.
     */
    public final static Path UPPER_RIGHT_GROUP = new Path("drawinglayout",
                    "upperright");
    /**
     * creates a constant String LOWER_RIGHT_GROUP where th path for the lower right drawing layout is stored.
     */
    public final static Path LOWER_RIGHT_GROUP = new Path("drawinglayout",
                    "lowerright");

    private CWorkingArea fDrawingsArea;

    /**
     * static logger for the WorkbenchImpl.class
     */
    public static org.apache.log4j.Logger logger = org.apache.log4j.Logger
                    .getLogger(WorkbenchImpl.class);

    private final static int fmaxConsoleLines = 80;

    /**
     * creates the main Workbench window and the menubar for it.
     * It also binds the suitable controller fpr the mainframe
     *
     * @param menuBar the MenuBar for the main workbench window.
     * @param frame the JFrame, where you initialize the the workbench window.
     */
    public WorkbenchImpl(RenewMenuBar menuBar, JFrame frame) {

        UIManager.installLookAndFeel("Darcula",
                        "com.bulenkov.darcula.DarculaLaf");

        logger.debug("Starting Workbench.");

        fMenu = menuBar;

        fStaticPanel = new JPanel();
        fStaticPanel.setLayout(
                        new BoxLayout(fStaticPanel, BoxLayout.PAGE_AXIS));
        fStaticPanel.setBackground(Color.lightGray);

        mainframe = frame;
        mainframe.setTitle("RENEW - the Reference Net Workshop");
        mainframe.setLayout(new BorderLayout());
        mainframe.setMinimumSize(new Dimension(800, 600));
        controller = new CControl(mainframe);
        mainframe.add(controller.getContentArea());

        // There currently is no proper hotkey management.
        // To not overwrite hotkeys in use by other plugins,
        // for now disable all controller hotkeys.
        controller.putProperty(CControl.KEY_MAXIMIZE_CHANGE, null);
        controller.putProperty(CControl.KEY_GOTO_MAXIMIZED, null);
        controller.putProperty(CControl.KEY_GOTO_MINIMIZED, null);
        controller.putProperty(CControl.KEY_CANCEL_OPERATION, null);
        controller.putProperty(CControl.KEY_CLOSE, null);
        controller.putProperty(CControl.KEY_GOTO_EXTERNALIZED, null);
        controller.putProperty(CControl.KEY_GOTO_NORMALIZED, null);

        mainframe.getContentPane().add(fStaticPanel, "North");
        mainframe.setJMenuBar(fMenu);

        // This is an old workaround to dispatch key events from externalized windows to the menu
        new HotDrawFocusManager(fMenu);

        PreferenceModel preferences = new CPreferenceModel(controller);
        controller.setPreferenceModel(preferences);
        RootMenuPiece system = new RootMenuPiece("System", false);
        system.add(new SeparatingMenuPiece(
                        new CLookAndFeelMenuPiece(controller), false, false,
                        true));

        RootMenuPiece layout = new RootMenuPiece("Layout", false);
        layout.add(new CLayoutChoiceMenuPiece(controller, true));
        system.add(new SeparatingMenuPiece(layout, false, false, true));
        system.add(new CPreferenceMenuPiece(controller));

        RootMenuPiece windows = new RootMenuPiece("Windows", false);
        windows.add(new SingleCDockableListMenuPiece(controller));

        try {
            fMenu.addMenu(system.getMenu(), RenewMenuBar.SYSTEMS_SECTION);
            fMenu.addMenu(windows.getMenu(), RenewMenuBar.SYSTEMS_SECTION);
        } catch (UnknownMenuSectionException e) {
            e.printStackTrace();
        }

        mainframe.getContentPane().add(controller.getContentArea(),
                        BorderLayout.CENTER);
        fDrawingsArea = controller.createWorkingArea("drawing area");
        initializeControl(controller);

        fStatusLine = new JTextField(40);
        fStatusLine.setEditable(false);
        mainframe.add(fStatusLine, BorderLayout.SOUTH);
        showStatus("Welcome to Renew - the Reference Net Workshop");

        // Maximize screen size if possible
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        mainframe.setSize(screenSize.width, screenSize.height);
        mainframe.pack();
        mainframe.setVisible(true);

        // Update internal components after look&feel changes
        ComponentCollector collector = new DockableCollector(
                        controller.intern());
        LookAndFeelList.getDefaultList().addComponentCollector(collector);

        // Try to load settings from disk
        CDockFrontend frontend = controller.intern();
        File file = new File(PluginManager.getPreferencesLocation(),
                        "renewLayoutSettings");
        if (file.exists()) {
            try {
                FileInputStream fileInputStream = new FileInputStream(file);
                DataInputStream dataInputStream = new DataInputStream(
                                fileInputStream);
                frontend.read(dataInputStream);
            } catch (IOException e) {
                e.printStackTrace();
            }
        }


        EventQueue.invokeLater(new Runnable() {
            @Override
            public void run() {
                controller.setTheme(ThemeMap.KEY_FLAT_THEME);
            }
        });

    }

    /**
     * Provides the {@link CControl} with its initial perspective.
     * <p>
     * ____  __________  ____
     * | ul ||          || ur |
     * |____||  drawing ||____|
     * |    ||          ||    |
     * | low||   area   || low|
     * | <- ||          || -> |
     * |____||__________||____|
     *
     * @param control
     */
    private void initializeControl(CControl control) {

        logger.debug("Initializing base perspective.");

        CPerspective perspective = control.getPerspectives()
                        .createEmptyPerspective();
        CGridPerspective center = perspective.getContentArea().getCenter();

        center.gridPlaceholder(0, 0, 20, 10, UPPER_LEFT_GROUP);
        center.gridPlaceholder(0, 10, 20, 20, LOWER_LEFT_GROUP);
        center.gridAdd(20, 0, 50, 30, (CDockablePerspective) perspective
                        .getStation("drawing area"));
        center.gridPlaceholder(70, 0, 20, 10, UPPER_RIGHT_GROUP);
        center.gridPlaceholder(70, 10, 20, 20, LOWER_RIGHT_GROUP);

        control.getPerspectives().setPerspective("basePerspective", perspective,
                        true);
        control.load("basePerspective");
    }

    @Override
    public JFrame getMainFrame() {

        return mainframe;
    }

    @Override
    public void showStatus(final String string) {

        Runnable r = new Runnable() {
            @Override
            public void run() {
                fStatusLine.setText(string);
            }
        };

        if (EventQueue.isDispatchThread()) {
            r.run();
        } else {
            EventQueue.invokeLater(r);
        }
    }

    @Override
    public void addStaticPanel(JPanel panel) {
        fStaticPanel.add(panel);
        mainframe.pack();
    }

    @Override
    public void addEditorWindow(final MultipleCDockableLayout layout,
                                final String factoryID) {

        logger.debug("Adding new Editor window for layout "
                        + layout.toString());
        logger.debug("Using factoryID " + factoryID);

        Runnable r = new Runnable() {
            @Override
            public void run() {

                MultipleCDockableFactory factory = controller
                                .getMultipleDockableFactory(factoryID);
                MultipleCDockable editor = factory.read(layout);
                controller.addDockable(editor);
                editor.setWorkingArea(fDrawingsArea);

                if (editor instanceof DefaultMultipleCDockable) {
                    ((DefaultMultipleCDockable) editor).setTitleShown(true);
                }

                editor.setVisible(true);
            }
        };

        if (EventQueue.isDispatchThread()) {

            r.run();

        } else {

            EventQueue.invokeLater(r);
        }

    }

    @Override
    public void newEditorWindow(String factoryID) {

        logger.debug("Attempting to create new empty editor template for factoryID "
                        + factoryID);
        MultipleCDockableFactory factory = controller
                        .getMultipleDockableFactory(factoryID);
        addEditorWindow(factory.create(), factoryID);
    }


    @Override
    public void registerControlListener(CControlListener listener) {

        logger.debug("Add control listener " + listener.toString());
        controller.addControlListener(listener);
    }


    @Override
    public synchronized void registerEditorFactory(MultipleCDockableFactory factory,
                                                   String id) {

        logger.debug("Registering new editor factory under id " + id);

        if (controller.getMultipleDockableFactory(id) != null) {
            return;
        }
        controller.addMultipleDockableFactory(id, factory);

    }


    @Override
    public void registerMenu(JMenu menu) {
        fMenu.add(menu);
        mainframe.pack();
    }


    @Override
    public void registerMenu(JMenu menu, int index)
                    throws UnknownMenuSectionException {
        fMenu.addMenu(menu, index);
        mainframe.pack();

    }


    @Override
    public void removeMenu(JMenu menu) {
        fMenu.remove(menu);
        mainframe.pack();
    }


    @Override
    public void close(MultipleCDockable dockable) {
        dockable.setVisible(false);
        if (!dockable.isVisible()) {
            // If container is still visible, operation was canceled by user.
            controller.removeDockable(dockable);
        }
    }

    @Override
    public void close(SingleCDockable dockable) {
        dockable.setVisible(false);
        if (!dockable.isVisible()) {
            // If container is still visible, operation was canceled by user.
            controller.removeDockable(dockable);
        }
    }


    @Override
    public void addViewWindow(final DefaultSingleCDockable dockable,
                              final Path path) {

        logger.debug("Add new view " + dockable.getTitleText());
        logger.debug("At position " + path.toString());

        Runnable r = new Runnable() {
            @Override
            public void run() {
                dockable.setGrouping(new PlaceholderGrouping(controller, path));
                controller.addDockable(dockable);
                dockable.setVisible(true);
            }
        };

        if (EventQueue.isDispatchThread()) {
            r.run();
        } else {
            EventQueue.invokeLater(r);
        }

    }

    /**
     * Destroy the mainframe and controller.
     */
    protected void destroy() {

        logger.debug("Saving current layout to disk.");

        CDockFrontend frontend = controller.intern();
        File file = new File(PluginManager.getPreferencesLocation(),
                        "renewLayoutSettings");
        // Create new file if it doesn't exist
        try {
            file.createNewFile();

            FileOutputStream fileOutputStream = new FileOutputStream(file);
            DataOutputStream dataOutputStream = new DataOutputStream(
                            fileOutputStream);
            frontend.write(dataOutputStream);
        } catch (IOException e) {
            e.printStackTrace();
        }

        mainframe.setVisible(false);

        // There might be unfinished operations requiring access to the mainframe.
        // Since all of those and this method run on the awt EventQueue thread, we can ensure
        // they are finished and the proxy is freed before finally destroying the mainframe.
        EventQueue.invokeLater(new Runnable() {
            @Override
            public void run() {
                logger.debug("Destroying controller and mainframe.");

                controller.destroy();
                mainframe.dispose();
            }
        });

    }

    public void registerShutdownHook(RenewShutdownHook hook) {
        // This is done in the proxy, so no need for this method here.
    }

    @Override
    public void openGui() {
        mainframe.setVisible(true);
    }

    @Override
    public void registerFocusListener(CFocusListener listener) {
        controller.addFocusListener(listener);
    }

    @Override
    public void registerVetoClosingListener(CVetoClosingListener listener) {
        controller.addVetoClosingListener(listener);
    }


    public void registerKeyboardListener(CKeyboardListener listener) {
        controller.addKeyboardListener(listener);
    }
}
