package de.renew.windowmanagement;

import bibliothek.gui.dock.common.*;
import bibliothek.gui.dock.common.event.CControlListener;
import bibliothek.gui.dock.common.event.CFocusListener;
import bibliothek.gui.dock.common.event.CKeyboardListener;
import bibliothek.gui.dock.common.event.CVetoClosingListener;
import bibliothek.util.Path;

import javax.swing.*;
import java.awt.*;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Map;

/**
 * This class is the port to access to a workbench.
 * It manages different listeners and functions to a Workbench.
 */
public class WorkbenchProxy implements Workbench {

    private boolean fwithGui = false;
    private WorkbenchImpl proxy;
    private HashMap<String, MultipleCDockableFactory> fFactoryMap = new HashMap();

    private LinkedHashSet<JPanel> fPanelSet = new LinkedHashSet<>();
    private ArrayList<MenuBarEntry> fmenu = new ArrayList<>();
    private HashMap<DefaultSingleCDockable, Path> fViewMap = new HashMap();
    private final ArrayList<RenewShutdownHook> fShutdownHooks = new ArrayList<RenewShutdownHook>();

    // Registered Listeners. They are kept as weak references, so plugins don't have to deregister.
    private LinkedHashSet<WeakReference> fControlListeners = new LinkedHashSet<>();
    private LinkedHashSet<WeakReference> fFocusListeners = new LinkedHashSet<>();
    private LinkedHashSet<WeakReference> fVetoListeners = new LinkedHashSet<>();
    private LinkedHashSet<WeakReference> fKeyboardListeners = new LinkedHashSet<>();

    @Override
    public JFrame getMainFrame() {
        if (fwithGui) {
            return ensureProxyInitialized().getMainFrame();
        }

        return null;
    }

    @Override
    public void showStatus(String string) {
        if (fwithGui) {
            ensureProxyInitialized().showStatus(string);
        }
    }

    @Override
    public void addStaticPanel(JPanel panel) {
        if (fwithGui) {
            ensureProxyInitialized().addStaticPanel(panel);
        } else {
            fPanelSet.add(panel);
        }
    }

    @Override
    public void addEditorWindow(MultipleCDockableLayout editor,
                                String factoryID) {
        if (fwithGui) {
            ensureProxyInitialized().addEditorWindow(editor, factoryID);
        }
    }

    @Override
    public void newEditorWindow(String factoryID) {
        if (fwithGui) {
            ensureProxyInitialized().newEditorWindow(factoryID);
        }
    }

    @Override
    public void registerControlListener(CControlListener listener) {
        if (fwithGui) {
            ensureProxyInitialized().registerControlListener(listener);
        } else {
            fControlListeners.add(new WeakReference(listener));
        }
    }

    @Override
    public void registerEditorFactory(MultipleCDockableFactory factory,
                                      String id) {
        if (fwithGui) {
            ensureProxyInitialized().registerEditorFactory(factory, id);
        } else {
            fFactoryMap.put(id, factory);
        }

    }

    @Override
    public void registerMenu(JMenu menu) {

        fmenu.add(new MenuBarEntry(menu, RenewMenuBar.MISC_SECTION));
        if (fwithGui) {
            ensureProxyInitialized().registerMenu(menu);
        }
    }

    @Override
    public void registerMenu(JMenu menu, int index)
                    throws UnknownMenuSectionException {

        fmenu.add(new MenuBarEntry(menu, index));
        if (fwithGui) {
            ensureProxyInitialized().registerMenu(menu, index);
        }
    }

    @Override
    public void removeMenu(JMenu menu) {
        fmenu.remove(new MenuBarEntry(menu, 0));

        //invokeLater is needed, because ensureProxyFreed()
        //would be called too late resulting in fwithGui still being
        //true while the proxy was already destroyed
        EventQueue.invokeLater(() -> {
            if (fwithGui) {
                ensureProxyInitialized().removeMenu(menu);
            }
        });
    }

    @Override
    public void close(MultipleCDockable dockable) {
        if (fwithGui) {
            ensureProxyInitialized().close(dockable);
        }
    }

    @Override
    public void close(SingleCDockable dockable) {
        //invokeLater is needed, because ensureProxyFreed()
        //would be called too late resulting in fwithGui still being
        //true while the proxy was already destroyed
        EventQueue.invokeLater(() -> {
            if (fwithGui) {
                ensureProxyInitialized().close(dockable);
            }
        });
    }

    @Override
    public void addViewWindow(DefaultSingleCDockable dockable, Path path) {
        if (fwithGui) {
            ensureProxyInitialized().addViewWindow(dockable, path);
        } else {
            fViewMap.put(dockable, path);
        }
    }

    /**
     * Request closing of this application.
     * All shutdown hooks will be informed.
     * Any hook objecting to closing the application may intervene.
     */
    protected synchronized void requestClose() {

        assert EventQueue.isDispatchThread();

        if (!fwithGui) {
            return;
        }

        for (RenewShutdownHook hook : fShutdownHooks) {

            if (hook.canClose()) {
                hook.exit();
            } else {
                // Some hook objected and the process is canceled.
                // Currently no rollback is done.
                fwithGui = true;
                return;
            }

        }

        proxy.destroy();
        ensureProxyFreed();
    }

    @Override
    public void registerShutdownHook(RenewShutdownHook hook) {
        fShutdownHooks.add(hook);
    }

    private synchronized WorkbenchImpl ensureProxyInitialized() {

        if (proxy == null) {

            RenewMenuBar bar = new RenewMenuBar();
            for (MenuBarEntry entry : fmenu) {

                try {
                    bar.addMenu(entry._menu, entry._section);
                } catch (UnknownMenuSectionException e) {
                    e.printStackTrace();
                }
            }

            JFrame frame = new JFrame();
            addWindowListeners(frame);

            proxy = new WorkbenchImpl(bar, frame);

            EventQueue.invokeLater(new Runnable() {
                @Override
                public void run() {

                    for (Map.Entry<String, MultipleCDockableFactory> entryset : fFactoryMap
                                    .entrySet()) {
                        proxy.registerEditorFactory(entryset.getValue(),
                                        entryset.getKey());
                    }

                    for (WeakReference listener : fControlListeners) {
                        proxy.registerControlListener(
                                        (CControlListener) listener.get());
                    }

                    for (WeakReference listener : fFocusListeners) {
                        proxy.registerFocusListener(
                                        (CFocusListener) listener.get());
                    }

                    for (WeakReference listener : fVetoListeners) {
                        proxy.registerVetoClosingListener(
                                        (CVetoClosingListener) listener.get());
                    }

                    for (WeakReference listener : fKeyboardListeners) {
                        proxy.registerKeyboardListener(
                                        (CKeyboardListener) listener.get());
                    }

                    for (JPanel panel : fPanelSet) {
                        proxy.addStaticPanel(panel);
                    }

                    for (Map.Entry<DefaultSingleCDockable, Path> view : fViewMap
                                    .entrySet()) {
                        proxy.addViewWindow(view.getKey(), view.getValue());
                    }
                }
            });

        }

        return proxy;
    }

    private void ensureProxyFreed() {
        proxy = null;
        fwithGui = false;
    }


    public void openGui() {
        fwithGui = true;
        ensureProxyInitialized();
    }

    @Override
    public void registerFocusListener(CFocusListener listener) {
        if (fwithGui) {
            ensureProxyInitialized().registerFocusListener(listener);
        } else {
            fFocusListeners.add(new WeakReference<CFocusListener>(listener));
        }
    }

    @Override
    public void registerVetoClosingListener(CVetoClosingListener listener) {
        if (fwithGui) {
            ensureProxyInitialized().registerVetoClosingListener(listener);
        } else {
            fVetoListeners.add(new WeakReference(listener));
        }
    }


    public void registerKeyboardListener(CKeyboardListener listener) {
        if (fwithGui) {
            ensureProxyInitialized().registerKeyboardListener(listener);
        } else {
            fKeyboardListeners.add(new WeakReference(listener));
        }
    }

    /**
     * Registers the listeners for the main window
     *
     * @param frame the frame where you want to add the listener to.
     */
    protected void addWindowListeners(JFrame frame) {
        frame.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
        frame.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent event) {
                EventQueue.invokeLater(new Runnable() {
                    @Override
                    public void run() {
                        requestClose();
                    }
                });
            }

            public void windowDeactivated(WindowEvent event) {
            }
        });
    }

    /*
     * Data container for toolbar entries
     */
    private class ToolbarEntry {
        protected String _name;
        protected Map<AbstractButton, Component> _buttonBoxMap;
        protected boolean _immediatelyVisible;

        ToolbarEntry(String name, Map<AbstractButton, Component> buttonBoxMap,
                     boolean immediatelyVisible) {
            _name = name;
            _buttonBoxMap = buttonBoxMap;
            _immediatelyVisible = immediatelyVisible;
        }

        @Override
        public boolean equals(Object otherEntry) {

            if (otherEntry instanceof ToolbarEntry) {
                return ((ToolbarEntry) otherEntry)._name == this._name;
            }
            return false;
        }
    }

    private class MenuBarEntry {

        protected JMenu _menu;
        protected int _section;

        MenuBarEntry(JMenu menu, int section) {
            _menu = menu;
            _section = section;
        }

        @Override
        public boolean equals(Object o) {
            if (o instanceof MenuBarEntry) {
                return ((MenuBarEntry) o)._menu == _menu;
            }
            return false;
        }
    }
}
